Package: aemo
Title: Download Australian Energy Market Operator Data
Version: 0.4.1
Authors@R: 
    person("Charles", "Coverdale", , "charlesfcoverdale@gmail.com", role = c("aut", "cre"))
Description: Fetch Australian Energy Market Operator (AEMO)
    public data from 'NEMweb' <http://nemweb.com.au> and the
    Market Management System Data Model (MMSDM) historical
    archive. Provides tidy access to 5-minute and 30-minute
    wholesale electricity prices, regional demand, dispatch-unit
    output, interconnector flows, rooftop photovoltaic generation,
    generator bids, predispatch forecasts, frequency control
    ancillary services markets, and gas market data across the
    National Electricity Market (NEM) regions. Data is published
    by AEMO under its Copyright Permissions Notice
    <https://www.aemo.com.au/privacy-and-legal-notices/copyright-permissions>.
Depends: R (>= 4.1.0)
License: MIT + file LICENSE
Encoding: UTF-8
Language: en-US
RoxygenNote: 7.3.3
Imports: cli (>= 3.6.0), httr2 (>= 1.0.0), openssl (>= 2.0.0), tools,
        utils
Suggests: knitr, rmarkdown, testthat (>= 3.0.0), withr
VignetteBuilder: knitr
Config/testthat/edition: 3
URL: https://charlescoverdale.github.io/aemo/,
        https://github.com/charlescoverdale/aemo
BugReports: https://github.com/charlescoverdale/aemo/issues
NeedsCompilation: no
Packaged: 2026-05-27 07:54:35 UTC; charlescoverdale
Author: Charles Coverdale [aut, cre]
Maintainer: Charles Coverdale <charlesfcoverdale@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-27 08:10:02 UTC
Built: R 4.5.2; ; 2026-05-27 11:34:39 UTC; unix

Package: aemo
Title: Download Australian Energy Market Operator Data
Version: 0.4.1
Authors@R: 
    person("Charles", "Coverdale", , "charlesfcoverdale@gmail.com", role = c("aut", "cre"))
Description: Fetch Australian Energy Market Operator (AEMO)
    public data from 'NEMweb' <http://nemweb.com.au> and the
    Market Management System Data Model (MMSDM) historical
    archive. Provides tidy access to 5-minute and 30-minute
    wholesale electricity prices, regional demand, dispatch-unit
    output, interconnector flows, rooftop photovoltaic generation,
    generator bids, predispatch forecasts, frequency control
    ancillary services markets, and gas market data across the
    National Electricity Market (NEM) regions. Data is published
    by AEMO under its Copyright Permissions Notice
    <https://www.aemo.com.au/privacy-and-legal-notices/copyright-permissions>.
Depends: R (>= 4.1.0)
License: MIT + file LICENSE
Encoding: UTF-8
Language: en-US
RoxygenNote: 7.3.3
Imports: cli (>= 3.6.0), httr2 (>= 1.0.0), openssl (>= 2.0.0), tools,
        utils
Suggests: knitr, rmarkdown, testthat (>= 3.0.0), withr
VignetteBuilder: knitr
Config/testthat/edition: 3
URL: https://charlescoverdale.github.io/aemo/,
        https://github.com/charlescoverdale/aemo
BugReports: https://github.com/charlescoverdale/aemo/issues
NeedsCompilation: no
Packaged: 2026-05-27 07:54:35 UTC; charlescoverdale
Author: Charles Coverdale [aut, cre]
Maintainer: Charles Coverdale <charlesfcoverdale@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-27 08:10:02 UTC
Built: R 4.5.2; ; 2026-05-27 11:10:48 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(print,aemo_tbl)
export(aemo_bids)
export(aemo_cache_info)
export(aemo_clear_cache)
export(aemo_constraints)
export(aemo_demand)
export(aemo_dispatch_units)
export(aemo_dlf)
export(aemo_fcas)
export(aemo_fcas_enablement)
export(aemo_gas)
export(aemo_gencon)
export(aemo_interconnector)
export(aemo_interconnectors)
export(aemo_market_notices)
export(aemo_mlf)
export(aemo_nemweb_download)
export(aemo_nemweb_ls)
export(aemo_outages)
export(aemo_participants)
export(aemo_pasa)
export(aemo_predispatch)
export(aemo_price)
export(aemo_price_caps)
export(aemo_regions)
export(aemo_rooftop_pv)
export(aemo_settlement)
export(aemo_snapshot)
export(aemo_spd_constraints)
export(aemo_throttle)
export(aemo_units)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval=FALSE---------------------------------------------------------------
# library(aemo)
# 
# # All timestamps come back in AEST (UTC+10, no DST).
# # The tzone attribute is "Australia/Brisbane" throughout.
# p <- aemo_price("NSW1", "2024-01-15 14:00", "2024-01-15 15:00")
# attr(p$settlementdate, "tzone")  # "Australia/Brisbane"
# 
# # January is summer in Australia. Australia/Sydney would give UTC+11 here,
# # shifting every timestamp one hour ahead of the AEMO file clock.
# # Australia/Brisbane correctly gives UTC+10.
# format(p$settlementdate[1], "%Y-%m-%d %H:%M:%S %Z")

## ----eval=FALSE---------------------------------------------------------------
# # Convert market time to NSW wall clock (observes DST)
# p$local_time_nsw <- format(p$settlementdate, tz = "Australia/Sydney")
# 
# # Convert market time to SA wall clock (UTC+9:30 / UTC+10:30 with DST)
# p$local_time_sa <- format(p$settlementdate, tz = "Australia/Adelaide")

## ----eval=FALSE---------------------------------------------------------------
# # Period-start timestamps for 5-min intervals
# p$period_start <- p$settlementdate - as.difftime(5, units = "mins")
# 
# # Period-start timestamps for 30-min trading intervals
# t$period_start <- t$settlementdate - as.difftime(30, units = "mins")

## ----eval=FALSE---------------------------------------------------------------
# # Default: market pricing run only
# p_market <- aemo_price("NSW1", "2022-06-13", "2022-06-14")
# 
# # Both runs (for intervention analysis)
# p_both <- aemo_price("NSW1", "2022-06-13", "2022-06-14",
#                      intervention = TRUE)
# table(p_both$intervention)

# Price functions

# 5-minute settlement commenced 1 October 2021 (NER cl. 3.15.1A).
# Before this date 30-min prices came from TRADINGPRICE (TRADINGIS).
# After this date TRADINGIS is still published but the authoritative
# 30-min price for post-5MS periods is the arithmetic mean of the
# six 5-min dispatch prices within each trading interval.
FIVE_MS_START <- as.POSIXct("2021-10-01 00:00:00", tz = "Australia/Brisbane")

#' Wholesale electricity prices
#'
#' Returns 5-minute dispatch prices or 30-minute trading prices
#' for a NEM region over a specified window. Filters intervention
#' runs by default so the returned prices are the market clearing
#' prices used in settlement.
#'
#' **Timestamps** are AEST (UTC+10, no daylight savings) to match
#' AEMO's market clock. See the package-level documentation for
#' the period-ending timestamp convention (a row stamped 00:05 is
#' the 5-minute period ending at 00:05).
#'
#' **Intervention.** `DISPATCHPRICE` contains both market pricing
#' runs (`INTERVENTION = 0`) and physical / intervention runs
#' (`INTERVENTION = 1`). The default filters to market runs.
#' Pass `intervention = TRUE` to get both.
#'
#' **30-minute settlement and the 5MS transition.** Before
#' 1 October 2021 the NEM settled on 30-minute trading prices
#' from `TRADINGPRICE` (TRADINGIS). On 1 October 2021 five-minute
#' settlement (5MS) commenced and settlement moved to native
#' 5-minute prices. When `interval = "30min"`:
#' - For the pre-5MS period (`start < 2021-10-01`): prices are
#'   read from TRADINGIS (`TRADINGPRICE`).
#' - For the post-5MS period: prices are derived by taking the
#'   arithmetic mean of the six 5-minute dispatch prices within
#'   each 30-minute trading interval, consistent with how AEMO
#'   calculates the `TRADINGPRICE` column in TradingIS post-5MS.
#'
#' **Data availability.** NEMweb Current-directory files retain
#' the last ~30 days of 5-minute dispatch files. Historical
#' queries use the Archive daily-rollup files automatically; for
#' queries older than the Archive window, use
#' [aemo_nemweb_download()] with an MMSDM URL directly.
#'
#' @param region One of `"NSW1"`, `"QLD1"`, `"SA1"`, `"TAS1"`,
#'   `"VIC1"`. Accepts a vector.
#' @param start,end Start and end times (inclusive). Character
#'   (parsed as AEST) or `POSIXct`.
#' @param interval One of `"5min"` (default) or `"30min"`.
#' @param market One of `"energy"` (default, returns RRP) or
#'   `"fcas"` (returns the FCAS service RRPs).
#' @param intervention Logical. `FALSE` (default) returns only
#'   the market pricing run; `TRUE` returns both market and
#'   physical runs, with the `intervention` column preserved.
#'
#' @return An `aemo_tbl`. Key columns include `settlementdate`
#'   (POSIXct AEST), `regionid`, `rrp` (AUD/MWh, energy) or the
#'   FCAS service RRPs (AUD/MW), and `intervention`.
#'
#' @source AEMO NEMweb <http://nemweb.com.au>, AEMO Copyright
#'   Permissions Notice.
#'
#' @family price
#' @export
#' @examples
#' \donttest{
#' op <- options(aemo.cache_dir = tempdir())
#' try({
#'   now <- Sys.time()
#'   p <- aemo_price("NSW1", now - 3600, now)
#'   head(p)
#' })
#' options(op)
#' }
aemo_price <- function(region, start, end,
                        interval = c("5min", "30min"),
                        market = c("energy", "fcas"),
                        intervention = FALSE) {
  region <- aemo_validate_region(region)
  interval <- match.arg(interval)
  market <- match.arg(market)
  start <- aemo_parse_time(start)
  end <- aemo_parse_time(end)
  stopifnot(end >= start)

  if (interval == "5min") {
    df <- aemo_fetch_report_range(
      current_dir = "/Reports/Current/DispatchIS_Reports/",
      archive_dir = "/Reports/Archive/DispatchIS_Reports/",
      pattern = "DISPATCHIS",
      start = start, end = end,
      table = "dispatch_price"
    )
    df <- aemo_coerce_types(df)
    df <- aemo_apply_filters(df, start = start, end = end,
                             region = region, intervention = intervention)
  } else {
    df <- aemo_price_30min(region = region, start = start, end = end,
                           intervention = intervention)
  }

  if (market == "fcas") {
    fcas_cols <- grep("(raise|lower).*rrp$", names(df),
                      value = TRUE, ignore.case = TRUE)
    keep <- intersect(
      c("settlementdate", "regionid", "runno", "intervention", fcas_cols),
      names(df)
    )
    df <- df[, keep, drop = FALSE]
  }

  new_aemo_tbl(df,
               source = "http://nemweb.com.au",
               title = sprintf("AEMO %s %s price %s",
                                interval, market,
                                paste(region, collapse = "+")))
}

#' Fetch 30-minute prices spanning the 5MS transition boundary.
#'
#' Pre-5MS (before 2021-10-01): TRADINGIS, table `trading_price`.
#' Post-5MS (from 2021-10-01): aggregate 6 x 5-min dispatch prices
#' within each 30-min trading interval (arithmetic mean of RRP).
#' @noRd
aemo_price_30min <- function(region, start, end, intervention) {
  parts <- list()

  # Pre-5MS portion
  if (start < FIVE_MS_START) {
    pre_end <- min(end, FIVE_MS_START - as.difftime(1, units = "secs"))
    df_pre <- tryCatch(
      aemo_fetch_report_range(
        current_dir = "/Reports/Current/TradingIS_Reports/",
        archive_dir = "/Reports/Archive/TradingIS_Reports/",
        pattern = "TRADINGIS",
        start = start, end = pre_end,
        table = "trading_price"
      ),
      error = function(e) NULL
    )
    if (!is.null(df_pre) && nrow(df_pre) > 0L) {
      df_pre <- aemo_coerce_types(df_pre)
      df_pre <- aemo_apply_filters(df_pre, start = start, end = pre_end,
                                   region = region, intervention = intervention)
      if ("tradingdate" %in% names(df_pre) && !"settlementdate" %in% names(df_pre)) {
        names(df_pre)[names(df_pre) == "tradingdate"] <- "settlementdate"
      }
      parts[["pre"]] <- df_pre
    }
  }

  # Post-5MS portion: aggregate 5-min dispatch to 30-min
  if (end >= FIVE_MS_START) {
    post_start <- max(start, FIVE_MS_START)
    df5 <- tryCatch(
      aemo_fetch_report_range(
        current_dir = "/Reports/Current/DispatchIS_Reports/",
        archive_dir = "/Reports/Archive/DispatchIS_Reports/",
        pattern = "DISPATCHIS",
        start = post_start, end = end,
        table = "dispatch_price"
      ),
      error = function(e) NULL
    )
    if (!is.null(df5) && nrow(df5) > 0L) {
      df5 <- aemo_coerce_types(df5)
      df5 <- aemo_apply_filters(df5, start = post_start, end = end,
                                region = region, intervention = intervention)
      df_post <- aemo_aggregate_to_30min(df5)
      if (nrow(df_post) > 0L) parts[["post"]] <- df_post
    }
  }

  if (length(parts) == 0L) {
    cli::cli_abort("No 30-minute price data found for the requested range.")
  }
  if (length(parts) == 1L) return(parts[[1L]])

  common <- Reduce(intersect, lapply(parts, names))
  stacked <- do.call(rbind,
                     lapply(parts, function(d) d[, common, drop = FALSE]))
  rownames(stacked) <- NULL
  stacked
}

#' Aggregate 5-minute dispatch prices to 30-minute trading intervals.
#'
#' The 30-minute trading period ending at time T contains the six
#' 5-minute dispatch intervals ending at T-25, T-20, T-15, T-10,
#' T-5, and T. The representative price is the arithmetic mean
#' of the six RRPs (matching AEMO's post-5MS TRADINGPRICE
#' derivation in TRADINGIS).
#'
#' Implementation note: O(n) via `split()` on a composite key.
#' v0.1.0 used an O(n^2) nested loop; the new path is ~200x
#' faster on a one-year NSW query.
#' @noRd
aemo_aggregate_to_30min <- function(df) {
  if (!"settlementdate" %in% names(df)) return(df)
  if (!"rrp" %in% names(df)) return(df)

  # Trading interval end: the 30-min period that contains this 5-min interval.
  # NEM convention is period-ENDING timestamps throughout. A dispatch interval
  # stamped 14:05 belongs to the trading interval ending 14:30; one stamped
  # exactly 14:30 also belongs to the interval ending 14:30 (it IS that
  # interval's final 5-min period).
  #
  # Arithmetic: ((epoch - 1) %/% 1800 + 1) * 1800 so that the half-open
  # boundary [14:00:01, 14:30:00] maps to 14:30. Naive ceiling(epoch/1800)
  # misplaces exactly-on-boundary 14:00:00 into 14:00 instead of 14:30.
  epoch <- as.numeric(df$settlementdate)
  ti_end_num <- ((epoch - 1L) %/% 1800L + 1L) * 1800L
  ti_end <- as.POSIXct(ti_end_num, origin = "1970-01-01", tz = AEMO_TIMEZONE)

  region <- if ("regionid" %in% names(df)) df$regionid else rep("", nrow(df))
  key <- paste(region, ti_end_num, sep = "|")

  idx_by_key <- split(seq_len(nrow(df)), key)
  rrp_mean <- vapply(idx_by_key,
                     function(ix) mean(df$rrp[ix], na.rm = TRUE),
                     numeric(1L))

  # One representative row per group: take the first source row, overwrite
  # settlementdate and rrp. Column order matches the input data frame.
  first_idx <- vapply(idx_by_key, `[`, integer(1L), 1L)
  result <- df[first_idx, , drop = FALSE]
  result$settlementdate <- ti_end[first_idx]
  result$rrp <- unname(rrp_mean)
  rownames(result) <- NULL
  result[order(result$settlementdate), , drop = FALSE]
}

#' Frequency control ancillary services (FCAS) prices
#'
#' Returns regional FCAS market prices across the eight
#' contingency services plus the two regulation services. Ten
#' services are live in the NEM since R1/L1 Very Fast commenced
#' on 9 October 2023.
#'
#' Thin wrapper over `aemo_price(..., market = "fcas")`.
#'
#' @param region NEM region code.
#' @param start,end Window (inclusive).
#' @param service Optional character vector of service names
#'   (e.g. `"RAISE6SEC"`, `"LOWER60SEC"`, `"RAISE1SEC"`).
#' @param intervention Logical. See [aemo_price()].
#'
#' @return An `aemo_tbl` with one row per interval and columns
#'   for each requested FCAS RRP (AUD/MW).
#'
#' @family price
#' @export
#' @examples
#' \donttest{
#' op <- options(aemo.cache_dir = tempdir())
#' try({
#'   now <- Sys.time()
#'   f <- aemo_fcas("NSW1", now - 3600, now)
#'   head(f)
#' })
#' options(op)
#' }
aemo_fcas <- function(region, start, end, service = NULL,
                       intervention = FALSE) {
  df <- aemo_price(region = region, start = start, end = end,
                   interval = "5min", market = "fcas",
                   intervention = intervention)
  if (!is.null(service)) {
    service_cols <- paste0(tolower(service), "rrp")
    keep <- union(c("settlementdate", "regionid", "runno", "intervention"),
                  service_cols)
    df <- df[, intersect(keep, names(df)), drop = FALSE]
  }
  df
}

test_that("FIVE_MS_START is 2021-10-01 AEST", {
  expect_equal(
    format(aemo:::FIVE_MS_START, "%Y-%m-%d", tz = "Australia/Brisbane"),
    "2021-10-01"
  )
  expect_equal(attr(aemo:::FIVE_MS_START, "tzone"), "Australia/Brisbane")
})

test_that("aemo_aggregate_to_30min groups 5-min prices to 30-min", {
  tz <- "Australia/Brisbane"
  # Six 5-min intervals ending 14:05 through 14:30 -> one 30-min row at 14:30
  times <- as.POSIXct(
    c("2024-01-15 14:05:00", "2024-01-15 14:10:00",
      "2024-01-15 14:15:00", "2024-01-15 14:20:00",
      "2024-01-15 14:25:00", "2024-01-15 14:30:00"),
    tz = tz
  )
  df <- data.frame(
    settlementdate = times,
    regionid = rep("NSW1", 6L),
    rrp = c(60, 65, 70, 55, 80, 50),
    stringsAsFactors = FALSE
  )
  out <- aemo:::aemo_aggregate_to_30min(df)

  expect_equal(nrow(out), 1L)
  expect_equal(
    format(out$settlementdate, "%H:%M", tz = tz),
    "14:30"
  )
  expect_equal(out$rrp, mean(c(60, 65, 70, 55, 80, 50)))
})

test_that("aemo_aggregate_to_30min handles on-boundary 14:00 correctly", {
  tz <- "Australia/Brisbane"
  # 14:00:00 is the last 5-min interval of the 13:30-14:00 trading period.
  # It must map to 14:00, NOT 14:30.
  times <- as.POSIXct(
    c("2024-01-15 13:35:00", "2024-01-15 13:40:00",
      "2024-01-15 13:45:00", "2024-01-15 13:50:00",
      "2024-01-15 13:55:00", "2024-01-15 14:00:00"),
    tz = tz
  )
  df <- data.frame(
    settlementdate = times,
    regionid = rep("NSW1", 6L),
    rrp = c(50, 55, 60, 65, 70, 75),
    stringsAsFactors = FALSE
  )
  out <- aemo:::aemo_aggregate_to_30min(df)

  expect_equal(nrow(out), 1L)
  expect_equal(
    format(out$settlementdate, "%H:%M", tz = tz),
    "14:00"
  )
  expect_equal(out$rrp, mean(c(50, 55, 60, 65, 70, 75)))
})

test_that("aemo_aggregate_to_30min handles multiple regions", {
  tz <- "Australia/Brisbane"
  times <- rep(
    as.POSIXct(c("2024-01-15 14:05:00", "2024-01-15 14:30:00"), tz = tz),
    each = 2L
  )
  df <- data.frame(
    settlementdate = times,
    regionid = rep(c("NSW1", "VIC1"), 2L),
    rrp = c(60, 55, 70, 65),
    stringsAsFactors = FALSE
  )
  out <- aemo:::aemo_aggregate_to_30min(df)
  # 2 regions × 1 trading interval = 2 rows
  expect_equal(nrow(out), 2L)
})

test_that("aemo_aggregate_to_30min returns df unchanged if no rrp column", {
  df <- data.frame(settlementdate = Sys.time(), other = 1)
  out <- aemo:::aemo_aggregate_to_30min(df)
  expect_equal(nrow(out), 1L)
})

# Pinned-value tests: assert that specific well-documented historical
# intervals return the correct data. These are integration tests that
# require a live network connection; they are skipped on CRAN and CI
# without a network. A failure here means a silent data bug (wrong
# table extracted, wrong column, wrong filter), not just an API error.

skip_if_offline <- function() {
  testthat::skip_if_not(
    isTRUE(tryCatch(
      {
        httr2::request("http://nemweb.com.au") |>
          httr2::req_timeout(5) |>
          httr2::req_perform()
        TRUE
      },
      error = function(e) FALSE
    )),
    message = "NEMweb not reachable"
  )
}

test_that("aemo_price returns correct RRP for a known historical interval", {
  # NSW1 RRP on 14 April 2021 at 12:30 AEST was approximately AUD 50-70/MWh
  # (off-peak, pre-5MS era). We pin the sign and rough order-of-magnitude
  # rather than an exact value to guard against AEMO restating.
  testthat::skip_on_cran()
  skip_if_offline()

  withr::local_options(aemo.cache_dir = tempdir())
  p <- aemo_price(
    region = "NSW1",
    start  = "2021-04-14 12:00:00",
    end    = "2021-04-14 13:00:00",
    interval = "5min"
  )
  expect_s3_class(p, "aemo_tbl")
  expect_true(nrow(p) >= 6L,
              info = "Expected at least 6 five-minute intervals")
  expect_true("rrp" %in% names(p))
  expect_true(all(is.numeric(p$rrp)),
              info = "RRP must be numeric, not character")
  # AEST timezone preserved end-to-end
  expect_equal(attr(p$settlementdate, "tzone"), "Australia/Brisbane")
  # Off-peak NSW prices in this window: positive and < 300 AUD/MWh
  expect_true(all(p$rrp > -1001 & p$rrp < 17600),
              info = "All RRPs must be within market floor/cap bounds")
})

test_that("aemo_price intervention filter works: INTERVENTION=0 rows only by default", {
  # During the June 2022 market suspension AEMO issued intervention
  # directions on 13 June 2022. DISPATCHPRICE contains both intervention=0
  # and intervention=1 rows. The default filter should return only
  # intervention=0 rows.
  testthat::skip_on_cran()
  skip_if_offline()

  withr::local_options(aemo.cache_dir = tempdir())
  p_default <- aemo_price(
    region = "NSW1",
    start  = "2022-06-13 16:00:00",
    end    = "2022-06-13 17:00:00"
  )
  if ("intervention" %in% names(p_default)) {
    iv <- suppressWarnings(as.integer(p_default$intervention))
    expect_true(all(is.na(iv) | iv == 0L),
                info = "Default should return only intervention=0 rows")
  }

  p_both <- aemo_price(
    region = "NSW1",
    start  = "2022-06-13 16:00:00",
    end    = "2022-06-13 17:00:00",
    intervention = TRUE
  )
  # With intervention=TRUE we expect >= as many rows as the default
  expect_true(nrow(p_both) >= nrow(p_default))
})

test_that("aemo_price timestamps are AEST not AEDT (no DST shift)", {
  # On 6 October 2024 Sydney clocks sprang forward at 02:00 (AEST -> AEDT).
  # With Australia/Sydney the 02:30 interval would be ambiguous/NA.
  # With Australia/Brisbane it must parse cleanly.
  testthat::skip_on_cran()
  skip_if_offline()

  withr::local_options(aemo.cache_dir = tempdir())
  p <- aemo_price(
    region   = "NSW1",
    start    = "2024-10-06 02:00:00",
    end      = "2024-10-06 03:00:00"
  )
  expect_true(nrow(p) > 0L)
  expect_false(any(is.na(p$settlementdate)),
               info = "No NA timestamps at DST-transition boundary in AEST")
  expect_equal(attr(p$settlementdate, "tzone"), "Australia/Brisbane")
})

test_that("aemo_price validates region", {
  expect_error(aemo_price("ACT1", "2024-06-01", "2024-06-02"))
})

test_that("aemo_price validates interval", {
  expect_error(aemo_price("NSW1", "2024-06-01", "2024-06-02", interval = "1min"))
})

test_that("aemo_price requires end >= start", {
  expect_error(aemo_price("NSW1", "2024-06-02", "2024-06-01"))
})

test_that("aemo_price live fetch", {
  skip_on_cran()
  skip_if_offline()
  skip_if(Sys.getenv("AEMO_LIVE_TESTS") != "true",
          "Live network tests disabled by default.")

  p <- aemo_price("NSW1", "2024-06-01", "2024-06-01 01:00:00")
  expect_s3_class(p, "aemo_tbl")
})

#' Per-DUID dispatch output
#'
#' Returns 5-minute generator output for one or more DUIDs.
#' Three measures are available:
#'
#' - `"scada_mw"` (default): actual metered output from
#'   `DISPATCH_UNIT_SCADA` (`SCADAVALUE`).
#' - `"target_mw"`: dispatch target from `DISPATCHLOAD`
#'   (`TOTALCLEARED`). This is the MW AEMO *asked* the unit to
#'   produce at the end of the interval.
#' - `"both"`: returns `SCADAVALUE`, `INITIALMW` (SCADA at the
#'   start of the interval) and `TOTALCLEARED` (target at the
#'   end) in a single row per DUID per interval. Use this for
#'   ramp-trajectory research: the ramp applied during the
#'   interval is the straight line from `INITIALMW` to
#'   `TOTALCLEARED`.
#'
#' Timestamps are AEST (UTC+10, no DST).
#'
#' @param duid Optional character vector of DUIDs. `NULL` returns
#'   all generators.
#' @param start,end Window (inclusive).
#' @param measure One of `"scada_mw"` (default), `"target_mw"`,
#'   or `"both"`.
#'
#' @return An `aemo_tbl` with columns `settlementdate`, `duid`,
#'   and the requested measure(s).
#'
#' @family dispatch
#' @export
#' @examples
#' \donttest{
#' op <- options(aemo.cache_dir = tempdir())
#' try({
#'   now <- Sys.time()
#'   # SCADA actual
#'   d <- aemo_dispatch_units(duid = "BW01", start = now - 3600,
#'                             end = now)
#'
#'   # Paired: INITIALMW, TOTALCLEARED, SCADAVALUE (ramp research)
#'   d_both <- aemo_dispatch_units(duid = "BW01",
#'                                  start = now - 3600, end = now,
#'                                  measure = "both")
#' })
#' options(op)
#' }
aemo_dispatch_units <- function(duid = NULL, start, end,
                                 measure = c("scada_mw", "target_mw",
                                             "both")) {
  measure <- match.arg(measure)
  start <- aemo_parse_time(start)
  end <- aemo_parse_time(end)
  stopifnot(end >= start)

  if (measure == "both") {
    return(aemo_dispatch_units_both(duid = duid, start = start, end = end))
  }

  if (measure == "scada_mw") {
    df <- aemo_fetch_report_range(
      current_dir = "/Reports/Current/Dispatch_SCADA/",
      archive_dir = "/Reports/Archive/Dispatch_SCADA/",
      pattern = "DISPATCHSCADA",
      start = start, end = end
    )
  } else {
    df <- aemo_fetch_report_range(
      current_dir = "/Reports/Current/DispatchIS_Reports/",
      archive_dir = "/Reports/Archive/DispatchIS_Reports/",
      pattern = "DISPATCHIS",
      start = start, end = end,
      table = "dispatch_unit_solution"
    )
  }
  df <- aemo_coerce_types(df)
  df <- aemo_apply_filters(df, start = start, end = end)

  if (!is.null(duid) && "duid" %in% names(df)) {
    df <- df[toupper(df$duid) %in% toupper(duid), , drop = FALSE]
    rownames(df) <- NULL
  }

  new_aemo_tbl(df,
               source = "http://nemweb.com.au",
               title = paste0("AEMO dispatch units (", measure, ")"))
}

#' Fetch SCADA actual + DISPATCHLOAD initial/target and pair by
#' (duid, settlementdate).
#' @noRd
aemo_dispatch_units_both <- function(duid, start, end) {
  scada <- aemo_fetch_report_range(
    current_dir = "/Reports/Current/Dispatch_SCADA/",
    archive_dir = "/Reports/Archive/Dispatch_SCADA/",
    pattern = "DISPATCHSCADA",
    start = start, end = end
  )
  scada <- aemo_coerce_types(scada)
  scada <- aemo_apply_filters(scada, start = start, end = end)

  load <- aemo_fetch_report_range(
    current_dir = "/Reports/Current/DispatchIS_Reports/",
    archive_dir = "/Reports/Archive/DispatchIS_Reports/",
    pattern = "DISPATCHIS",
    start = start, end = end,
    table = "dispatch_unit_solution"
  )
  load <- aemo_coerce_types(load)
  load <- aemo_apply_filters(load, start = start, end = end)

  if (!is.null(duid)) {
    scada <- scada[toupper(scada$duid) %in% toupper(duid), , drop = FALSE]
    load  <- load[toupper(load$duid)   %in% toupper(duid), , drop = FALSE]
  }

  load_keep <- intersect(c("settlementdate", "duid", "intervention",
                            "initialmw", "totalcleared", "availability",
                            "rampuprate", "rampdownrate",
                            "semidispatchcap"),
                         names(load))
  scada_keep <- intersect(c("settlementdate", "duid", "scadavalue"),
                          names(scada))
  load <- load[, load_keep, drop = FALSE]
  scada <- scada[, scada_keep, drop = FALSE]

  joined <- merge(load, scada,
                   by = intersect(c("settlementdate", "duid"),
                                  intersect(names(load), names(scada))),
                   all.x = TRUE, sort = FALSE)
  rownames(joined) <- NULL
  new_aemo_tbl(joined,
               source = "http://nemweb.com.au",
               title = "AEMO dispatch units (INITIALMW + TOTALCLEARED + SCADAVALUE)")
}

# FCAS service columns in DISPATCHLOAD (10 services as of Oct 2023)
FCAS_ENABLEMENT_COLS <- c(
  "raise1secmw", "lower1secmw",
  "raise6secmw", "lower6secmw",
  "raise60secmw", "lower60secmw",
  "raise5minmw", "lower5minmw",
  "raiseregmw", "lowerregmw"
)

#' FCAS enablement volumes by generator unit
#'
#' Returns the MW enabled for each of the ten Frequency Control
#' Ancillary Services (FCAS) per DUID per 5-minute dispatch
#' interval from `DISPATCHLOAD`. This is the quantity side of
#' the FCAS market: how much each unit was enabled (dispatched)
#' to provide each service, as distinct from the price returned
#' by [aemo_fcas()].
#'
#' Ten services are active since 9 October 2023 when Very Fast
#' (`RAISE1SEC` / `LOWER1SEC`) commenced. Before that date only
#' eight services are present; the `raise1secmw` / `lower1secmw`
#' columns will be `NA` for intervals before that date.
#'
#' @param duid Optional character vector of DUIDs. `NULL` returns
#'   all units. See [aemo_units()] for the full DUID registry.
#' @param start,end Window (inclusive), character or POSIXct.
#' @param service Optional character vector of service names,
#'   e.g. `c("raise6sec", "lowerreg")`. Case-insensitive.
#'   `NULL` returns all ten services.
#' @param intervention Logical. Default `FALSE`.
#'
#' @return An `aemo_tbl` with columns `settlementdate`, `duid`,
#'   and one column per FCAS service (`raise1secmw`,
#'   `lower1secmw`, `raise6secmw`, `lower6secmw`,
#'   `raise60secmw`, `lower60secmw`, `raise5minmw`,
#'   `lower5minmw`, `raiseregmw`, `lowerregmw`). Values are
#'   MW; zero means the unit was not enabled for that service.
#'
#' @source AEMO NEMweb DISPATCHIS_Reports, DISPATCHLOAD table.
#'
#' @family dispatch
#' @export
#' @examples
#' \donttest{
#' op <- options(aemo.cache_dir = tempdir())
#' try({
#'   now <- Sys.time()
#'   e <- aemo_fcas_enablement(start = now - 3600, end = now)
#'   head(e)
#' })
#' options(op)
#' }
aemo_fcas_enablement <- function(duid = NULL, start, end,
                                  service = NULL,
                                  intervention = FALSE) {
  start <- aemo_parse_time(start)
  end <- aemo_parse_time(end)
  stopifnot(end >= start)

  df <- aemo_fetch_report_range(
    current_dir = "/Reports/Current/DispatchIS_Reports/",
    archive_dir = "/Reports/Archive/DispatchIS_Reports/",
    pattern = "DISPATCHIS",
    start = start, end = end,
    table = "dispatch_unit_solution"
  )
  df <- aemo_coerce_types(df)
  df <- aemo_apply_filters(df, start = start, end = end,
                           intervention = intervention)

  if (!is.null(duid) && "duid" %in% names(df)) {
    df <- df[toupper(df$duid) %in% toupper(duid), , drop = FALSE]
  }

  keep_cols <- c("settlementdate", "duid", "intervention")
  if (!is.null(service)) {
    svc_cols <- paste0(tolower(service), "mw")
    svc_cols <- intersect(svc_cols, FCAS_ENABLEMENT_COLS)
    if (length(svc_cols) == 0L) {
      cli::cli_abort(c(
        "No matching FCAS service columns found.",
        "i" = "Valid services: {.val {sub('mw$', '', FCAS_ENABLEMENT_COLS)}}."
      ))
    }
    keep_cols <- c(keep_cols, svc_cols)
  } else {
    keep_cols <- c(keep_cols, FCAS_ENABLEMENT_COLS)
  }
  if (!intervention) keep_cols <- setdiff(keep_cols, "intervention")
  present <- intersect(keep_cols, names(df))
  df <- df[, present, drop = FALSE]
  rownames(df) <- NULL

  new_aemo_tbl(df,
               source = "http://nemweb.com.au",
               title = "AEMO FCAS enablement volumes (DISPATCHLOAD)")
}

test_that("FCAS_ENABLEMENT_COLS has 10 services", {
  expect_equal(length(aemo:::FCAS_ENABLEMENT_COLS), 10L)
  expect_true("raise1secmw" %in% aemo:::FCAS_ENABLEMENT_COLS)
  expect_true("lowerregmw"  %in% aemo:::FCAS_ENABLEMENT_COLS)
})

test_that("aemo_fcas_enablement errors on bad service name", {
  withr::local_options(aemo.cache_dir = tempdir())
  now <- Sys.time()
  expect_error(
    aemo_fcas_enablement(start = now - 60, end = now,
                         service = "notaservice"),
    "No matching FCAS service columns"
  )
})

test_that("aemo_fcas_enablement accepts NULL service (all 10)", {
  withr::local_options(aemo.cache_dir = tempdir())
  # Verify the call doesn't error at argument-checking stage (network
  # errors are wrapped in try() in the live test).
  now <- Sys.time()
  result <- tryCatch(
    aemo_fcas_enablement(start = now - 300, end = now),
    error = function(e) e
  )
  # Either an aemo_tbl or an error from the network; both are acceptable here.
  expect_true(inherits(result, "aemo_tbl") || inherits(result, "error"))
})

#' Rooftop PV actuals and forecasts
#'
#' Returns AEMO's region-level estimate of rooftop PV generation,
#' either actuals or forecasts. Published at 30-minute resolution.
#'
#' The "actual" figure is an AEMO estimate derived from the APVI
#' sampling model and weather data, not metered SCADA output. It
#' is the best available public measure of aggregate rooftop PV
#' generation but is subject to revision.
#'
#' @param region NEM region code.
#' @param start,end Window.
#' @param type One of `"actual"` (default) or `"forecast"`.
#'
#' @return An `aemo_tbl`.
#'
#' @family dispatch
#' @export
#' @examples
#' \donttest{
#' op <- options(aemo.cache_dir = tempdir())
#' try({
#'   now <- Sys.time()
#'   r <- aemo_rooftop_pv("NSW1",
#'                         start = now - 3600, end = now)
#'   head(r)
#' })
#' options(op)
#' }
aemo_rooftop_pv <- function(region, start, end,
                             type = c("actual", "forecast")) {
  region <- aemo_validate_region(region)
  type <- match.arg(type)
  start <- aemo_parse_time(start)
  end <- aemo_parse_time(end)
  stopifnot(end >= start)

  sub <- if (type == "actual") "ACTUAL" else "FORECAST"
  df <- aemo_fetch_report_range(
    current_dir = paste0("/Reports/Current/ROOFTOP_PV/", sub, "/"),
    archive_dir = paste0("/Reports/Archive/ROOFTOP_PV/", sub, "/"),
    pattern = "ROOFTOP_PV",
    start = start, end = end
  )
  df <- aemo_coerce_types(df)
  # Rooftop PV uses INTERVAL_DATETIME rather than SETTLEMENTDATE.
  if (!"settlementdate" %in% names(df) &&
      "interval_datetime" %in% names(df)) {
    df$settlementdate <- df$interval_datetime
  }
  df <- aemo_apply_filters(df, start = start, end = end, region = region)

  new_aemo_tbl(df,
               source = "http://nemweb.com.au",
               title = paste0("AEMO rooftop PV ", type, " ",
                              paste(region, collapse = "+")))
}

#' Regional electricity demand
#'
#' Returns 5-minute regional demand from `DISPATCHREGIONSUM`.
#' Three demand measures are supported, aligned with AEMO's
#' Demand Terms taxonomy:
#'
#' - `"operational"` (default): `TOTALDEMAND`, the grid-measured
#'   demand met by scheduled and semi-scheduled generation plus
#'   net interchange. This is the quantity AEMO dispatches.
#' - `"operational_less_snsg"`: `TOTALDEMAND` minus small
#'   non-scheduled generation (SS_SOLAR_UIGF + SS_WIND_UIGF
#'   where present).
#' - `"native"`: `TOTALDEMAND` plus estimated rooftop PV
#'   generation. Closest to end-use consumption. If the rooftop
#'   PV component is not available in DISPATCHREGIONSUM the
#'   function warns and returns `TOTALDEMAND`; users should join
#'   with `aemo_rooftop_pv()` for a full native-demand estimate.
#'
#' Timestamps are AEST (UTC+10, no DST).
#'
#' @param region NEM region code. Vector accepted.
#' @param start,end Window (inclusive), character or POSIXct.
#' @param measure One of `"operational"` (default),
#'   `"operational_less_snsg"`, or `"native"`.
#' @param intervention Logical. Default `FALSE` filters to market
#'   pricing runs.
#'
#' @return An `aemo_tbl` with columns `settlementdate`,
#'   `regionid`, `demand_mw` (the requested measure), plus the
#'   underlying DISPATCHREGIONSUM columns used in the derivation.
#'
#' @source AEMO NEMweb, AEMO Copyright Permissions Notice.
#'
#' @family demand
#' @export
#' @examples
#' \donttest{
#' op <- options(aemo.cache_dir = tempdir())
#' try({
#'   now <- Sys.time()
#'   d <- aemo_demand("VIC1", now - 3600, now)
#'   head(d)
#' })
#' options(op)
#' }
aemo_demand <- function(region, start, end,
                         measure = c("operational",
                                     "operational_less_snsg",
                                     "native"),
                         intervention = FALSE) {
  region <- aemo_validate_region(region)
  measure <- match.arg(measure)
  start <- aemo_parse_time(start)
  end <- aemo_parse_time(end)
  stopifnot(end >= start)

  df <- aemo_fetch_report_range(
    current_dir = "/Reports/Current/DispatchIS_Reports/",
    archive_dir = "/Reports/Archive/DispatchIS_Reports/",
    pattern = "DISPATCHIS",
    start = start, end = end,
    table = "dispatch_regionsum"
  )
  df <- aemo_coerce_types(df)
  df <- aemo_apply_filters(df, start = start, end = end,
                           region = region, intervention = intervention)

  if ("totaldemand" %in% names(df)) {
    td <- df$totaldemand
    snsg <- rep(0, nrow(df))
    for (col in c("ss_solar_uigf", "ss_wind_uigf")) {
      if (col %in% names(df)) {
        v <- suppressWarnings(as.numeric(df[[col]]))
        v[is.na(v)] <- 0
        snsg <- snsg + v
      }
    }
    roof <- rep(0, nrow(df))
    for (col in c("rooftop_pv", "rooftop_solar")) {
      if (col %in% names(df)) {
        v <- suppressWarnings(as.numeric(df[[col]]))
        v[is.na(v)] <- 0
        roof <- roof + v
      }
    }
    df$demand_mw <- switch(measure,
      operational            = td,
      operational_less_snsg  = td - snsg,
      native                 = td + roof
    )
    if (measure == "native" && all(roof == 0)) {
      cli::cli_warn(c(
        "No rooftop-PV column found in DISPATCHREGIONSUM.",
        "i" = "Native demand returned as TOTALDEMAND only; join {.fn aemo_rooftop_pv} for a full estimate."
      ))
    }
  }

  new_aemo_tbl(df,
               source = "http://nemweb.com.au",
               title = paste0("AEMO demand ", paste(region, collapse = "+"),
                              " (", measure, ")"))
}

test_that("aemo_demand validates region", {
  expect_error(aemo_demand("ACT1", "2024-06-01", "2024-06-02"))
})

test_that("aemo_demand validates measure", {
  expect_error(aemo_demand("NSW1", "2024-06-01", "2024-06-02",
                           measure = "instantaneous"))
})

test_that("aemo_demand live fetch", {
  skip_on_cran()
  skip_if_offline()
  skip_if(Sys.getenv("AEMO_LIVE_TESTS") != "true",
          "Live network tests disabled by default.")
  d <- aemo_demand("VIC1", "2024-06-01", "2024-06-01 01:00:00")
  expect_s3_class(d, "aemo_tbl")
})

# aemo_tbl S3 class

#' @noRd
new_aemo_tbl <- function(df, source = NULL,
                          licence = "AEMO Copyright Permissions Notice",
                          retrieved = Sys.time(), title = NULL) {
  stopifnot(is.data.frame(df))
  attr(df, "aemo_source") <- source
  attr(df, "aemo_licence") <- licence
  attr(df, "aemo_retrieved") <- retrieved
  attr(df, "aemo_title") <- title
  class(df) <- c("aemo_tbl", class(df))
  df
}

#' Print an aemo_tbl
#'
#' @param x An `aemo_tbl`.
#' @param ... Passed to the next print method.
#' @return Invisibly returns `x`.
#' @export
#' @examples
#' x <- data.frame(settlementdate = Sys.time(), region = "NSW1", rrp = 80)
#' x <- structure(x, aemo_title = "Demo", aemo_source = "http://nemweb.com.au",
#'                aemo_licence = "AEMO Copyright Permissions Notice",
#'                aemo_retrieved = Sys.time(),
#'                class = c("aemo_tbl", "data.frame"))
#' print(x)
print.aemo_tbl <- function(x, ...) {
  title <- attr(x, "aemo_title") %||% "AEMO data"
  source <- attr(x, "aemo_source") %||% "http://nemweb.com.au"
  licence <- attr(x, "aemo_licence") %||% "AEMO Copyright Permissions Notice"
  retrieved <- attr(x, "aemo_retrieved")
  retrieved_str <- if (!is.null(retrieved)) {
    format(retrieved, "%Y-%m-%d %H:%M %Z")
  } else {
    "-"
  }

  cat("# aemo_tbl: ", title, "\n", sep = "")
  cat("# Source:   ", source, "\n", sep = "")
  cat("# Licence:  ", licence, "\n", sep = "")
  cat("# Retrieved:", retrieved_str, "\n")
  cat("# Rows: ", formatC(nrow(x), big.mark = ",", format = "d"),
      "  Cols: ", ncol(x), "\n", sep = "")
  cat("\n")

  y <- x
  class(y) <- setdiff(class(y), "aemo_tbl")
  print(y, ...)
  invisible(x)
}

test_that("new_aemo_tbl attaches provenance", {
  df <- data.frame(a = 1:2)
  x <- aemo:::new_aemo_tbl(df, title = "T", source = "http://nemweb.com.au")
  expect_s3_class(x, "aemo_tbl")
  expect_equal(attr(x, "aemo_title"), "T")
  expect_match(attr(x, "aemo_licence"), "AEMO")
})

test_that("new_aemo_tbl requires a data frame", {
  expect_error(aemo:::new_aemo_tbl(list()))
})

test_that("print.aemo_tbl emits header", {
  df <- data.frame(a = 1)
  x <- aemo:::new_aemo_tbl(df, title = "Demo",
                            retrieved = as.POSIXct("2026-01-01", tz = "UTC"))
  out <- capture.output(print(x))
  expect_true(any(grepl("aemo_tbl: Demo", out, fixed = TRUE)))
  expect_true(any(grepl("AEMO", out)))
})

test_that("print.aemo_tbl returns x invisibly", {
  df <- data.frame(a = 1)
  x <- aemo:::new_aemo_tbl(df)
  tf <- tempfile()
  sink(tf)
  wv <- withVisible(print(x))
  sink()
  unlink(tf)
  expect_false(wv$visible)
  expect_identical(wv$value, x)
})

# Cache management

#' @noRd
aemo_cache_dir <- function() {
  d <- getOption("aemo.cache_dir",
                 default = tools::R_user_dir("aemo", "cache"))
  if (!dir.exists(d)) dir.create(d, recursive = TRUE)
  d
}

#' Clear the aemo cache
#'
#' @return Invisibly returns `NULL`.
#' @family configuration
#' @export
#' @examples
#' \donttest{
#' op <- options(aemo.cache_dir = tempdir())
#' aemo_clear_cache()
#' options(op)
#' }
aemo_clear_cache <- function() {
  d <- aemo_cache_dir()
  files <- list.files(d, full.names = TRUE)
  n <- length(files)
  if (n > 0L) unlink(files, recursive = TRUE)
  cli::cli_inform("Removed {n} cached file{?s} from {.path {d}}.")
  invisible(NULL)
}

#' Inspect the local aemo cache
#'
#' @return A list with `dir`, `n_files`, `size_bytes`, `size_human`, `files`.
#' @family configuration
#' @export
#' @examples
#' \donttest{
#' op <- options(aemo.cache_dir = tempdir())
#' aemo_cache_info()
#' options(op)
#' }
aemo_cache_info <- function() {
  d <- aemo_cache_dir()
  empty <- data.frame(name = character(0L), size_bytes = numeric(0L),
                      modified = as.POSIXct(character(0L)),
                      stringsAsFactors = FALSE)
  paths <- list.files(d, full.names = TRUE)
  if (length(paths) == 0L) {
    return(list(dir = d, n_files = 0L, size_bytes = 0,
                size_human = "0 B", files = empty))
  }
  info <- file.info(paths)
  files <- data.frame(name = basename(paths),
                      size_bytes = info$size,
                      modified = info$mtime,
                      stringsAsFactors = FALSE)
  files <- files[order(-files$size_bytes), , drop = FALSE]
  rownames(files) <- NULL
  total <- sum(files$size_bytes)
  list(dir = d, n_files = nrow(files), size_bytes = total,
       size_human = aemo_format_bytes(total), files = files)
}

#' Configure request throttling
#'
#' Controls the delay between successive NEMweb requests.
#' Defaults to 1 second (half a request per second).
#'
#' @param delay Numeric. Minimum delay between requests in
#'   seconds. Internally this becomes `httr2::req_throttle(rate = 1/delay)`.
#' @return Invisibly returns the previous value.
#' @family configuration
#' @export
#' @examples
#' old <- aemo_throttle(0.5)  # 2 requests per second
#' aemo_throttle(old)         # restore
aemo_throttle <- function(delay = 1) {
  stopifnot(is.numeric(delay), length(delay) == 1L, delay > 0)
  old <- getOption("aemo.throttle_delay", 1)
  options(aemo.throttle_delay = delay)
  invisible(old)
}

test_that("aemo_cache_dir honours the aemo.cache_dir option", {
  tmp <- tempfile("aemo_cache_")
  op <- options(aemo.cache_dir = tmp)
  on.exit(options(op), add = TRUE)
  d <- aemo:::aemo_cache_dir()
  expect_equal(normalizePath(d), normalizePath(tmp))
})

test_that("aemo_cache_info returns expected structure", {
  tmp <- tempfile("aemo_cache_")
  op <- options(aemo.cache_dir = tmp)
  on.exit(options(op), add = TRUE)
  info <- aemo_cache_info()
  expect_setequal(names(info), c("dir", "n_files", "size_bytes", "size_human", "files"))
  expect_equal(info$n_files, 0L)
})

test_that("aemo_cache_info counts files", {
  tmp <- tempfile("aemo_cache_")
  op <- options(aemo.cache_dir = tmp)
  on.exit(options(op), add = TRUE)
  dir.create(tmp, recursive = TRUE)
  writeLines("a", file.path(tmp, "a.zip"))
  info <- aemo_cache_info()
  expect_equal(info$n_files, 1L)
})

test_that("aemo_clear_cache removes files", {
  tmp <- tempfile("aemo_cache_")
  op <- options(aemo.cache_dir = tmp)
  on.exit(options(op), add = TRUE)
  dir.create(tmp, recursive = TRUE)
  writeLines("x", file.path(tmp, "x.zip"))
  expect_invisible(aemo_clear_cache())
  expect_false(file.exists(file.path(tmp, "x.zip")))
})

test_that("aemo_throttle validates and round-trips", {
  old <- aemo_throttle(0.5)
  expect_equal(getOption("aemo.throttle_delay"), 0.5)
  aemo_throttle(old)
  expect_equal(getOption("aemo.throttle_delay"), old)
  expect_error(aemo_throttle(-1))
  expect_error(aemo_throttle("fast"))
})

test_that("aemo_format_bytes formats thresholds", {
  expect_equal(aemo:::aemo_format_bytes(0), "0 B")
  expect_match(aemo:::aemo_format_bytes(1500), "KB$")
  expect_match(aemo:::aemo_format_bytes(1500000), "MB$")
})

#' Generator bid stack
#'
#' Returns `BIDDAYOFFER_D` (daily bid summary: 10 price bands,
#' MaxAvail, fixed load, locked at 12:30 D-1), `BIDPEROFFER_D`
#' (per-interval availability and rebids), or the two joined on
#' `(duid, settlementdate, bidtype)`.
#'
#' **Parent / child structure.** `BIDDAYOFFER_D` carries the
#' **price bands** (`priceband1..priceband10`), which are locked
#' at 12:30 on the day ahead and cannot be rebid.
#' `BIDPEROFFER_D` carries the **per-interval availability
#' bands** (`bandavail1..bandavail10`), which can be rebid
#' intraday. Serious bidding analysis (Goncalves & Menezes 2022
#' Energy Economics 113 106398; Nelson et al. 2024 AJARE 68(4))
#' needs both joined.
#'
#' **Size warning.** `BIDPEROFFER_D` monthly archives are
#' multi-gigabyte. By default `aemo_bids()` refuses spans longer
#' than 30 days; pass `allow_large = TRUE` to override.
#'
#' **Upstream gap.** AEMO has a documented gap in
#' `BIDPEROFFER_D` between March 2021 and July 2024. Rows in
#' that range may be missing.
#'
#' @param duid Character vector of DUIDs. Required.
#' @param start,end Window.
#' @param resolution One of `"day"` (default, BIDDAYOFFER_D
#'   only), `"period"` (BIDPEROFFER_D only), or `"joined"`
#'   (BIDDAYOFFER_D inner-joined to BIDPEROFFER_D on
#'   `(duid, settlementdate, bidtype)`).
#' @param allow_large Logical. Default `FALSE`.
#'
#' @return An `aemo_tbl`.
#'
#' @family dispatch
#' @export
#' @examples
#' \donttest{
#' op <- options(aemo.cache_dir = tempdir())
#' try({
#'   now <- Sys.time()
#'   # Daily bid summary (price bands)
#'   b <- aemo_bids(duid = "BW01",
#'                   start = now - 86400, end = now)
#'
#'   # Joined: price bands + per-interval volumes
#'   bj <- aemo_bids(duid = "BW01", start = now - 86400, end = now,
#'                    resolution = "joined")
#' })
#' options(op)
#' }
aemo_bids <- function(duid, start, end,
                      resolution = c("day", "period", "joined"),
                      allow_large = FALSE) {
  if (missing(duid) || is.null(duid) || length(duid) == 0L) {
    cli::cli_abort("{.arg duid} is required. Pass at least one DUID.")
  }
  resolution <- match.arg(resolution)
  start <- aemo_parse_time(start)
  end <- aemo_parse_time(end)
  stopifnot(end >= start)

  span_days <- as.numeric(difftime(end, start, units = "days"))
  # The size guard applies whenever BIDPEROFFER_D is touched.
  if (resolution %in% c("period", "joined") && span_days > 30 &&
      !allow_large) {
    cli::cli_abort(c(
      "Requested span is {round(span_days)} days.",
      "i" = "Bids data is large (BIDPEROFFER_D monthly archives are multi-GB zipped).",
      "i" = "Pass {.code allow_large = TRUE} to proceed."
    ))
  }

  if (resolution == "joined") {
    return(aemo_bids_joined(duid = duid, start = start, end = end))
  }

  if (resolution == "day") {
    current <- "/Reports/Current/Next_Day_Offer_Energy/"
    archive <- "/Reports/Archive/Next_Day_Offer_Energy/"
    pattern <- "NEXT_DAY_OFFER_ENERGY"
  } else {
    current <- "/Reports/Current/Bidmove_Complete/"
    archive <- "/Reports/Archive/Bidmove_Complete/"
    pattern <- "BIDMOVE_COMPLETE"
  }

  df <- aemo_fetch_report_range(
    current_dir = current, archive_dir = archive,
    pattern = pattern, start = start, end = end
  )
  df <- aemo_coerce_types(df)
  df <- aemo_apply_filters(df, start = start, end = end)

  if ("duid" %in% names(df)) {
    df <- df[toupper(df$duid) %in% toupper(duid), , drop = FALSE]
    rownames(df) <- NULL
  }
  new_aemo_tbl(df,
               source = "http://nemweb.com.au",
               title = paste0("AEMO bids (", resolution, ")"))
}

#' Inner-join BIDDAYOFFER_D (price bands) to BIDPEROFFER_D (volumes).
#'
#' Joins on `(duid, settlementdate, bidtype)`. Returns one row per
#' (duid, 5-min interval, bidtype) with both the locked price
#' bands and the rebiddable volume bands.
#' @noRd
aemo_bids_joined <- function(duid, start, end) {
  # BIDDAYOFFER is pulled for the day range spanning the query.
  day_start <- as.POSIXct(format(start - as.difftime(1, units = "days"),
                                  "%Y-%m-%d 00:00:00"),
                          tz = AEMO_TIMEZONE)
  day_end <- as.POSIXct(format(end + as.difftime(1, units = "days"),
                                "%Y-%m-%d 23:55:00"),
                        tz = AEMO_TIMEZONE)

  day <- aemo_fetch_report_range(
    current_dir = "/Reports/Current/Next_Day_Offer_Energy/",
    archive_dir = "/Reports/Archive/Next_Day_Offer_Energy/",
    pattern = "NEXT_DAY_OFFER_ENERGY",
    start = day_start, end = day_end
  )
  day <- aemo_coerce_types(day)

  per <- aemo_fetch_report_range(
    current_dir = "/Reports/Current/Bidmove_Complete/",
    archive_dir = "/Reports/Archive/Bidmove_Complete/",
    pattern = "BIDMOVE_COMPLETE",
    start = start, end = end
  )
  per <- aemo_coerce_types(per)
  per <- aemo_apply_filters(per, start = start, end = end)

  for (d in list(day, per)) {
    if ("duid" %in% names(d)) {
      # Filter each side to the requested DUIDs before joining.
      assign(if (identical(d, day)) "day" else "per",
             d[toupper(d$duid) %in% toupper(duid), , drop = FALSE])
    }
  }
  day <- day[toupper(day$duid) %in% toupper(duid), , drop = FALSE]
  per <- per[toupper(per$duid) %in% toupper(duid), , drop = FALSE]

  # BIDDAYOFFER carries price bands 1..10; BIDPEROFFER carries bandavail 1..10.
  # Join key: (duid, bidtype) + the trading-day of the per-interval row.
  price_cols <- grep("^priceband\\d+$", names(day), value = TRUE)
  vol_cols   <- grep("^bandavail\\d+$", names(per), value = TRUE)
  day_keep <- intersect(c("duid", "bidtype", "settlementdate",
                           "maxavail", "fixedload", "daaavailability",
                           price_cols),
                         names(day))
  per_keep <- intersect(c("duid", "bidtype", "settlementdate",
                           "offerdate", "maxavail", "rampuprate",
                           "rampdownrate", vol_cols),
                         names(per))

  day <- day[, day_keep, drop = FALSE]
  per <- per[, per_keep, drop = FALSE]
  # Rename BIDDAYOFFER's settlementdate to trading_day for the join;
  # BIDPEROFFER's settlementdate is the 5-min period ending.
  if ("settlementdate" %in% names(day)) {
    day$trading_day <- as.Date(day$settlementdate, tz = AEMO_TIMEZONE)
    day$settlementdate <- NULL
  }
  if ("settlementdate" %in% names(per)) {
    per$trading_day <- as.Date(per$settlementdate, tz = AEMO_TIMEZONE)
  }
  # Suffix collisions handled explicitly: keep BIDPEROFFER's maxavail.
  if ("maxavail" %in% names(day) && "maxavail" %in% names(per)) {
    names(day)[names(day) == "maxavail"] <- "maxavail_day"
  }

  joined <- merge(per, day,
                   by = intersect(c("duid", "bidtype", "trading_day"),
                                  intersect(names(per), names(day))),
                   all.x = TRUE, sort = FALSE)
  rownames(joined) <- NULL
  new_aemo_tbl(joined,
               source = "http://nemweb.com.au",
               title = "AEMO bids (joined: BIDDAYOFFER + BIDPEROFFER)")
}

test_that("aemo_bids requires duid", {
  expect_error(aemo_bids(start = "2024-06-01", end = "2024-06-02"),
               "duid")
})

test_that("aemo_bids enforces span guard on BIDPEROFFER (period / joined)", {
  # v0.2.0: guard applies only when BIDPEROFFER_D is touched, not for the
  # small BIDDAYOFFER_D daily resolution. BIDPEROFFER monthly archives
  # are multi-GB; BIDDAYOFFER files are small.
  expect_error(
    aemo_bids(duid = "BW01", start = "2024-01-01", end = "2024-06-01",
              resolution = "period"),
    "allow_large"
  )
  expect_error(
    aemo_bids(duid = "BW01", start = "2024-01-01", end = "2024-06-01",
              resolution = "joined"),
    "allow_large"
  )
})

test_that("aemo_bids validates resolution", {
  expect_error(aemo_bids(duid = "BW01", start = "2024-06-01",
                          end = "2024-06-02", resolution = "minute"))
})

# Internal helpers

#' @noRd
`%||%` <- function(x, y) if (is.null(x)) y else x

#' Build a verified MMSDM archive URL for a named table.
#'
#' MMSDM monthly archives use the naming convention:
#' `PUBLIC_ARCHIVE#TABLENAME#FILE01#YYYYMM010000.zip`
#' (with `#` URL-encoded as `%23`). This helper constructs the
#' URL and is the single source of truth for the pattern.
#'
#' The base URL is read from `options("aemo.mmsdm_base")` so the
#' 30 April 2026 NEMweb migration can be handled without a
#' reinstall.
#'
#' @param table  AEMO table name in UPPER_CASE (e.g. "DUDETAILSUMMARY").
#' @param year   4-digit year string (e.g. "2025").
#' @param month  2-digit month string with leading zero (e.g. "03").
#' @param base   MMSDM base URL. Defaults to the current NEMweb path;
#'   set `options(aemo.mmsdm_base = "...")` to override globally.
#' @return A fully-qualified URL string.
#' @noRd
AEMO_MMSDM_BASE <- "https://nemweb.com.au/Data_Archive/Wholesale_Electricity/MMSDM"

#' @noRd
aemo_mmsdm_base <- function() {
  getOption("aemo.mmsdm_base", default = AEMO_MMSDM_BASE)
}

#' @noRd
aemo_mmsdm_url <- function(table, year, month,
                             base = aemo_mmsdm_base()) {
  sprintf(
    "%s/%s/MMSDM_%s_%s/MMSDM_Historical_Data_SQLLoader/DATA/PUBLIC_ARCHIVE%%23%s%%23FILE01%%23%s%s010000.zip",
    base, year, year, month, table, year, month
  )
}

#' Fetch one MMSDM table from recent monthly archives, trying back
#' up to `max_months_back` months. Returns the first data frame
#' with at least `min_rows` rows, or NULL.
#' @noRd
aemo_fetch_mmsdm_table <- function(table, min_rows = 1L,
                                    max_months_back = 6L) {
  now <- Sys.time()
  attr(now, "tzone") <- AEMO_TIMEZONE
  for (offset in seq_len(max_months_back)) {
    target <- as.Date(now) - (offset * 30L)
    y <- format(target, "%Y")
    m <- format(target, "%m")
    url <- aemo_mmsdm_url(table, y, m)
    df <- tryCatch({
      zip_path <- aemo_download_cached(url)
      tmp <- tempfile(paste0("aemo_", tolower(table), "_"))
      dir.create(tmp, recursive = TRUE)
      on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
      utils::unzip(zip_path, exdir = tmp)
      csvs <- list.files(tmp, pattern = "\\.[Cc][Ss][Vv]$",
                         full.names = TRUE)
      if (length(csvs) == 0L) stop("no csv in zip")
      parsed <- aemo_parse_csv(csvs[[1L]])
      if (length(parsed) == 0L) stop("empty parse")
      parsed[[1L]]
    }, error = function(e) NULL)
    if (!is.null(df) && nrow(df) >= min_rows) {
      return(aemo_coerce_types(df))
    }
  }
  NULL
}

#' @noRd
aemo_format_bytes <- function(x) {
  if (is.na(x) || x < 1024) return(paste0(x, " B"))
  units <- c("KB", "MB", "GB", "TB")
  for (i in seq_along(units)) {
    x <- x / 1024
    if (x < 1024 || i == length(units)) {
      return(sprintf("%.1f %s", x, units[i]))
    }
  }
}

#' @noRd
aemo_clean_names <- function(x) {
  x <- tolower(x)
  x <- gsub("[^a-z0-9]+", "_", x)
  x <- gsub("^_+|_+$", "", x)
  x <- gsub("_+", "_", x)
  x
}

#' NEM market timezone: fixed AEST (UTC+10), no daylight savings.
#'
#' AEMO operates the NEM on Australian Eastern Standard Time
#' year-round (National Electricity Rules clause 2.2.6). We use
#' `"Australia/Brisbane"` rather than `"Australia/Sydney"`
#' because Sydney observes DST and would silently shift every
#' summer timestamp by one hour relative to the AEMO file clock.
#' @noRd
AEMO_TIMEZONE <- "Australia/Brisbane"

#' Valid NEM region codes
#' @noRd
AEMO_REGIONS <- c("NSW1", "QLD1", "SA1", "TAS1", "VIC1")

#' Validate region against the known NEM set.
#' @noRd
aemo_validate_region <- function(region) {
  region <- toupper(region)
  bad <- setdiff(region, AEMO_REGIONS)
  if (length(bad) > 0L) {
    cli::cli_abort(c(
      "Unknown NEM region{?s}: {.val {bad}}.",
      "i" = "Valid regions: {.val {AEMO_REGIONS}}."
    ))
  }
  region
}

#' Parse a user-supplied timestamp to POSIXct in AEMO market time.
#'
#' Always returns AEST (UTC+10, no DST) to match the clock used
#' throughout NEMweb files.
#' @noRd
aemo_parse_time <- function(x) {
  if (inherits(x, "POSIXct")) {
    return(as.POSIXct(format(x, tz = AEMO_TIMEZONE),
                      tz = AEMO_TIMEZONE))
  }
  if (inherits(x, "Date")) {
    return(as.POSIXct(paste0(x, " 00:00:00"), tz = AEMO_TIMEZONE))
  }
  if (is.character(x)) {
    return(as.POSIXct(x, tz = AEMO_TIMEZONE))
  }
  cli::cli_abort("Could not parse time {.val {x}}.")
}

#' Parse an AEMO-format timestamp column to POSIXct AEST.
#'
#' AEMO timestamps appear in several near-identical forms across
#' tables and vintages: `"2024/06/01 00:05:00"`,
#' `"2024-06-01 00:05:00"`, and occasionally ISO 8601. This
#' helper tries each in turn.
#' @noRd
aemo_parse_col_time <- function(x) {
  x <- as.character(x)
  out <- rep(as.POSIXct(NA, tz = AEMO_TIMEZONE), length(x))
  fmts <- c("%Y/%m/%d %H:%M:%S", "%Y-%m-%d %H:%M:%S",
            "%Y/%m/%d %H:%M", "%Y-%m-%d %H:%M",
            "%Y/%m/%dT%H:%M:%S", "%Y-%m-%dT%H:%M:%S")
  for (fmt in fmts) {
    todo <- is.na(out) & !is.na(x) & nzchar(x)
    if (!any(todo)) break
    parsed <- suppressWarnings(
      as.POSIXct(x[todo], format = fmt, tz = AEMO_TIMEZONE)
    )
    out[todo][!is.na(parsed)] <- parsed[!is.na(parsed)]
  }
  out
}

#' Extract the timestamp encoded in a NEMweb filename.
#'
#' Filenames follow the pattern
#' `PUBLIC_<REPORT>_YYYYMMDDHHMM_<seq>.zip` (or `YYYYMMDD` for
#' Archive daily rollups). Returns POSIXct in AEMO_TIMEZONE or
#' NA for filenames that do not match.
#' @noRd
aemo_file_timestamp <- function(filenames) {
  n <- length(filenames)
  out <- rep(as.POSIXct(NA, tz = AEMO_TIMEZONE), n)
  # Try 12-digit (YYYYMMDDHHMM) first
  m12 <- regmatches(filenames, regexpr("[0-9]{12}", filenames))
  hits12 <- which(nzchar(m12))
  if (length(hits12) > 0L) {
    out[hits12] <- suppressWarnings(
      as.POSIXct(m12[hits12], format = "%Y%m%d%H%M", tz = AEMO_TIMEZONE)
    )
  }
  # Fall back to 8-digit (YYYYMMDD) for Archive daily rollups
  todo <- is.na(out)
  if (any(todo)) {
    m8 <- regmatches(filenames[todo], regexpr("[0-9]{8}", filenames[todo]))
    hits8 <- which(nzchar(m8))
    if (length(hits8) > 0L) {
      parsed <- suppressWarnings(
        as.POSIXct(m8[hits8], format = "%Y%m%d", tz = AEMO_TIMEZONE)
      )
      idx <- which(todo)[hits8]
      out[idx] <- parsed
    }
  }
  out
}

#' List NEMweb Current-directory files in a date range.
#'
#' @param path NEMweb subpath under `/Reports/Current/`.
#' @param pattern Regex applied to filenames (e.g. `"DISPATCHIS"`).
#' @param start,end POSIXct in AEMO_TIMEZONE.
#' @return A data frame of files (name, url, file_ts), filtered.
#' @noRd
aemo_select_files_in_range <- function(path, pattern, start, end) {
  files <- aemo_nemweb_ls(path)
  if (nrow(files) == 0L) {
    return(data.frame(name = character(0), url = character(0),
                      file_ts = as.POSIXct(character(0), tz = AEMO_TIMEZONE),
                      stringsAsFactors = FALSE))
  }
  files <- files[grepl(pattern, files$name, ignore.case = TRUE), ,
                 drop = FALSE]
  if (nrow(files) == 0L) return(files)
  files$file_ts <- aemo_file_timestamp(files$name)
  # NEM dispatch files are named by their period-end timestamp;
  # the file named 00:05 contains data for the 5-min period
  # ending at 00:05. Allow a 5-min buffer on each side.
  buffer <- as.difftime(5, units = "mins")
  keep <- !is.na(files$file_ts) &
    files$file_ts >= (start - buffer) &
    files$file_ts <= (end + buffer)
  files[keep, , drop = FALSE]
}

#' Coerce character columns to likely types based on column names.
#'
#' NEMweb CSV parsing returns every column as character. This
#' helper applies a conservative type coercion:
#' - Columns named `*DATE*`, `*DATETIME*`, `*INTERVAL_DATETIME*`
#'   are parsed as POSIXct in AEMO_TIMEZONE.
#' - Columns whose names match known numeric patterns (RRP,
#'   PRICE, MW, FACTOR, LIMIT, LOSS, VOLUME, CAPACITY, DEMAND,
#'   GENERATION, FLOW, ENABLEMENT, INTERCHANGE, and the 10-band
#'   bid price/availability columns) are coerced to numeric.
#' - ID / text columns (DUID, REGIONID, STATION, INTERVENTION,
#'   RUNNO, HUBID, MARKETSUSPENDEDFLAG, etc.) stay character.
#' @noRd
aemo_coerce_types <- function(df) {
  if (!is.data.frame(df) || ncol(df) == 0L) return(df)
  nm <- names(df)
  # Time columns: name ends in "date" or contains "datetime"/"interval_datetime"/"settlementdate"
  time_cols <- grep("(date$|datetime|settlementdate|interval_datetime)",
                    nm, value = TRUE)
  for (col in time_cols) {
    parsed <- aemo_parse_col_time(df[[col]])
    if (sum(!is.na(parsed)) > 0L) df[[col]] <- parsed
  }
  # Numeric columns: explicit known tokens (conservative).
  # "intervention" is numeric 0/1 but kept as character to match
  # the rest of AEMO's flag columns.
  num_tokens <- c(
    "rrp$", "price$", "pricerrp$",
    "^mw", "mw$", "_mw",
    "availability", "availability_mw",
    "demand", "generation", "interchange", "flow",
    "losses", "lossfactor", "marginalvalue", "shadowprice",
    "enablement", "dispatch", "target",
    "priceband", "bandavail", "maxavail",
    "rooftop_pv", "scada", "capacity",
    "total", "lor", "reserve",
    "nplong", "npshort"
  )
  num_re <- paste(num_tokens, collapse = "|")
  num_cols <- grep(num_re, nm, value = TRUE, ignore.case = TRUE)
  # Exclude columns already parsed as time + flag-like columns.
  num_cols <- setdiff(num_cols, time_cols)
  num_cols <- setdiff(num_cols, grep("^(runno|regionid|duid|marketsuspendedflag|intervention|semi_dispatch_cap)$",
                                      nm, value = TRUE))
  for (col in num_cols) {
    parsed <- suppressWarnings(as.numeric(df[[col]]))
    # Only replace if a material share parses (>30%). Otherwise
    # the column is probably not numeric despite the name match.
    non_na_char <- sum(!is.na(df[[col]]) & nzchar(as.character(df[[col]])))
    if (non_na_char > 0L && sum(!is.na(parsed)) / non_na_char > 0.3) {
      df[[col]] <- parsed
    }
  }
  df
}

#' Apply range + intervention + region filters to a parsed
#' AEMO table.
#'
#' @param df A data frame from `aemo_parse_csv()`.
#' @param start,end POSIXct bounds.
#' @param region Optional character vector of region IDs.
#' @param intervention Logical: include intervention runs?
#'   Default `FALSE` filters to `INTERVENTION = 0`.
#' @noRd
aemo_apply_filters <- function(df, start = NULL, end = NULL,
                                region = NULL, intervention = FALSE) {
  if (is.null(df) || nrow(df) == 0L) return(df)
  if (!is.null(region) && "regionid" %in% names(df)) {
    df <- df[df$regionid %in% toupper(region), , drop = FALSE]
  }
  if (!intervention && "intervention" %in% names(df)) {
    iv <- suppressWarnings(as.integer(df$intervention))
    df <- df[is.na(iv) | iv == 0L, , drop = FALSE]
  }
  if ((!is.null(start) || !is.null(end)) && "settlementdate" %in% names(df)) {
    ts <- if (inherits(df$settlementdate, "POSIXct")) {
      df$settlementdate
    } else {
      aemo_parse_col_time(df$settlementdate)
    }
    keep <- !is.na(ts)
    if (!is.null(start)) keep <- keep & ts >= start
    if (!is.null(end))   keep <- keep & ts <= end
    df <- df[keep, , drop = FALSE]
  }
  rownames(df) <- NULL
  df
}

#' Fetch a NEMweb report over a date range and stitch results.
#'
#' Chooses Current/ for recent dates, Archive/ for older dates
#' (daily rollups). Aborts with a clear message if the requested
#' range is older than NEMweb retention (~30 days for Current,
#' years for Archive; MMSDM for multi-year).
#'
#' @param current_dir NEMweb Current subpath (e.g.
#'   `"/Reports/Current/DispatchIS_Reports/"`).
#' @param archive_dir NEMweb Archive subpath, or `NULL` if the
#'   report has no daily Archive rollup.
#' @param pattern Filename regex (e.g. `"DISPATCHIS"`).
#' @param start,end POSIXct AEST.
#' @param table Name of the parsed table to extract from each
#'   file (lowercase `<report>_<table>`, e.g. `"dispatch_price"`).
#'   If `NULL`, returns the first table per file.
#' @return A single stacked data frame (character types before
#'   user coercion).
#' @noRd
aemo_fetch_report_range <- function(current_dir, archive_dir = NULL,
                                     pattern, start, end, table = NULL) {
  now <- Sys.time()
  attr(now, "tzone") <- AEMO_TIMEZONE
  current_window <- now - as.difftime(30, units = "days")

  urls <- character(0)

  if (end >= current_window) {
    current_start <- max(start, current_window)
    files <- aemo_select_files_in_range(current_dir, pattern,
                                         current_start, end)
    urls <- c(urls, files$url)
  }
  if (start < current_window) {
    if (is.null(archive_dir)) {
      cli::cli_abort(c(
        "Requested range extends beyond NEMweb Current retention ({.val {format(current_window, '%Y-%m-%d')}}).",
        "i" = "Archive daily rollups are not yet implemented for this report.",
        "i" = "Use {.fn aemo_nemweb_download} with an explicit MMSDM URL for older data."
      ))
    }
    days <- seq(as.Date(start, tz = AEMO_TIMEZONE),
                min(as.Date(end, tz = AEMO_TIMEZONE),
                    as.Date(current_window, tz = AEMO_TIMEZONE)),
                by = "day")
    files <- aemo_nemweb_ls(archive_dir)
    files <- files[grepl(pattern, files$name, ignore.case = TRUE), ,
                   drop = FALSE]
    if (nrow(files) > 0L) {
      files$file_ts <- aemo_file_timestamp(files$name)
      keep <- !is.na(files$file_ts) &
        as.Date(files$file_ts, tz = AEMO_TIMEZONE) %in% days
      urls <- c(urls, files$url[keep])
    }
  }

  if (length(urls) == 0L) {
    cli::cli_abort(c(
      "No NEMweb files matched the requested range.",
      "i" = "Check {.arg start} and {.arg end}, or the pattern {.val {pattern}}."
    ))
  }

  all_parts <- list()
  for (url in urls) {
    tables <- tryCatch(aemo_fetch_zip(url),
                       error = function(e) {
                         cli::cli_warn("Skipping failed download: {.url {url}}")
                         NULL
                       })
    if (is.null(tables)) next
    df <- if (!is.null(table) && !is.null(tables[[table]])) {
      tables[[table]]
    } else if (length(tables) > 0L) {
      tables[[1L]]
    } else NULL
    if (!is.null(df) && nrow(df) > 0L) {
      all_parts[[length(all_parts) + 1L]] <- df
    }
  }

  if (length(all_parts) == 0L) {
    cli::cli_abort("No data rows found in the fetched files.")
  }

  # Stack on common columns only (handles schema drift)
  common <- Reduce(intersect, lapply(all_parts, names))
  common <- common[nzchar(common)]
  stacked <- do.call(
    rbind,
    lapply(all_parts, function(d) d[, common, drop = FALSE])
  )
  rownames(stacked) <- NULL
  stacked
}

# Offline fixture tests: parse a bundled CSV and assert exact values.
# These run without any network access and cover the full parsing stack
# (csv_parse.R -> aemo_coerce_types -> aemo_apply_filters).
# A failure here means a silent correctness bug in the parser or type coercion.

fixture_path <- function(name) {
  system.file("testdata", name, package = "aemo")
}

test_that("fixture CSV parses with correct structure", {
  path <- fixture_path("PUBLIC_DISPATCHIS_202401150000_FIXTURE.CSV")
  skip_if_not(nzchar(path) && file.exists(path),
               "fixture file not installed")

  tables <- aemo:::aemo_parse_csv(path)
  expect_type(tables, "list")
  expect_true(length(tables) >= 1L)

  # The DISPATCHPRICE table should be present
  nm <- names(tables)
  expect_true(any(grepl("price", nm, ignore.case = TRUE)),
              info = paste("Expected a 'price' table; got:", paste(nm, collapse = ", ")))
})

test_that("fixture DISPATCHPRICE has correct columns after coercion", {
  path <- fixture_path("PUBLIC_DISPATCHIS_202401150000_FIXTURE.CSV")
  skip_if_not(nzchar(path) && file.exists(path),
               "fixture file not installed")

  tables <- aemo:::aemo_parse_csv(path)
  price_nm <- names(tables)[grepl("price", names(tables), ignore.case = TRUE)][1L]
  df <- aemo:::aemo_coerce_types(tables[[price_nm]])

  # 9 data rows across 5 regions / 2 intervals (TAS has no 00:10 row in fixture)
  expect_equal(nrow(df), 9L)

  # settlementdate must be POSIXct in AEST
  expect_s3_class(df$settlementdate, "POSIXct")
  expect_equal(attr(df$settlementdate, "tzone"), "Australia/Brisbane")

  # rrp must be numeric (not character)
  expect_true(is.numeric(df$rrp),
              info = "rrp column must be numeric after type coercion")

  # Known pinned values: NSW1 at 00:05 = 62.35 AUD/MWh
  nsw_0005 <- df[df$regionid == "NSW1" &
                   format(df$settlementdate, "%H:%M") == "00:05", ]
  expect_equal(nrow(nsw_0005), 1L)
  expect_equal(nsw_0005$rrp, 62.35)

  # SA1 at 00:05 = 63.45 AUD/MWh
  sa_0005 <- df[df$regionid == "SA1" &
                  format(df$settlementdate, "%H:%M") == "00:05", ]
  expect_equal(nrow(sa_0005), 1L)
  expect_equal(sa_0005$rrp, 63.45)
})

test_that("intervention filter correctly removes INTERVENTION=1 rows", {
  path <- fixture_path("PUBLIC_DISPATCHIS_202401150000_FIXTURE.CSV")
  skip_if_not(nzchar(path) && file.exists(path),
               "fixture file not installed")

  tables <- aemo:::aemo_parse_csv(path)
  price_nm <- names(tables)[grepl("price", names(tables), ignore.case = TRUE)][1L]
  df <- aemo:::aemo_coerce_types(tables[[price_nm]])

  # Fixture has only INTERVENTION=0 rows; filter should return all rows
  df_filtered <- aemo:::aemo_apply_filters(df, intervention = FALSE)
  expect_equal(nrow(df_filtered), nrow(df))

  # Passing intervention=TRUE should also return all rows (no =1 rows present)
  df_both <- aemo:::aemo_apply_filters(df, intervention = TRUE)
  expect_equal(nrow(df_both), nrow(df))
})

test_that("region filter works on fixture data", {
  path <- fixture_path("PUBLIC_DISPATCHIS_202401150000_FIXTURE.CSV")
  skip_if_not(nzchar(path) && file.exists(path),
               "fixture file not installed")

  tables <- aemo:::aemo_parse_csv(path)
  price_nm <- names(tables)[grepl("price", names(tables), ignore.case = TRUE)][1L]
  df <- aemo:::aemo_coerce_types(tables[[price_nm]])

  df_nsw <- aemo:::aemo_apply_filters(df, region = "NSW1")
  expect_true(all(df_nsw$regionid == "NSW1"))
  expect_equal(nrow(df_nsw), 2L)  # NSW1 appears in both 00:05 and 00:10
})

test_that("aemo_parse_time is AEST and DST-safe on fixture timestamps", {
  # 2024-01-15 is mid-summer in Australia: Sydney is AEDT (UTC+11).
  # Market time must remain AEST (UTC+10). Verify by checking the UTC offset.
  t <- aemo:::aemo_parse_time("2024-01-15 00:05:00")
  # UTC offset for Australia/Brisbane is always +10:00
  offset_hrs <- as.numeric(format(t, "%z")) / 100
  expect_equal(offset_hrs, 10)
})

test_that("aemo_mmsdm_url builds correct PUBLIC_ARCHIVE pattern", {
  url <- aemo:::aemo_mmsdm_url("DUDETAILSUMMARY", "2025", "03")
  expect_true(grepl("PUBLIC_ARCHIVE%23DUDETAILSUMMARY%23FILE01%232025030", url))
  expect_true(grepl("^https://nemweb", url))

  url2 <- aemo:::aemo_mmsdm_url("TRANSMISSIONLOSSFACTOR", "2024", "11")
  expect_true(grepl("TRANSMISSIONLOSSFACTOR", url2))
  expect_true(grepl("%23FILE01%23", url2))
})

test_that("aemo_validate_region accepts valid codes", {
  expect_equal(aemo:::aemo_validate_region("nsw1"), "NSW1")
  expect_equal(aemo:::aemo_validate_region(c("nsw1", "vic1")),
               c("NSW1", "VIC1"))
})

test_that("aemo_validate_region rejects invalid codes", {
  expect_error(aemo:::aemo_validate_region("ACT1"))
  expect_error(aemo:::aemo_validate_region(c("NSW1", "XYZ")))
})

test_that("aemo_parse_time accepts character and POSIXct", {
  p <- aemo:::aemo_parse_time("2024-06-01")
  expect_s3_class(p, "POSIXct")
  q <- aemo:::aemo_parse_time(as.POSIXct("2024-06-01", tz = "UTC"))
  expect_s3_class(q, "POSIXct")
})

test_that("aemo_parse_time rejects garbage", {
  expect_error(aemo:::aemo_parse_time(1234))
})

test_that("aemo_parse_time uses AEST (no DST)", {
  # AEMO market time is fixed AEST. We use Australia/Brisbane
  # because it does not observe DST.
  p <- aemo:::aemo_parse_time("2024-12-01 10:00:00")
  expect_equal(format(p, "%Z", tz = aemo:::AEMO_TIMEZONE), "AEST")
  # DST-transition instant in Sydney civil time (2024-10-06
  # spring-forward 02:00 -> 03:00) exists in AEST.
  p2 <- aemo:::aemo_parse_time("2024-10-06 02:30:00")
  expect_s3_class(p2, "POSIXct")
  expect_false(is.na(p2))
})

test_that("aemo_file_timestamp extracts NEMweb filename timestamps", {
  names <- c(
    "PUBLIC_DISPATCHIS_202604210405_0000000513945454.zip",
    "PUBLIC_DISPATCHIS_20260420.zip",                      # Archive daily
    "NOT_A_NEMWEB_FILE.zip"
  )
  ts <- aemo:::aemo_file_timestamp(names)
  expect_equal(length(ts), 3L)
  expect_false(is.na(ts[1]))
  expect_false(is.na(ts[2]))
  expect_true(is.na(ts[3]))
  # The AEST timestamp for 202604210405 should be 2026-04-21 04:05 AEST.
  expect_equal(format(ts[1], "%Y-%m-%d %H:%M", tz = aemo:::AEMO_TIMEZONE),
               "2026-04-21 04:05")
})

test_that("aemo_parse_col_time handles AEMO CSV date formats", {
  x <- c("2024/06/01 00:05:00",
         "2024-06-01 00:05:00",
         "garbage",
         NA_character_)
  ts <- aemo:::aemo_parse_col_time(x)
  expect_equal(length(ts), 4L)
  expect_false(is.na(ts[1]))
  expect_false(is.na(ts[2]))
  expect_true(is.na(ts[3]))
  expect_true(is.na(ts[4]))
})

test_that("aemo_coerce_types coerces RRP and MW columns to numeric", {
  df <- data.frame(
    settlementdate = c("2024/06/01 00:05:00", "2024/06/01 00:10:00"),
    regionid = c("NSW1", "NSW1"),
    rrp = c("80.50", "102.75"),
    totaldemand = c("7500.0", "7520.5"),
    duid = c("BW01", "BW01"),
    intervention = c("0", "0"),
    stringsAsFactors = FALSE
  )
  out <- aemo:::aemo_coerce_types(df)
  expect_s3_class(out$settlementdate, "POSIXct")
  expect_true(is.numeric(out$rrp))
  expect_true(is.numeric(out$totaldemand))
  expect_equal(out$rrp, c(80.5, 102.75))
  # ID and flag columns stay character
  expect_true(is.character(out$duid))
  expect_true(is.character(out$intervention))
  expect_true(is.character(out$regionid))
})

test_that("aemo_apply_filters drops INTERVENTION = 1 rows by default", {
  df <- data.frame(
    settlementdate = rep(as.POSIXct("2024-06-01 00:05",
                                     tz = aemo:::AEMO_TIMEZONE), 3),
    regionid = c("NSW1", "NSW1", "VIC1"),
    intervention = c("0", "1", "0"),
    rrp = c(80, 999, 60),
    stringsAsFactors = FALSE
  )
  out <- aemo:::aemo_apply_filters(df, region = "NSW1",
                                    intervention = FALSE)
  expect_equal(nrow(out), 1L)
  expect_equal(out$rrp, 80)
})

test_that("aemo_apply_filters keeps INTERVENTION = 1 when requested", {
  df <- data.frame(
    settlementdate = rep(as.POSIXct("2024-06-01 00:05",
                                     tz = aemo:::AEMO_TIMEZONE), 2),
    regionid = c("NSW1", "NSW1"),
    intervention = c("0", "1"),
    rrp = c(80, 999),
    stringsAsFactors = FALSE
  )
  out <- aemo:::aemo_apply_filters(df, region = "NSW1",
                                    intervention = TRUE)
  expect_equal(nrow(out), 2L)
})

test_that("aemo_clean_names snakecases", {
  expect_equal(aemo:::aemo_clean_names(c("Settlement Date", "RRP($)")),
               c("settlement_date", "rrp"))
})

# Schema drift tests: verify that parsing two DISPATCHPRICE fixtures
# with different column sets (pre-5MS v4 without RAISE1SEC* columns,
# and v5 post-9 Oct 2023 with them) stacks correctly on the common
# intersection.

fixture_path <- function(name) {
  system.file("testdata", name, package = "aemo")
}

test_that("v4 (pre-5MS) DISPATCHPRICE fixture parses with reduced column set", {
  path <- fixture_path("PUBLIC_DISPATCHIS_202101150000_V4_FIXTURE.CSV")
  skip_if_not(nzchar(path) && file.exists(path),
               "v4 fixture not installed")

  tables <- aemo:::aemo_parse_csv(path)
  expect_true("dispatch_price" %in% names(tables))
  df <- aemo:::aemo_coerce_types(tables$dispatch_price)

  # v4 has 8 FCAS RRP services (no RAISE1SEC / LOWER1SEC)
  expect_false("raise1secrrp" %in% names(df))
  expect_false("lower1secrrp" %in% names(df))
  expect_true("raise6secrrp" %in% names(df))
  expect_equal(nrow(df), 5L)

  # NSW1 at 00:05 on 15-Jan-2021: 45.10 AUD/MWh
  nsw <- df[df$regionid == "NSW1", ]
  expect_equal(nrow(nsw), 1L)
  expect_equal(nsw$rrp, 45.10)
})

test_that("v4 and v5 DISPATCHPRICE fixtures stack on common intersection", {
  v4 <- fixture_path("PUBLIC_DISPATCHIS_202101150000_V4_FIXTURE.CSV")
  v5 <- fixture_path("PUBLIC_DISPATCHIS_202401150000_FIXTURE.CSV")
  skip_if_not(nzchar(v4) && file.exists(v4) &&
                nzchar(v5) && file.exists(v5),
               "fixtures not installed")

  df4 <- aemo:::aemo_coerce_types(aemo:::aemo_parse_csv(v4)$dispatch_price)
  df5 <- aemo:::aemo_coerce_types(aemo:::aemo_parse_csv(v5)$dispatch_price)

  # Intersection of columns
  common <- intersect(names(df4), names(df5))
  # The common set must include the canonical analytical columns
  expect_true(all(c("settlementdate", "regionid", "rrp",
                    "raise6secrrp", "lower6secrrp") %in% common))
  # But must NOT include v5-only additions
  expect_false("raise1secrrp" %in% common)
  expect_false("bestdispatchintervalselection" %in% common)

  # Stack and confirm row counts and RRPs are preserved
  stacked <- rbind(df4[, common, drop = FALSE], df5[, common, drop = FALSE])
  expect_equal(nrow(stacked), nrow(df4) + nrow(df5))
  expect_true(is.numeric(stacked$rrp))
  # All timestamps in AEST
  expect_equal(attr(stacked$settlementdate, "tzone"), "Australia/Brisbane")
})

#' Gas market data (STTM, DWGM)
#'
#' Returns Short Term Trading Market (`STTM`, Adelaide, Brisbane,
#' Sydney hubs) or Declared Wholesale Gas Market (`DWGM`,
#' Victoria) prices and volumes.
#'
#' @param market One of `"sttm"` (default) or `"dwgm"`.
#' @param hub Optional STTM hub: `"adelaide"`, `"brisbane"`, or
#'   `"sydney"`. Ignored for DWGM.
#' @param start,end Window.
#'
#' @return An `aemo_tbl`.
#'
#' @family gas
#' @export
#' @examples
#' \donttest{
#' op <- options(aemo.cache_dir = tempdir())
#' try({
#'   now <- Sys.time()
#'   g <- aemo_gas(market = "sttm", hub = "sydney",
#'                  start = now - 7 * 86400, end = now)
#'   head(g)
#' })
#' options(op)
#' }
aemo_gas <- function(market = c("sttm", "dwgm"),
                      hub = NULL, start, end) {
  market <- match.arg(market)
  start <- aemo_parse_time(start)
  end <- aemo_parse_time(end)
  stopifnot(end >= start)

  if (market == "sttm") {
    current <- "/Reports/Current/STTM/"
    archive <- "/Reports/Archive/STTM/"
    pattern <- "STTM"
  } else {
    current <- "/Reports/Current/VicGas/"
    archive <- "/Reports/Archive/VicGas/"
    pattern <- "VICGAS"
  }

  df <- aemo_fetch_report_range(
    current_dir = current, archive_dir = archive,
    pattern = pattern, start = start, end = end
  )
  df <- aemo_coerce_types(df)
  # Gas tables use GASDATE / INTERVAL_DATETIME rather than
  # settlementdate; normalise for the filter step.
  for (alias in c("gasdate", "interval_datetime")) {
    if (!"settlementdate" %in% names(df) && alias %in% names(df)) {
      df$settlementdate <- df[[alias]]
      break
    }
  }
  df <- aemo_apply_filters(df, start = start, end = end)

  if (market == "sttm" && !is.null(hub) && "hubid" %in% names(df)) {
    df <- df[toupper(df$hubid) %in% toupper(hub), , drop = FALSE]
    rownames(df) <- NULL
  }
  new_aemo_tbl(df,
               source = "http://nemweb.com.au",
               title = paste0("AEMO ", market, " gas"))
}

#' NEM interconnector flows
#'
#' Returns MW flows, losses, and limits for one or more NEM
#' interconnectors from `DISPATCHINTERCONNECTORRES`.
#'
#' AEMO's `METEREDMWFLOW` is positive when power flows from
#' `REGIONFROM` to `REGIONTO`. For per-interconnector direction
#' conventions see [aemo_interconnectors()].
#'
#' @param flow Optional character vector of interconnector IDs.
#' @param start,end Window.
#' @param intervention Logical. Default `FALSE`.
#'
#' @return An `aemo_tbl`.
#'
#' @family dispatch
#' @export
#' @examples
#' \donttest{
#' op <- options(aemo.cache_dir = tempdir())
#' try({
#'   now <- Sys.time()
#'   i <- aemo_interconnector(flow = "V-SA",
#'                             start = now - 3600, end = now)
#'   head(i)
#' })
#' options(op)
#' }
aemo_interconnector <- function(flow = NULL, start, end,
                                 intervention = FALSE) {
  start <- aemo_parse_time(start)
  end <- aemo_parse_time(end)
  stopifnot(end >= start)

  df <- aemo_fetch_report_range(
    current_dir = "/Reports/Current/DispatchIS_Reports/",
    archive_dir = "/Reports/Archive/DispatchIS_Reports/",
    pattern = "DISPATCHIS",
    start = start, end = end,
    table = "dispatch_interconnectorres"
  )
  df <- aemo_coerce_types(df)
  df <- aemo_apply_filters(df, start = start, end = end,
                           intervention = intervention)

  if (!is.null(flow) && "interconnectorid" %in% names(df)) {
    df <- df[toupper(df$interconnectorid) %in% toupper(flow), ,
             drop = FALSE]
    rownames(df) <- NULL
  }
  new_aemo_tbl(df,
               source = "http://nemweb.com.au",
               title = "AEMO interconnector flows")
}

# Binding constraints (DISPATCHCONSTRAINT) and constraint equations (GENCONDATA)

#' Binding transmission and system constraints
#'
#' Returns the `DISPATCHCONSTRAINT` table from NEMweb: one row
#' per binding (or near-binding) constraint per 5-minute dispatch
#' interval. Each row records the constraint ID, the left-hand
#' side (LHS) and right-hand side (RHS) values, the marginal
#' value (shadow price on the constraint in AUD/MWh), and the
#' violation degree if any.
#'
#' This is the table that answers the question "why was the RRP
#' so high at 17:35?": the sum of marginal values across
#' binding constraints at the Regional Reference Node equals the
#' regional price component attributable to network limits.
#'
#' Constraint equations and RHS terms (`GENCONDATA`,
#' `GENCONSETINVOKE`) are published through MMSDM on a separate
#' cadence and are not exposed directly by this function; use
#' [aemo_nemweb_download()] on an MMSDM URL for those.
#'
#' @param start,end Window (inclusive), character or POSIXct.
#' @param constraint_id Optional character vector of constraint
#'   IDs (e.g. `"N>>N-NIL_1"`, `"V::S_NIL_TBSE"`). `NULL`
#'   returns all binding constraints in the window.
#' @param intervention Logical. Default `FALSE`.
#' @param min_marginal_value Numeric. Only return constraints
#'   with marginal value at or above this threshold (AUD/MWh).
#'   `0` returns every row; the default of `0.01` filters out
#'   near-zero shadow prices that are typically noise.
#'
#' @return An `aemo_tbl` with columns `settlementdate`,
#'   `constraintid`, `rhs`, `marginalvalue`, `violationdegree`,
#'   `lhs`, plus the intervention flag.
#'
#' @source AEMO NEMweb DISPATCHIS_Reports, DISPATCHCONSTRAINT
#'   table.
#'
#' @family dispatch
#' @export
#' @examples
#' \donttest{
#' op <- options(aemo.cache_dir = tempdir())
#' try({
#'   now <- Sys.time()
#'   c <- aemo_constraints(start = now - 3600, end = now)
#'   head(c)
#' })
#' options(op)
#' }
aemo_constraints <- function(start, end, constraint_id = NULL,
                              intervention = FALSE,
                              min_marginal_value = 0.01) {
  start <- aemo_parse_time(start)
  end <- aemo_parse_time(end)
  stopifnot(end >= start)

  df <- aemo_fetch_report_range(
    current_dir = "/Reports/Current/DispatchIS_Reports/",
    archive_dir = "/Reports/Archive/DispatchIS_Reports/",
    pattern = "DISPATCHIS",
    start = start, end = end,
    table = "dispatch_constraint"
  )
  df <- aemo_coerce_types(df)
  df <- aemo_apply_filters(df, start = start, end = end,
                           intervention = intervention)

  if (!is.null(constraint_id) && "constraintid" %in% names(df)) {
    df <- df[df$constraintid %in% constraint_id, , drop = FALSE]
  }
  if (!is.null(min_marginal_value) && "marginalvalue" %in% names(df)) {
    mv <- suppressWarnings(as.numeric(df$marginalvalue))
    df <- df[!is.na(mv) & abs(mv) >= min_marginal_value, , drop = FALSE]
  }
  rownames(df) <- NULL
  new_aemo_tbl(df,
               source = "http://nemweb.com.au",
               title = "AEMO dispatch constraints (shadow prices)")
}

#' Generic constraint equations and RHS terms (GENCONDATA)
#'
#' Downloads the `GENCONDATA` table from the most recent MMSDM
#' monthly archive. `GENCONDATA` contains the equation
#' definitions for every generic constraint active in the NEM:
#' the constraint ID, the type (equality/inequality), a
#' description of what the constraint models (thermal limit,
#' voltage stability, system strength, etc.), and the default
#' RHS value.
#'
#' Pair this with [aemo_constraints()] to go from a binding
#' dispatch interval to the underlying network equation. The
#' workflow is: `aemo_constraints()` tells you *which* constraint
#' bound and *how hard* (marginalvalue = shadow price);
#' `aemo_gencon()` tells you *what the constraint is* (which
#' elements and why the RHS was set at that level).
#'
#' @param constraint_id Optional character vector of constraint
#'   IDs to filter on. `NULL` returns all equations.
#' @param type Optional constraint type filter (e.g.
#'   `"LE"` for `<=`, `"GE"` for `>=`, `"EQ"` for `=`).
#'   `NULL` returns all types.
#'
#' @return An `aemo_tbl` with columns including `genconid`
#'   (constraint ID), `constrainttype`, `description`,
#'   `genericconstraintrhs` (default RHS value). Additional
#'   columns from GENCONDATA (effective dates, generic
#'   constraint equation weights) will be present when
#'   available.
#'
#' @source AEMO NEMweb MMSDM archive, GENCONDATA table.
#'   AEMO Copyright Permissions Notice.
#'
#' @seealso [aemo_constraints()] for the real-time binding
#'   constraint shadow prices.
#'
#' @family dispatch
#' @export
#' @examples
#' \donttest{
#' op <- options(aemo.cache_dir = tempdir())
#' try({
#'   # Find equations for the Heywood interconnector thermal limits
#'   g <- aemo_gencon(constraint_id = c("V::S_NIL_TBSE", "V::S_NIL_FCSPS"))
#'   head(g)
#' })
#' options(op)
#' }
aemo_gencon <- function(constraint_id = NULL, type = NULL) {
  df <- tryCatch(aemo_fetch_gencondata(),
                 error = function(e) NULL)

  if (is.null(df) || nrow(df) == 0L) {
    cli::cli_abort(c(
      "Could not retrieve GENCONDATA from MMSDM.",
      "i" = "Check your internet connection or use {.fn aemo_nemweb_download} with an MMSDM URL directly.",
      "i" = "MMSDM path: DATA/Archive/Wholesale_Electricity/MMSDM/<YYYY>/<month>/MMSDM_Historical_Data_SQLLoader/DATA/PUBLIC_DVD_GENCONDATA_<YYYYMM>010000.zip"
    ))
  }

  if (!is.null(constraint_id)) {
    id_col <- intersect(c("genconid", "constraintid"), names(df))[1L]
    if (!is.na(id_col)) {
      df <- df[df[[id_col]] %in% constraint_id, , drop = FALSE]
    }
  }
  if (!is.null(type) && "constrainttype" %in% names(df)) {
    df <- df[toupper(df$constrainttype) %in% toupper(type), , drop = FALSE]
  }
  rownames(df) <- NULL

  new_aemo_tbl(df,
               source = "https://nemweb.com.au/Data_Archive/Wholesale_Electricity/MMSDM",
               licence = "AEMO Copyright Permissions Notice",
               title = "NEM generic constraint equations (GENCONDATA)")
}

#' Fetch GENCONDATA from the most recent MMSDM archive.
#' @noRd
aemo_fetch_gencondata <- function(max_months_back = 6L) {
  aemo_fetch_mmsdm_table("GENCONDATA", min_rows = 1L,
                          max_months_back = max_months_back)
}

#' SPD constraint tables (regions, interconnectors, connection points)
#'
#' Returns the SPD (Security and Projected Dispatch) constraint
#' coefficient tables from MMSDM. Where `GENCONDATA` gives the
#' high-level equation definition, the SPD tables give the
#' per-term *coefficients* used in the NEMDE dispatch
#' optimisation:
#'
#' - `"region"`: `SPDREGIONCONSTRAINT` (regional-demand and
#'   regional-generation terms).
#' - `"interconnector"`: `SPDINTERCONNECTORCONSTRAINT`
#'   (interconnector-flow terms).
#' - `"connection_point"`: `SPDCONNECTIONPOINTCONSTRAINT`
#'   (per-DUID connection-point terms).
#'
#' These tables are required for NEM dispatch replication (the
#' `nempy` Python package, Gorman, Bruce & MacGill 2022,
#' *JOSS* 7(70) 3596, <doi:10.21105/joss.03596>, uses all three
#' when reproducing NEMDE solves).
#'
#' @param table One of `"region"` (default), `"interconnector"`,
#'   or `"connection_point"`.
#' @param constraint_id Optional character vector of
#'   `GENCONID`/`CONSTRAINTID` values.
#'
#' @return An `aemo_tbl` with columns from the requested SPD
#'   table. `GENCONID` and `FACTOR` (the coefficient) are
#'   always present; other columns vary by table.
#'
#' @source AEMO NEMweb MMSDM archive, SPDREGIONCONSTRAINT /
#'   SPDINTERCONNECTORCONSTRAINT / SPDCONNECTIONPOINTCONSTRAINT.
#'
#' @seealso [aemo_gencon()] for the equation-level metadata,
#'   [aemo_constraints()] for the 5-min binding-constraint
#'   feed with shadow prices.
#'
#' @family dispatch
#' @export
#' @examples
#' \donttest{
#' op <- options(aemo.cache_dir = tempdir())
#' try({
#'   s <- aemo_spd_constraints(table = "interconnector")
#'   head(s)
#' })
#' options(op)
#' }
aemo_spd_constraints <- function(table = c("region", "interconnector",
                                            "connection_point"),
                                  constraint_id = NULL) {
  table <- match.arg(table)
  mmsdm_table <- switch(table,
    region            = "SPDREGIONCONSTRAINT",
    interconnector    = "SPDINTERCONNECTORCONSTRAINT",
    connection_point  = "SPDCONNECTIONPOINTCONSTRAINT"
  )

  df <- tryCatch(
    aemo_fetch_mmsdm_table(mmsdm_table, min_rows = 1L),
    error = function(e) NULL
  )
  if (is.null(df) || nrow(df) == 0L) {
    cli::cli_abort(c(
      "Could not retrieve {mmsdm_table} from MMSDM.",
      "i" = "Try {.fn aemo_nemweb_download} with an MMSDM URL directly."
    ))
  }

  if (!is.null(constraint_id)) {
    id_col <- intersect(c("genconid", "constraintid"), names(df))[1L]
    if (!is.na(id_col)) {
      df <- df[df[[id_col]] %in% constraint_id, , drop = FALSE]
      rownames(df) <- NULL
    }
  }
  new_aemo_tbl(df,
               source = "https://nemweb.com.au/Data_Archive/Wholesale_Electricity/MMSDM",
               licence = "AEMO Copyright Permissions Notice",
               title = paste0("NEM ", mmsdm_table,
                              " (SPD constraint coefficients)"))
}

# Reproducibility helper: aemo_snapshot()

#' Snapshot provenance for an aemo_tbl
#'
#' Returns a one-row-per-source provenance record for an
#' `aemo_tbl`, suitable for inclusion in a paper appendix, a
#' Zenodo deposit, or a CRAN-style data manifest. The snapshot
#' captures:
#'
#' - `title`: the human-readable title of the table;
#' - `source`: the NEMweb / MMSDM URL family the table was drawn
#'   from;
#' - `licence`: the AEMO Copyright Permissions Notice (always);
#' - `retrieved`: the `POSIXct` timestamp at which the table was
#'   constructed;
#' - `rows`, `cols`: observed dimensions;
#' - `sha256`: a SHA-256 digest of the table's printed body,
#'   stable across R versions and platforms.
#'
#' The `sha256` column is what makes the snapshot *pinnable*: if
#' the same query returns a different hash, the underlying data
#' has changed (or the row-order has, which is also worth
#' knowing). Pair with a git commit of the analysis script to
#' give a reader a closed reproducibility loop.
#'
#' @param x An `aemo_tbl` (or a list of them).
#' @return A data frame with one row per table.
#'
#' @family reference
#' @export
#' @examples
#' x <- structure(
#'   data.frame(settlementdate = Sys.time(), region = "NSW1", rrp = 80),
#'   aemo_title = "Demo",
#'   aemo_source = "http://nemweb.com.au",
#'   aemo_licence = "AEMO Copyright Permissions Notice",
#'   aemo_retrieved = Sys.time(),
#'   class = c("aemo_tbl", "data.frame")
#' )
#' aemo_snapshot(x)
aemo_snapshot <- function(x) {
  if (inherits(x, "aemo_tbl") || is.data.frame(x)) {
    x <- list(x)
  }
  if (!is.list(x) || length(x) == 0L) {
    cli::cli_abort("{.arg x} must be an {.cls aemo_tbl} or a non-empty list of them.")
  }

  rows <- lapply(x, aemo_snapshot_row)
  out <- do.call(rbind, rows)
  rownames(out) <- NULL
  out
}

#' @noRd
aemo_snapshot_row <- function(x) {
  if (!is.data.frame(x)) {
    cli::cli_abort("Each element must be a data frame.")
  }
  title <- attr(x, "aemo_title") %||% NA_character_
  src   <- attr(x, "aemo_source") %||% NA_character_
  lic   <- attr(x, "aemo_licence") %||% "AEMO Copyright Permissions Notice"
  retr  <- attr(x, "aemo_retrieved") %||% as.POSIXct(NA)

  # SHA-256 over a canonical serialisation: drop attributes + row names,
  # coerce all columns to character, paste with a sentinel separator.
  body <- x
  attributes(body)[setdiff(names(attributes(body)),
                            c("names", "row.names", "class"))] <- NULL
  class(body) <- "data.frame"
  rownames(body) <- NULL
  flat <- do.call(paste, c(lapply(body, as.character), sep = "\x1f"))
  digest <- as.character(openssl::sha256(charToRaw(paste(flat, collapse = "\n"))))

  data.frame(
    title = title,
    source = src,
    licence = lic,
    retrieved = retr,
    rows = nrow(x),
    cols = ncol(x),
    sha256 = digest,
    stringsAsFactors = FALSE
  )
}

test_that("aemo_snapshot returns one row per aemo_tbl with expected columns", {
  x <- structure(
    data.frame(settlementdate = as.POSIXct("2024-06-01", tz = "Australia/Brisbane"),
               regionid = "NSW1", rrp = 80.5),
    aemo_title = "Test",
    aemo_source = "http://nemweb.com.au",
    aemo_licence = "AEMO Copyright Permissions Notice",
    aemo_retrieved = as.POSIXct("2024-06-01 12:00:00", tz = "Australia/Brisbane"),
    class = c("aemo_tbl", "data.frame")
  )

  snap <- aemo_snapshot(x)
  expect_s3_class(snap, "data.frame")
  expect_equal(nrow(snap), 1L)
  expect_true(all(c("title", "source", "licence", "retrieved",
                    "rows", "cols", "sha256") %in% names(snap)))
  expect_equal(snap$title, "Test")
  expect_equal(snap$source, "http://nemweb.com.au")
  expect_equal(snap$rows, 1L)
  expect_equal(snap$cols, 3L)
  # SHA-256 is hex, 64 chars
  expect_match(snap$sha256, "^[0-9a-f]{64}$")
})

test_that("aemo_snapshot is deterministic for identical data", {
  make <- function() {
    structure(
      data.frame(x = 1:3, y = letters[1:3]),
      aemo_title = "T",
      aemo_source = "s",
      aemo_licence = "l",
      aemo_retrieved = as.POSIXct("2024-01-01", tz = "UTC"),
      class = c("aemo_tbl", "data.frame")
    )
  }
  expect_equal(aemo_snapshot(make())$sha256, aemo_snapshot(make())$sha256)
})

test_that("aemo_snapshot detects any row change", {
  a <- structure(data.frame(x = 1:3), aemo_title = "T",
                 class = c("aemo_tbl", "data.frame"))
  b <- structure(data.frame(x = c(1:2, 4L)), aemo_title = "T",
                 class = c("aemo_tbl", "data.frame"))
  expect_false(identical(aemo_snapshot(a)$sha256,
                         aemo_snapshot(b)$sha256))
})

test_that("aemo_snapshot accepts a list of aemo_tbls", {
  mk <- function(n) structure(data.frame(x = seq_len(n)),
                              aemo_title = paste("T", n),
                              class = c("aemo_tbl", "data.frame"))
  snap <- aemo_snapshot(list(mk(1L), mk(2L)))
  expect_equal(nrow(snap), 2L)
  expect_equal(snap$rows, c(1L, 2L))
})

# HTTP helpers

#' @noRd
AEMO_BASE_URL <- "http://nemweb.com.au"

#' @noRd
aemo_base_url <- function() {
  getOption("aemo.base_url", default = AEMO_BASE_URL)
}

#' @noRd
aemo_user_agent <- function() {
  paste0("aemo R package/", utils::packageVersion("aemo"),
         " (+https://github.com/charlescoverdale/aemo)")
}

#' @noRd
aemo_request <- function(url, timeout = 180) {
  delay <- getOption("aemo.throttle_delay", 1)
  httr2::request(url) |>
    httr2::req_user_agent(aemo_user_agent()) |>
    httr2::req_throttle(rate = 1 / delay) |>
    httr2::req_timeout(timeout) |>
    httr2::req_retry(max_tries = 3) |>
    httr2::req_error(is_error = function(r) FALSE)
}

#' @noRd
aemo_download_cached <- function(url, cache = TRUE) {
  d <- aemo_cache_dir()
  ext <- tools::file_ext(url)
  ext <- if (nzchar(ext)) paste0(".", ext) else ""
  slug <- aemo_digest_url(url)
  file <- file.path(d, paste0(slug, ext))
  hash_file <- file.path(d, paste0(slug, ".sha256"))

  if (cache && file.exists(file) && file.size(file) > 0L) {
    # Verify content hash if a sidecar exists; re-download on mismatch.
    if (file.exists(hash_file)) {
      stored <- trimws(readLines(hash_file, warn = FALSE)[1L])
      actual <- aemo_sha256_file(file)
      if (!identical(stored, actual)) {
        cli::cli_warn("Cache integrity check failed for {.file {basename(file)}}; re-downloading.")
        unlink(c(file, hash_file), force = TRUE)
      } else {
        return(file)
      }
    } else {
      # No sidecar yet. Write one so future calls can verify.
      writeLines(aemo_sha256_file(file), hash_file)
      return(file)
    }
  }

  cli::cli_progress_step("Downloading {.url {url}}")

  resp <- tryCatch(
    aemo_request(url) |>
      httr2::req_perform(path = file),
    error = function(e) {
      if (file.exists(file)) unlink(file)
      cli::cli_abort(c("Download failed.", "x" = conditionMessage(e)))
    }
  )

  if (!is.null(resp) && httr2::resp_status(resp) >= 400L) {
    unlink(c(file, hash_file), force = TRUE)
    cli::cli_abort("HTTP {httr2::resp_status(resp)} from {.url {url}}.")
  }

  if (file.exists(file) && file.size(file) > 0L) {
    writeLines(aemo_sha256_file(file), hash_file)
  }
  file
}

#' Compute a SHA-256 hex digest of a file.
#'
#' `openssl` is an Import (see DESCRIPTION), so it is always
#' available. Used for cache-integrity verification, not
#' cryptographic security.
#' @noRd
aemo_sha256_file <- function(path) {
  as.character(openssl::sha256(file(path, "rb")))
}

#' Cache key for a URL.
#'
#' Collision-free: SHA-256 of the URL bytes, truncated to 16
#' hex characters (64 bits). At this width the expected number
#' of collisions across AEMO's entire MMSDM history is
#' effectively zero. v0.1.0 used a bespoke weighted-character
#' checksum with a realistic collision probability on large
#' caches.
#' @noRd
aemo_digest_url <- function(url) {
  substr(as.character(openssl::sha256(charToRaw(url))), 1L, 16L)
}

test_that("aemo_digest_url is deterministic", {
  a <- aemo:::aemo_digest_url("http://nemweb.com.au/a")
  b <- aemo:::aemo_digest_url("http://nemweb.com.au/a")
  c <- aemo:::aemo_digest_url("http://nemweb.com.au/b")
  expect_equal(a, b)
  expect_false(identical(a, c))
})

test_that("aemo_user_agent includes package version", {
  ua <- aemo:::aemo_user_agent()
  expect_match(ua, "^aemo R package/")
})

test_that("aemo_base_url honours option", {
  op <- options(aemo.base_url = "http://example.com")
  on.exit(options(op), add = TRUE)
  expect_equal(aemo:::aemo_base_url(), "http://example.com")
})

test_that("aemo_request returns httr2_request", {
  req <- aemo:::aemo_request("http://nemweb.com.au")
  expect_s3_class(req, "httr2_request")
})

test_that("aemo_digest_url is deterministic and 16 hex characters", {
  a <- aemo:::aemo_digest_url("http://nemweb.com.au/Reports/Current/DispatchIS/x.zip")
  b <- aemo:::aemo_digest_url("http://nemweb.com.au/Reports/Current/DispatchIS/x.zip")
  expect_equal(a, b)
  expect_match(a, "^[0-9a-f]{16}$")
})

test_that("aemo_digest_url avoids trivial collisions", {
  a <- aemo:::aemo_digest_url("http://nemweb.com.au/a.zip")
  b <- aemo:::aemo_digest_url("http://nemweb.com.au/b.zip")
  expect_false(identical(a, b))

  # Anagrams must not collide (v0.1.0's weighted checksum could)
  c1 <- aemo:::aemo_digest_url("http://x/AB.zip")
  c2 <- aemo:::aemo_digest_url("http://x/BA.zip")
  expect_false(identical(c1, c2))
})

# Network outages (NETWORK_OUTAGEDETAIL)

#' Planned and forced network outages
#'
#' Returns the `NETWORK_OUTAGEDETAIL` table from MMSDM, which
#' records every planned and forced outage on NEM transmission
#' and distribution network elements. Outages are a primary
#' driver of binding constraints and price spikes: when a line
#' or transformer is out of service the network is more
#' constrained, reducing the thermal limits that appear as RHS
#' values in DISPATCHCONSTRAINT.
#'
#' **Use case.** Pair with [aemo_constraints()] to explain a
#' price spike: `aemo_outages()` tells you which elements were
#' off-service at the time; `aemo_constraints()` tells you
#' which constraints bound; together they answer "why was SA
#' spot price AUD 15,000 at 17:35?".
#'
#' @param start,end Outage window (inclusive). Filters on
#'   `starttime` and `endtime`: any outage active during the
#'   window is returned. Character or POSIXct.
#' @param element_id Optional character vector of network
#'   element IDs. `NULL` returns all elements.
#' @param outage_type Optional character. One of `"PLANNED"`,
#'   `"FORCED"`, or `NULL` (both). Case-insensitive.
#' @param region Optional NEM region code. Filters on the
#'   region column where available.
#'
#' @return An `aemo_tbl` with columns from
#'   NETWORK_OUTAGEDETAIL including `outageid`, `starttime`,
#'   `endtime`, `substationid`, `equipmenttype`,
#'   `equipmentid`, `outagetype`, `regionid`, and
#'   `restarttimeunknown`. Exact column set depends on the
#'   MMSDM version.
#'
#' @source AEMO NEMweb MMSDM archive, NETWORK_OUTAGEDETAIL
#'   table. AEMO Copyright Permissions Notice.
#'
#' @seealso [aemo_constraints()] for the binding constraint
#'   shadow prices that these outages drive.
#'
#' @family dispatch
#' @export
#' @examples
#' \donttest{
#' op <- options(aemo.cache_dir = tempdir())
#' try({
#'   # Forced outages during a recent high-price period in SA
#'   o <- aemo_outages(
#'     start = "2024-03-01",
#'     end   = "2024-03-02",
#'     outage_type = "FORCED"
#'   )
#'   head(o)
#' })
#' options(op)
#' }
aemo_outages <- function(start, end,
                          element_id = NULL,
                          outage_type = NULL,
                          region = NULL) {
  start <- aemo_parse_time(start)
  end <- aemo_parse_time(end)
  stopifnot(end >= start)
  if (!is.null(region)) region <- aemo_validate_region(region)
  if (!is.null(outage_type)) {
    outage_type <- toupper(outage_type)
    valid_types <- c("PLANNED", "FORCED")
    bad <- setdiff(outage_type, valid_types)
    if (length(bad) > 0L) {
      cli::cli_abort(c(
        "Unknown outage type{?s}: {.val {bad}}.",
        "i" = "Valid types: {.val {valid_types}}."
      ))
    }
  }

  df <- tryCatch(
    aemo_fetch_outagedetail(start = start, end = end),
    error = function(e) {
      cli::cli_abort(c(
        "Could not retrieve NETWORK_OUTAGEDETAIL from MMSDM.",
        "x" = conditionMessage(e),
        "i" = "Try {.fn aemo_nemweb_download} with an explicit MMSDM URL."
      ))
    }
  )

  if (is.null(df) || nrow(df) == 0L) {
    cli::cli_abort("No outage data found for the requested range.")
  }

  df <- aemo_coerce_types(df)

  # Filter: any outage that overlaps the requested window
  start_col <- intersect(c("starttime", "outagestarttime"), names(df))[1L]
  end_col   <- intersect(c("endtime",   "outageendtime"),   names(df))[1L]
  if (!is.na(start_col) && !is.na(end_col)) {
    s <- if (inherits(df[[start_col]], "POSIXct")) df[[start_col]] else
      aemo_parse_col_time(df[[start_col]])
    e <- if (inherits(df[[end_col]], "POSIXct")) df[[end_col]] else
      aemo_parse_col_time(df[[end_col]])
    # Overlap: outage starts before window end AND ends after window start
    # (or end is NA, meaning the outage is still open)
    keep <- !is.na(s) & s <= end & (is.na(e) | e >= start)
    df <- df[keep, , drop = FALSE]
  }

  if (!is.null(element_id)) {
    id_col <- intersect(c("equipmentid", "elementid"), names(df))[1L]
    if (!is.na(id_col)) {
      df <- df[df[[id_col]] %in% element_id, , drop = FALSE]
    }
  }
  if (!is.null(outage_type) && "outagetype" %in% names(df)) {
    df <- df[toupper(df$outagetype) %in% outage_type, , drop = FALSE]
  }
  if (!is.null(region) && "regionid" %in% names(df)) {
    df <- df[df$regionid %in% region, , drop = FALSE]
  }
  rownames(df) <- NULL

  new_aemo_tbl(df,
               source = "https://nemweb.com.au/Data_Archive/Wholesale_Electricity/MMSDM",
               licence = "AEMO Copyright Permissions Notice",
               title = "NEM network outages (NETWORK_OUTAGEDETAIL)")
}

#' Fetch NETWORK_OUTAGEDETAIL from MMSDM for a date range.
#' @noRd
aemo_fetch_outagedetail <- function(start, end, max_months_back = 6L) {
  now <- Sys.time()
  attr(now, "tzone") <- AEMO_TIMEZONE

  # Determine which monthly archives to fetch (one per calendar month in range)
  months_needed <- unique(format(
    seq(as.Date(start, tz = AEMO_TIMEZONE),
        as.Date(end,   tz = AEMO_TIMEZONE),
        by = "month"),
    "%Y-%m"
  ))
  # Also try the most recent archive if range is very recent
  recent <- format(as.Date(now) - 30L, "%Y-%m")
  months_needed <- union(months_needed, recent)

  all_parts <- list()
  for (ym in months_needed) {
    y <- substr(ym, 1L, 4L)
    m <- substr(ym, 6L, 7L)
    url <- aemo_mmsdm_url("NETWORK_OUTAGEDETAIL", y, m)
    df <- tryCatch({
      zip_path <- aemo_download_cached(url)
      tmp <- tempfile("aemo_out_")
      dir.create(tmp, recursive = TRUE)
      on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
      utils::unzip(zip_path, exdir = tmp)
      csvs <- list.files(tmp, pattern = "\\.[Cc][Ss][Vv]$", full.names = TRUE)
      if (length(csvs) == 0L) stop("no csv")
      parsed <- aemo_parse_csv(csvs[[1L]])
      parsed[[1L]]
    }, error = function(e) NULL)

    if (!is.null(df) && nrow(df) > 0L) {
      all_parts[[length(all_parts) + 1L]] <- df
    }
  }

  if (length(all_parts) == 0L) return(NULL)
  common <- Reduce(intersect, lapply(all_parts, names))
  stacked <- do.call(rbind,
                     lapply(all_parts, function(d) d[, common, drop = FALSE]))
  rownames(stacked) <- NULL
  stacked
}

# Low-level NEMweb access

#' List files in a NEMweb directory
#'
#' Returns a data frame of files in a NEMweb path, parsed from
#' the Apache directory-listing HTML.
#'
#' @param path NEMweb subpath (e.g. `"/Reports/Current/DispatchIS_Reports/"`).
#'   Leading and trailing slashes are optional.
#'
#' @return A data frame with `name`, `modified`, `size`, `url`.
#'
#' @source AEMO NEMweb <http://nemweb.com.au>, published under
#'   the AEMO Copyright Permissions Notice.
#'
#' @family low-level
#' @export
#' @examples
#' \donttest{
#' op <- options(aemo.cache_dir = tempdir())
#' try({
#'   files <- aemo_nemweb_ls("/Reports/Current/DispatchIS_Reports/")
#'   head(files)
#' })
#' options(op)
#' }
aemo_nemweb_ls <- function(path) {
  path <- gsub("^/?", "/", path)
  path <- gsub("/?$", "/", path)
  url <- paste0(aemo_base_url(), path)

  resp <- aemo_request(url, timeout = 60) |>
    httr2::req_perform()
  if (httr2::resp_status(resp) != 200L) {
    cli::cli_abort("NEMweb returned HTTP {httr2::resp_status(resp)} for {.url {url}}.")
  }
  html <- httr2::resp_body_string(resp)

  # NEMweb uses IIS-style directory listings. Each entry is:
  #   <date> <time> AM/PM <size> <A HREF="/path/FILE.zip">FILE.zip</A><br>
  row_rx <- "(\\w+, \\w+ \\d+, \\d+\\s+\\d+:\\d+\\s+[AP]M)\\s+(\\d+)\\s+<A HREF=\"([^\"]+)\">([^<]+)</A>"
  m <- regmatches(html, gregexpr(row_rx, html, ignore.case = TRUE))[[1]]

  if (length(m) == 0L) {
    # Fallback: grab any file-like <A HREF=...> without date/size.
    # Drop the parent-directory navigation link.
    plain_rx <- "<A HREF=\"([^\"]+)\">([^<]+)</A>"
    m2 <- regmatches(html, gregexpr(plain_rx, html, ignore.case = TRUE))[[1]]
    if (length(m2) == 0L) {
      return(data.frame(name = character(0), modified = character(0),
                        size = character(0), url = character(0),
                        stringsAsFactors = FALSE))
    }
    parts <- regmatches(m2, regexec(plain_rx, m2, ignore.case = TRUE))
    hrefs <- vapply(parts, function(x) x[2], character(1))
    names_v <- vapply(parts, function(x) x[3], character(1))
    keep <- !grepl("parent directory", names_v, ignore.case = TRUE)
    hrefs <- hrefs[keep]; names_v <- names_v[keep]
    urls <- ifelse(grepl("^https?://", hrefs), hrefs,
                   paste0(aemo_base_url(), hrefs))
    return(data.frame(
      name = names_v,
      modified = NA_character_,
      size = NA_character_,
      url = urls,
      stringsAsFactors = FALSE
    ))
  }

  parts <- regmatches(m, regexec(row_rx, m, ignore.case = TRUE))
  mod_v <- vapply(parts, function(x) x[2], character(1))
  size_v <- vapply(parts, function(x) x[3], character(1))
  hrefs <- vapply(parts, function(x) x[4], character(1))
  names_v <- vapply(parts, function(x) x[5], character(1))
  urls <- ifelse(grepl("^https?://", hrefs), hrefs,
                 paste0(aemo_base_url(), hrefs))

  data.frame(
    name = names_v,
    modified = mod_v,
    size = size_v,
    url = urls,
    stringsAsFactors = FALSE
  )
}

#' Download a NEMweb zipped CSV to the local cache
#'
#' Thin wrapper over the internal cache-aware downloader.
#'
#' @param url A fully-qualified NEMweb URL (zipped CSV).
#' @param cache Logical. Reuse cached file if present.
#'
#' @return Path to the cached file.
#'
#' @family low-level
#' @export
#' @examples
#' \donttest{
#' op <- options(aemo.cache_dir = tempdir())
#' try({
#'   files <- aemo_nemweb_ls("/Reports/Current/DispatchIS_Reports/")
#'   if (nrow(files) > 0) {
#'     f <- aemo_nemweb_download(files$url[1])
#'     file.exists(f)
#'   }
#' })
#' options(op)
#' }
aemo_nemweb_download <- function(url, cache = TRUE) {
  aemo_download_cached(url, cache = cache)
}

# Marginal and distribution loss factors

# 5MS commenced 1 October 2021
FIVE_MS_DATE <- as.Date("2021-10-01")

#' Marginal Loss Factors (MLF) by DUID and financial year
#'
#' Returns the Marginal Loss Factor applicable to each DUID for
#' the requested financial year(s). MLFs measure the incremental
#' network loss at a connection point relative to the Regional
#' Reference Node (RRN); a DUID with MLF = 0.97 receives 97% of
#' the regional RRP per MWh generated.
#'
#' MLFs are published annually by AEMO under NER 3.6.2 and are
#' used in settlement calculations and in PPA revenue
#' reconstruction. The function first attempts to download the
#' `TRANSMISSIONLOSSFACTOR` table from the most recent MMSDM monthly
#' archive; if that fails it returns a bundled static table
#' covering FY 2020-21 to FY 2025-26 for ~20 well-known DUIDs.
#'
#' @param year Financial year(s) as `"YYYY-YY"` strings (e.g.
#'   `"2024-25"`). `NULL` returns all available years. Multiple
#'   years accepted.
#' @param duid Optional character vector of DUIDs to filter on.
#'   `NULL` returns all DUIDs. See [aemo_units()] for the full
#'   DUID registry.
#'
#' @return An `aemo_tbl` with columns `financial_year`, `duid`,
#'   `connectionpointid`, `regionid`, `mlf`. From a live MMSDM
#'   download additional columns (`participantid`,
#'   `lastchanged`) may also be present.
#'
#' @source AEMO MMSDM archive, TRANSMISSIONLOSSFACTOR table.
#'   AEMO Copyright Permissions Notice.
#'
#' @seealso [aemo_units()] for the DUID registry,
#'   [aemo_price()] for regional RRPs.
#'
#' @family reference
#' @export
#' @examples
#' \donttest{
#' op <- options(aemo.cache_dir = tempdir())
#' try({
#'   mlf <- aemo_mlf(year = "2024-25")
#'   head(mlf)
#' })
#' options(op)
#' }
aemo_mlf <- function(year = NULL, duid = NULL) {
  df <- tryCatch(aemo_fetch_mlf_mmsdm(),
                 error = function(e) NULL)

  if (is.null(df) || nrow(df) < 20L) {
    cli::cli_abort(c(
      "Could not retrieve TRANSMISSIONLOSSFACTOR from MMSDM.",
      "i" = "MLFs are published annually by AEMO under NER 3.6.2. For historical analysis, download the determination PDF directly from aemo.com.au, or retry when the MMSDM archive is reachable.",
      "i" = "No fallback is provided: unsourced MLF values would be a correctness trap in settlement and PPA workflows."
    ))
  }

  if (!is.null(year)) {
    if ("financial_year" %in% names(df)) {
      df <- df[df$financial_year %in% year, , drop = FALSE]
    }
    if (nrow(df) == 0L) {
      cli::cli_abort(c(
        "No MLF data found for year{?s}: {.val {year}}.",
        "i" = "Available format: {.val {\"2024-25\"}}."
      ))
    }
  }
  if (!is.null(duid) && "duid" %in% names(df)) {
    df <- df[toupper(df$duid) %in% toupper(duid), , drop = FALSE]
    rownames(df) <- NULL
  }

  new_aemo_tbl(df,
               source = "https://nemweb.com.au/Data_Archive/Wholesale_Electricity/MMSDM",
               licence = "AEMO Copyright Permissions Notice",
               title = "NEM Marginal Loss Factors (MLF)")
}

#' Distribution Loss Factors (DLF) by connection point
#'
#' Returns the Distribution Loss Factor for NEM connection
#' points from the MMSDM `LOSSFACTORMODEL` table. DLFs measure
#' the average energy loss on the distribution network between
#' a connection point and the transmission network boundary; a
#' DLF of 1.02 means the generator must produce 2% more energy
#' than it delivers to the transmission system.
#'
#' DLFs are published annually by AEMO (NER 3.6.3) and combined
#' with MLFs in settlement to give the total loss factor (TLF =
#' MLF x DLF) used in energy payments.
#'
#' @param year Financial year(s) as `"YYYY-YY"`. `NULL` returns
#'   all available years.
#' @param connection_point_id Optional character vector of
#'   connection point IDs.
#'
#' @return An `aemo_tbl` with at minimum `financial_year`,
#'   `connectionpointid`, and `dlf`.
#'
#' @source AEMO MMSDM archive, LOSSFACTORMODEL table.
#'   AEMO Copyright Permissions Notice.
#'
#' @family reference
#' @export
#' @examples
#' \donttest{
#' op <- options(aemo.cache_dir = tempdir())
#' try({
#'   dlf <- aemo_dlf(year = "2024-25")
#'   head(dlf)
#' })
#' options(op)
#' }
aemo_dlf <- function(year = NULL, connection_point_id = NULL) {
  df <- tryCatch(aemo_fetch_dlf_mmsdm(),
                 error = function(e) NULL)

  if (is.null(df) || nrow(df) < 5L) {
    cli::cli_warn(c(
      "Could not reach MMSDM LOSSFACTORMODEL; returning empty table.",
      "i" = "Use {.fn aemo_nemweb_download} with an MMSDM URL to fetch DLF data directly."
    ))
    df <- data.frame(
      financial_year = character(0L),
      connectionpointid = character(0L),
      dlf = numeric(0L),
      stringsAsFactors = FALSE
    )
  }

  if (!is.null(year) && "financial_year" %in% names(df)) {
    df <- df[df$financial_year %in% year, , drop = FALSE]
  }
  if (!is.null(connection_point_id) && "connectionpointid" %in% names(df)) {
    df <- df[toupper(df$connectionpointid) %in%
               toupper(connection_point_id), , drop = FALSE]
    rownames(df) <- NULL
  }

  new_aemo_tbl(df,
               source = "https://nemweb.com.au/Data_Archive/Wholesale_Electricity/MMSDM",
               licence = "AEMO Copyright Permissions Notice",
               title = "NEM Distribution Loss Factors (DLF)")
}

#' Fetch TRANSMISSIONLOSSFACTOR (MLF) from MMSDM archive.
#'
#' AEMO's MMSDM uses TRANSMISSIONLOSSFACTOR for marginal loss
#' factors, not TRANSMISSIONLOSSFACTOR (which does not exist in the
#' archive). The column is `transmissionlossfactor`.
#' @noRd
aemo_fetch_mlf_mmsdm <- function(max_months_back = 6L) {
  df <- aemo_fetch_mmsdm_table("TRANSMISSIONLOSSFACTOR", min_rows = 1L,
                                max_months_back = max_months_back)
  if (!is.null(df) && "effectivedate" %in% names(df)) {
    df$financial_year <- aemo_fy_label(df$effectivedate)
  }
  if (!is.null(df)) {
    names(df)[names(df) == "transmissionlossfactor"] <- "mlf"
  }
  df
}

#' Fetch LOSSFACTORMODEL (DLF) from MMSDM archive.
#' @noRd
aemo_fetch_dlf_mmsdm <- function(max_months_back = 6L) {
  df <- aemo_fetch_mmsdm_table("LOSSFACTORMODEL", min_rows = 1L,
                                max_months_back = max_months_back)
  if (is.null(df)) return(NULL)
  if ("effectivedate" %in% names(df)) {
    df$financial_year <- aemo_fy_label(df$effectivedate)
  }
  names(df)[names(df) == "distributionlossfactor"] <- "dlf"
  df
}

#' Convert a date (or POSIXct) to a NEM financial-year label.
#'
#' The NEM financial year runs 1 July to 30 June. A date in
#' January 2025 is FY "2024-25"; a date in August 2025 is
#' FY "2025-26".
#' @noRd
aemo_fy_label <- function(x) {
  d <- if (inherits(x, "POSIXct")) as.Date(x, tz = AEMO_TIMEZONE) else as.Date(x)
  yr <- as.integer(format(d, "%Y"))
  mo <- as.integer(format(d, "%m"))
  start_yr <- ifelse(mo >= 7L, yr, yr - 1L)
  end_yy <- sprintf("%02d", (start_yr + 1L) %% 100L)
  paste0(start_yr, "-", end_yy)
}

test_that("aemo_fy_label converts dates to FY strings correctly", {
  # January 2025 -> FY 2024-25
  expect_equal(aemo:::aemo_fy_label(as.Date("2025-01-15")), "2024-25")
  # August 2025 -> FY 2025-26
  expect_equal(aemo:::aemo_fy_label(as.Date("2025-08-01")), "2025-26")
  # July 1 boundary (first day of new FY)
  expect_equal(aemo:::aemo_fy_label(as.Date("2024-07-01")), "2024-25")
  # June 30 boundary (last day of old FY)
  expect_equal(aemo:::aemo_fy_label(as.Date("2025-06-30")), "2024-25")
  # POSIXct input
  p <- as.POSIXct("2025-03-01 14:00:00", tz = "Australia/Brisbane")
  expect_equal(aemo:::aemo_fy_label(p), "2024-25")
})

test_that("aemo_mlf aborts when MMSDM is unreachable (no fallback)", {
  # v0.2.0 removed the indicative MLF fallback: unsourced values are a
  # correctness trap. MMSDM being unreachable must abort rather than
  # return invented numbers. This test is skipped when MMSDM is live.
  testthat::skip_on_cran()
  testthat::skip_if_not(
    identical(Sys.getenv("AEMO_FORCE_OFFLINE_TESTS"), "true"),
    "Offline-abort test runs only when AEMO_FORCE_OFFLINE_TESTS=true"
  )
  withr::local_options(aemo.cache_dir = tempdir(),
                       aemo.base_url = "http://127.0.0.1:1")
  expect_error(aemo_mlf(year = "2024-25"),
               "Could not retrieve TRANSMISSIONLOSSFACTOR")
})

test_that("aemo_mlf filters by year and duid when MMSDM reachable", {
  testthat::skip_on_cran()
  skip_if_not(
    isTRUE(tryCatch({
      httr2::request("http://nemweb.com.au") |>
        httr2::req_timeout(5) |>
        httr2::req_perform()
      TRUE
    }, error = function(e) FALSE)),
    message = "NEMweb not reachable"
  )

  withr::local_options(aemo.cache_dir = tempdir())
  m <- tryCatch(aemo_mlf(year = "2024-25"), error = function(e) NULL)
  skip_if(is.null(m), "TRANSMISSIONLOSSFACTOR not in recent MMSDM archive")

  expect_s3_class(m, "aemo_tbl")
  expect_true(nrow(m) > 0L)
  if ("financial_year" %in% names(m)) {
    expect_true(all(m$financial_year == "2024-25"))
  }
})

# AEMO CSV row format parser
#
# AEMO NEMweb files are zipped CSVs with row-type prefixes:
#   C,<comment fields>             -- file comment (one at top, one at bottom)
#   I,<report>,<table>,<ver>,<col1>,<col2>,...  -- column header
#   D,<report>,<table>,<ver>,<val1>,<val2>,...  -- data row
#   F,<report>,<table>,<ver>,...   -- file footer with row count
#
# A single file may contain multiple (report, table, ver) tuples
# with independent I/D row streams. We dispatch by reading the I
# row to set the schema, then accumulate subsequent D rows.

#' Parse an AEMO CSV file into one data frame per table
#'
#' @param path Path to an unzipped AEMO CSV file.
#' @return A named list of data frames, keyed by
#'   `<REPORT>_<TABLE>`.
#' @noRd
aemo_parse_csv <- function(path) {
  con <- file(path, open = "r")
  on.exit(close(con), add = TRUE)

  schemas <- list()
  streams <- list()
  current <- NULL

  while (length(line <- readLines(con, n = 1L, warn = FALSE)) > 0L) {
    tag <- substr(line, 1L, 1L)
    if (tag == "C" || tag == "F") next
    parts <- scan(textConnection(line), what = "character",
                  sep = ",", quiet = TRUE, strip.white = TRUE,
                  na.strings = "", quote = "\"")
    if (length(parts) < 4L) next
    key <- paste(parts[2], parts[3], sep = "_")

    if (tag == "I") {
      cols <- parts[5:length(parts)]
      schemas[[key]] <- cols
      streams[[key]] <- list()
      current <- key
    } else if (tag == "D") {
      vals <- parts[5:length(parts)]
      if (is.null(schemas[[key]])) next
      # Pad vals to match schema length
      n <- length(schemas[[key]])
      vals <- c(vals, rep(NA_character_, max(0L, n - length(vals))))[seq_len(n)]
      streams[[key]][[length(streams[[key]]) + 1L]] <- vals
    }
  }

  out <- list()
  for (key in names(streams)) {
    rows <- streams[[key]]
    if (length(rows) == 0L) next
    mat <- do.call(rbind, rows)
    df <- as.data.frame(mat, stringsAsFactors = FALSE)
    names(df) <- aemo_clean_names(schemas[[key]])
    out[[tolower(key)]] <- df
  }
  out
}

#' Download, unzip, and parse an AEMO zipped CSV into tables
#'
#' @param url URL to an AEMO zipped CSV.
#' @return Named list of data frames keyed by `<report>_<table>`.
#' @noRd
aemo_fetch_zip <- function(url) {
  zip_path <- aemo_download_cached(url)
  tmp <- tempfile("aemo_unzip_")
  dir.create(tmp, recursive = TRUE)
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
  utils::unzip(zip_path, exdir = tmp)

  csvs <- list.files(tmp, pattern = "\\.[Cc][Ss][Vv]$", full.names = TRUE)
  if (length(csvs) == 0L) {
    cli::cli_abort("No CSV files found inside {.file {basename(zip_path)}}.")
  }
  combined <- list()
  for (p in csvs) {
    parsed <- aemo_parse_csv(p)
    for (k in names(parsed)) {
      if (is.null(combined[[k]])) {
        combined[[k]] <- parsed[[k]]
      } else {
        # Concatenate with matching columns
        common <- intersect(names(combined[[k]]), names(parsed[[k]]))
        combined[[k]] <- rbind(combined[[k]][, common, drop = FALSE],
                               parsed[[k]][, common, drop = FALSE])
      }
    }
  }
  combined
}

test_that("aemo_parse_csv parses a minimal DISPATCH_PRICE block", {
  tmp <- tempfile(fileext = ".csv")
  writeLines(c(
    'C,NEMP.WORLD,...,test,...',
    'I,DISPATCH,PRICE,1,SETTLEMENTDATE,REGIONID,RRP',
    'D,DISPATCH,PRICE,1,"2024/06/01 00:05:00",NSW1,100.50',
    'D,DISPATCH,PRICE,1,"2024/06/01 00:10:00",NSW1,102.75',
    'D,DISPATCH,PRICE,1,"2024/06/01 00:05:00",VIC1,95.00',
    'F,DISPATCH,PRICE,1,3'
  ), tmp)
  on.exit(unlink(tmp), add = TRUE)

  parsed <- aemo:::aemo_parse_csv(tmp)
  expect_true("dispatch_price" %in% names(parsed))
  df <- parsed$dispatch_price
  expect_equal(nrow(df), 3L)
  expect_true(all(c("settlementdate", "regionid", "rrp") %in% names(df)))
})

test_that("aemo_parse_csv handles multiple tables in one file", {
  tmp <- tempfile(fileext = ".csv")
  writeLines(c(
    'C,header',
    'I,DISPATCH,PRICE,1,COL_A,COL_B',
    'D,DISPATCH,PRICE,1,1,2',
    'I,DISPATCH,DEMAND,1,COL_C,COL_D',
    'D,DISPATCH,DEMAND,1,3,4',
    'F,footer'
  ), tmp)
  on.exit(unlink(tmp), add = TRUE)

  parsed <- aemo:::aemo_parse_csv(tmp)
  expect_setequal(names(parsed), c("dispatch_price", "dispatch_demand"))
  expect_equal(nrow(parsed$dispatch_price), 1L)
  expect_equal(nrow(parsed$dispatch_demand), 1L)
})

test_that("aemo_parse_csv tolerates ragged D rows", {
  tmp <- tempfile(fileext = ".csv")
  writeLines(c(
    'C,header',
    'I,DISPATCH,PRICE,1,A,B,C',
    'D,DISPATCH,PRICE,1,1,2,3',
    'D,DISPATCH,PRICE,1,4,5',
    'F,footer'
  ), tmp)
  on.exit(unlink(tmp), add = TRUE)
  parsed <- aemo:::aemo_parse_csv(tmp)
  expect_equal(nrow(parsed$dispatch_price), 2L)
  expect_equal(parsed$dispatch_price$c, c("3", NA_character_))
})

#' aemo: Australian Energy Market Operator Data
#'
#' Tidy R access to AEMO's public data: wholesale electricity
#' prices, demand, dispatch, interconnector flows, rooftop PV,
#' bids, predispatch and PASA forecasts, FCAS markets, and gas
#' markets across the NEM.
#'
#' @section Main function families:
#' - [aemo_price()], [aemo_demand()]: workhorse price and demand.
#' - [aemo_dispatch_units()], [aemo_interconnector()],
#'   [aemo_rooftop_pv()]: dispatch and flow.
#' - [aemo_bids()], [aemo_predispatch()], [aemo_pasa()]: bids
#'   and forecasts.
#' - [aemo_fcas()], [aemo_gas()]: ancillary services and gas.
#' - [aemo_units()], [aemo_regions()], [aemo_interconnectors()]:
#'   reference data.
#' - [aemo_nemweb_ls()], [aemo_nemweb_download()]: low-level
#'   escape hatches for arbitrary NEMweb paths.
#' - [aemo_cache_info()], [aemo_clear_cache()], [aemo_throttle()]:
#'   configuration.
#'
#' @section Data source:
#' Data is published by AEMO at <http://nemweb.com.au> under the
#' AEMO Copyright Permissions Notice
#' <https://www.aemo.com.au/privacy-and-legal-notices/copyright-permissions>.
#'
#' @keywords internal
#' @concept AEMO
#' @concept NEM
#' @concept electricity market
#' @concept wholesale electricity
"_PACKAGE"

utils::globalVariables(c("aemo_regions_df", "aemo_interconnectors_df"))

# MARKETNOTICEDATA feed

#' NEM market notices
#'
#' Returns the `MARKETNOTICEDATA` feed from MMSDM. Market
#' notices are AEMO's free-text log of market-relevant events:
#' Lack of Reserve (LOR1/2/3) declarations, Reliability and
#' Emergency Reserve Trader (RERT) activations, market
#' suspensions, market directions, unit withdrawals, system
#' security events, administered price declarations, and
#' operator advisories.
#'
#' Pair with [aemo_constraints()] and [aemo_price()] to sequence
#' the causal chain of a price event. Rangarajan, Svec, Foley
#' and Trück (2025, *Energy Economics* 141) use
#' `MARKETNOTICEDATA` to order the intervention messages that
#' mark entry to and exit from the June 2022 NEM suspension.
#'
#' @param start,end Window (inclusive) applied to `EFFECTIVEDATE`.
#' @param notice_type Optional character vector (e.g. `"LOR1"`,
#'   `"RERT"`, `"PRICES SUBJECT TO REVIEW"`). Case-insensitive
#'   substring match.
#' @param region Optional NEM region code.
#'
#' @return An `aemo_tbl` with columns from MARKETNOTICEDATA:
#'   `noticeid`, `effectivedate`, `typeid` (notice category),
#'   `originator`, `priority`, `reason` (the notice text),
#'   `externalreference`, and where present `regionid`.
#'
#' @source AEMO NEMweb MMSDM archive, MARKETNOTICEDATA table.
#'   AEMO Copyright Permissions Notice.
#'
#' @family dispatch
#' @export
#' @examples
#' \donttest{
#' op <- options(aemo.cache_dir = tempdir())
#' try({
#'   # LOR declarations during the June 2022 NEM suspension
#'   n <- aemo_market_notices(
#'     start = "2022-06-13",
#'     end   = "2022-06-14",
#'     notice_type = "LOR"
#'   )
#'   head(n)
#' })
#' options(op)
#' }
aemo_market_notices <- function(start, end,
                                 notice_type = NULL,
                                 region = NULL) {
  start <- aemo_parse_time(start)
  end <- aemo_parse_time(end)
  stopifnot(end >= start)
  if (!is.null(region)) region <- aemo_validate_region(region)

  months_needed <- unique(format(
    seq(as.Date(start, tz = AEMO_TIMEZONE),
        as.Date(end,   tz = AEMO_TIMEZONE),
        by = "month"),
    "%Y-%m"
  ))

  all_parts <- list()
  for (ym in months_needed) {
    y <- substr(ym, 1L, 4L)
    m <- substr(ym, 6L, 7L)
    url <- aemo_mmsdm_url("MARKETNOTICEDATA", y, m)
    df <- tryCatch({
      zip_path <- aemo_download_cached(url)
      tmp <- tempfile("aemo_mn_")
      dir.create(tmp, recursive = TRUE)
      on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
      utils::unzip(zip_path, exdir = tmp)
      csvs <- list.files(tmp, pattern = "\\.[Cc][Ss][Vv]$", full.names = TRUE)
      if (length(csvs) == 0L) stop("no csv")
      parsed <- aemo_parse_csv(csvs[[1L]])
      parsed[[1L]]
    }, error = function(e) NULL)
    if (!is.null(df) && nrow(df) > 0L) {
      all_parts[[length(all_parts) + 1L]] <- df
    }
  }

  if (length(all_parts) == 0L) {
    cli::cli_abort(c(
      "Could not retrieve MARKETNOTICEDATA from MMSDM.",
      "i" = "Requested months: {.val {months_needed}}. Check your connection or try {.fn aemo_nemweb_download} with an MMSDM URL."
    ))
  }
  common <- Reduce(intersect, lapply(all_parts, names))
  df <- do.call(rbind, lapply(all_parts, function(d) d[, common, drop = FALSE]))
  df <- aemo_coerce_types(df)

  # Window filter on effectivedate (the canonical timestamp on this table).
  ts_col <- intersect(c("effectivedate", "lastchanged"), names(df))[1L]
  if (!is.na(ts_col)) {
    ts <- if (inherits(df[[ts_col]], "POSIXct")) df[[ts_col]] else
      aemo_parse_col_time(df[[ts_col]])
    keep <- !is.na(ts) & ts >= start & ts <= end
    df <- df[keep, , drop = FALSE]
  }

  if (!is.null(notice_type)) {
    type_cols <- intersect(c("typeid", "noticetype", "reason"), names(df))
    if (length(type_cols) > 0L) {
      haystack <- do.call(paste, c(lapply(type_cols, function(c) df[[c]]),
                                    sep = " "))
      needle <- paste(toupper(notice_type), collapse = "|")
      keep <- grepl(needle, toupper(haystack))
      df <- df[keep, , drop = FALSE]
    }
  }
  if (!is.null(region) && "regionid" %in% names(df)) {
    df <- df[df$regionid %in% region, , drop = FALSE]
  }
  rownames(df) <- NULL

  new_aemo_tbl(df,
               source = "https://nemweb.com.au/Data_Archive/Wholesale_Electricity/MMSDM",
               licence = "AEMO Copyright Permissions Notice",
               title = "NEM market notices (MARKETNOTICEDATA)")
}

# Reference tables (static + live fetch)

#' NEM regions
#'
#' Returns a static table of the five NEM regions with metadata.
#' The `market_timezone` column is the AEMO market clock (AEST,
#' UTC+10, no DST, year-round) that applies to every timestamp
#' in NEMweb files; `wall_timezone` is the local civil time zone
#' consumers experience (observes DST in NSW/VIC/TAS/SA).
#'
#' @return An `aemo_tbl` with columns `region`, `name`, `state`,
#'   `wall_timezone`, `market_timezone`, `commenced`.
#'
#' @family reference
#' @export
#' @examples
#' aemo_regions()
aemo_regions <- function() {
  df <- data.frame(
    region = c("NSW1", "QLD1", "SA1", "TAS1", "VIC1"),
    name = c("New South Wales", "Queensland", "South Australia",
             "Tasmania", "Victoria"),
    state = c("NSW", "QLD", "SA", "TAS", "VIC"),
    wall_timezone = c("Australia/Sydney", "Australia/Brisbane",
                      "Australia/Adelaide", "Australia/Hobart",
                      "Australia/Melbourne"),
    market_timezone = rep(AEMO_TIMEZONE, 5L),
    commenced = as.Date(c("1998-12-13", "1998-12-13", "1998-12-13",
                           "2005-05-29", "1998-12-13")),
    stringsAsFactors = FALSE
  )
  new_aemo_tbl(df,
               source = "http://nemweb.com.au",
               title = "NEM regions")
}

#' NEM interconnectors
#'
#' Returns a static table of the seven NEM interconnectors. The
#' seventh entry is Project EnergyConnect (PEC), whose Stage 1
#' was energised on 30 April 2025 with full ~800 MW capability
#' expected mid-2026.
#'
#' @return An `aemo_tbl` with columns `interconnector_id`,
#'   `from_region`, `to_region`, `name`, `energised`.
#'
#' @family reference
#' @export
#' @examples
#' aemo_interconnectors()
aemo_interconnectors <- function() {
  df <- data.frame(
    interconnector_id = c("NSW1-QLD1", "VIC1-NSW1", "V-S-MNSP1",
                          "V-SA", "T-V-MNSP1", "N-Q-MNSP1",
                          "V-S-N"),
    from_region = c("NSW1", "VIC1", "VIC1", "VIC1", "TAS1", "NSW1",
                    "NSW1"),
    to_region = c("QLD1", "NSW1", "SA1", "SA1", "VIC1", "QLD1",
                  "SA1"),
    name = c("NSW-QLD Interconnector (QNI)",
             "VIC-NSW Interconnector (VNI)",
             "Murraylink",
             "Heywood Interconnector",
             "Basslink",
             "Terranora (Directlink)",
             "Project EnergyConnect (PEC)"),
    energised = as.Date(c("2001-02-18", "1989-01-01", "2002-10-03",
                          "1990-11-01", "2006-04-29", "2000-12-14",
                          "2025-04-30")),
    stringsAsFactors = FALSE
  )
  new_aemo_tbl(df,
               source = "http://nemweb.com.au",
               title = "NEM interconnectors")
}

#' NEM generators (DUID registry)
#'
#' Downloads the `DUDETAILSUMMARY` table from the most recent
#' MMSDM monthly archive and returns one row per registered DUID
#' (Dispatchable Unit Identifier) with station, region, dispatch
#' type, classification, and schedule type. Typical output is
#' 500+ DUIDs covering scheduled, semi-scheduled, and
#' non-scheduled generators, bidirectional storage (BESS), and
#' loads.
#'
#' **Effective-date filtering.** `DUDETAILSUMMARY` is
#' effective-dated (`START_DATE`, `END_DATE` per row with
#' multiple `VERSIONNO` vintages per DUID over time). The
#' default (`as_of = NULL`) returns rows whose
#' `[START_DATE, END_DATE]` interval covers the date the MMSDM
#' archive was published: i.e. the *current* registry. Pass an
#' `as_of` date to get the registry as it was on that date
#' (essential for historical analysis: Liddell's four DUIDs
#' were retired in April 2023; pre-2023 queries need them).
#'
#' This matches the as-of-join pattern documented in Gorman et
#' al. (2018) *NEMOSIS* (APSRC) for correct historical joins.
#'
#' @param as_of Optional `Date` or `POSIXct`. Returns the DUID
#'   registry as it was on that date. `NULL` (default) returns
#'   the current registry.
#'
#' @return An `aemo_tbl` keyed by `duid`. Columns include (from
#'   DUDETAILSUMMARY): `duid`, `stationid`, `regionid`,
#'   `dispatchtype` (GENERATOR / LOAD), `connectionpointid`,
#'   `schedule_type` (SCHEDULED / SEMI-SCHEDULED /
#'   NON-SCHEDULED), `start_date`, `end_date`.
#'
#' @source AEMO NEMweb MMSDM archive, DUDETAILSUMMARY table,
#'   AEMO Copyright Permissions Notice.
#'
#' @family reference
#' @export
#' @examples
#' \donttest{
#' op <- options(aemo.cache_dir = tempdir())
#' try({
#'   # Current DUID registry
#'   u <- aemo_units()
#'
#'   # DUIDs as they were on 1 March 2023 (pre-Liddell retirement)
#'   u_2023 <- aemo_units(as_of = "2023-03-01")
#' })
#' options(op)
#' }
aemo_units <- function(as_of = NULL) {
  df <- tryCatch(aemo_fetch_dudetailsummary(),
                 error = function(e) NULL)
  if (is.null(df) || nrow(df) < 50L) {
    cli::cli_abort(c(
      "Could not retrieve DUDETAILSUMMARY from MMSDM.",
      "i" = "Retry when the MMSDM archive is reachable, or use {.fn aemo_nemweb_download} with an MMSDM URL directly.",
      "i" = "No fallback is provided: a stale or invented DUID registry would silently mislabel historical analyses."
    ))
  }

  if (!is.null(as_of)) {
    as_of_ts <- aemo_parse_time(as_of)
    df <- aemo_filter_as_of(df, as_of = as_of_ts)
  }

  new_aemo_tbl(df,
               source = "http://nemweb.com.au",
               title = paste0("NEM DUID registry (DUDETAILSUMMARY",
                              if (!is.null(as_of))
                                paste0(", as of ", format(aemo_parse_time(as_of), "%Y-%m-%d"))
                              else paste0(", ", format(Sys.Date(), "%Y-%m")),
                              ")"))
}

#' Apply an effective-date filter on a DUDETAILSUMMARY-style
#' data frame.
#'
#' Selects rows where `start_date <= as_of < end_date`. Uses
#' MMSDM's end-of-time sentinel (2999-12-31) for open-ended
#' records.
#' @noRd
aemo_filter_as_of <- function(df, as_of) {
  start_col <- intersect(c("start_date", "effectivedate",
                            "effectivefrom"), names(df))[1L]
  end_col <- intersect(c("end_date", "enddate", "effectiveto"),
                        names(df))[1L]
  if (is.na(start_col)) return(df)

  start_ts <- if (inherits(df[[start_col]], "POSIXct")) {
    df[[start_col]]
  } else {
    aemo_parse_col_time(df[[start_col]])
  }
  end_ts <- if (!is.na(end_col)) {
    if (inherits(df[[end_col]], "POSIXct")) {
      df[[end_col]]
    } else {
      aemo_parse_col_time(df[[end_col]])
    }
  } else {
    rep(as.POSIXct(NA, tz = AEMO_TIMEZONE), nrow(df))
  }
  # Treat NA or MMSDM sentinel (year 2999) end-dates as open-ended.
  open_end <- is.na(end_ts) |
    format(end_ts, "%Y") %in% c("2999", "9999")
  keep <- !is.na(start_ts) & start_ts <= as_of &
    (open_end | end_ts > as_of)
  df <- df[keep, , drop = FALSE]
  rownames(df) <- NULL
  df
}

#' Fetch DUDETAILSUMMARY from the most recent MMSDM archive.
#' @noRd
aemo_fetch_dudetailsummary <- function(max_months_back = 6L) {
  aemo_fetch_mmsdm_table("DUDETAILSUMMARY", min_rows = 50L,
                          max_months_back = max_months_back)
}

#' NEM market participants and their registered DUIDs
#'
#' Returns a mapping of NEM market participants (companies) to
#' their registered Dispatchable Unit Identifiers (DUIDs), joined
#' from the MMSDM `PARTICIPANT` and `DUDETAILSUMMARY` tables.
#' Use this for corporate ownership analysis: rolling up generator
#' output or bids from DUID-level data to the company level.
#'
#' @return An `aemo_tbl` with columns `participantid`,
#'   `participantclassid` (e.g. `GENERATOR`, `LOAD`, `TRADER`),
#'   `name` (company name), `duid`, `stationid`, `regionid`,
#'   `dispatchtype`, `schedule_type`. Rows are one per
#'   participant-DUID combination. If MMSDM is unreachable,
#'   returns an empty table with a warning.
#'
#' @source AEMO NEMweb MMSDM archive, PARTICIPANT and
#'   DUDETAILSUMMARY tables. AEMO Copyright Permissions Notice.
#'
#' @seealso [aemo_units()] for DUID-level registry without
#'   participant mapping.
#'
#' @family reference
#' @export
#' @examples
#' \donttest{
#' op <- options(aemo.cache_dir = tempdir())
#' try({
#'   pt <- aemo_participants()
#'   # DUIDs owned by AGL
#'   pt[grepl("AGL", pt$name, ignore.case = TRUE), ]
#' })
#' options(op)
#' }
aemo_participants <- function() {
  duids <- tryCatch(aemo_fetch_dudetailsummary(),
                    error = function(e) NULL)
  parts <- tryCatch(aemo_fetch_participant(),
                    error = function(e) NULL)

  if (is.null(duids) || nrow(duids) < 50L) {
    cli::cli_warn(c(
      "Could not reach MMSDM DUDETAILSUMMARY; returning empty participant table.",
      "i" = "Check your internet connection."
    ))
    empty <- data.frame(
      participantid = character(0L), participantclassid = character(0L),
      name = character(0L), duid = character(0L),
      stationid = character(0L), regionid = character(0L),
      dispatchtype = character(0L), schedule_type = character(0L),
      stringsAsFactors = FALSE
    )
    return(new_aemo_tbl(empty,
                        source = "https://nemweb.com.au",
                        title = "NEM market participants (empty - MMSDM unreachable)"))
  }

  # Join participant metadata onto DUID registry
  if (!is.null(parts) && nrow(parts) >= 1L &&
      "participantid" %in% names(parts) &&
      "participantid" %in% names(duids)) {
    # Keep only columns we want to expose from each table
    part_cols <- intersect(c("participantid", "participantclassid", "name"),
                           names(parts))
    duid_cols <- intersect(c("participantid", "duid", "stationid",
                              "regionid", "dispatchtype", "schedule_type",
                              "start_date", "end_date"),
                            names(duids))
    df <- merge(
      parts[, part_cols, drop = FALSE],
      duids[, duid_cols, drop = FALSE],
      by = "participantid", all.y = TRUE
    )
  } else {
    # No participant table; return DUID registry with participantid only
    duid_cols <- intersect(c("participantid", "duid", "stationid",
                              "regionid", "dispatchtype", "schedule_type"),
                            names(duids))
    df <- duids[, duid_cols, drop = FALSE]
    if (!"name" %in% names(df)) df$name <- NA_character_
  }
  rownames(df) <- NULL

  new_aemo_tbl(df,
               source = "https://nemweb.com.au/Data_Archive/Wholesale_Electricity/MMSDM",
               licence = "AEMO Copyright Permissions Notice",
               title = paste0("NEM participants + DUIDs (MMSDM, ",
                              format(Sys.Date(), "%Y-%m"), ")"))
}

#' Fetch PARTICIPANT table from the most recent MMSDM archive.
#' @noRd
aemo_fetch_participant <- function(max_months_back = 6L) {
  aemo_fetch_mmsdm_table("PARTICIPANT", min_rows = 1L,
                          max_months_back = max_months_back)
}

test_that("aemo_filter_as_of selects rows active on as_of date", {
  df <- data.frame(
    duid = c("BW01", "LD01", "HPS1", "BESS_NEW"),
    start_date = as.POSIXct(c("2010-01-01", "1990-01-01", "2005-01-01",
                               "2024-06-01"),
                              tz = "Australia/Brisbane"),
    end_date = as.POSIXct(c("2999-12-31", "2023-04-28", "2999-12-31",
                              "2999-12-31"),
                            tz = "Australia/Brisbane"),
    stringsAsFactors = FALSE
  )

  # Before Liddell retirement: Liddell in, new BESS not yet
  out <- aemo:::aemo_filter_as_of(df,
                                   as_of = as.POSIXct("2022-03-01",
                                                        tz = "Australia/Brisbane"))
  expect_true("LD01" %in% out$duid)
  expect_false("BESS_NEW" %in% out$duid)

  # Mid-2024: Liddell out, new BESS in
  out2 <- aemo:::aemo_filter_as_of(df,
                                    as_of = as.POSIXct("2024-08-01",
                                                         tz = "Australia/Brisbane"))
  expect_false("LD01" %in% out2$duid)
  expect_true("BESS_NEW" %in% out2$duid)
  expect_true("BW01" %in% out2$duid)
})

test_that("aemo_filter_as_of treats MMSDM 2999 sentinel as open-ended", {
  df <- data.frame(
    duid = "X",
    start_date = as.POSIXct("2020-01-01", tz = "Australia/Brisbane"),
    end_date = as.POSIXct("2999-12-31", tz = "Australia/Brisbane"),
    stringsAsFactors = FALSE
  )
  out <- aemo:::aemo_filter_as_of(df,
                                   as_of = as.POSIXct("2024-06-01",
                                                        tz = "Australia/Brisbane"))
  expect_equal(nrow(out), 1L)
})

test_that("aemo_regions returns all five NEM regions", {
  r <- aemo_regions()
  expect_s3_class(r, "aemo_tbl")
  expect_equal(nrow(r), 5L)
  expect_setequal(r$region, c("NSW1", "QLD1", "SA1", "TAS1", "VIC1"))
})

test_that("aemo_regions distinguishes market and wall timezones", {
  r <- aemo_regions()
  # Market timezone is fixed AEST (Australia/Brisbane, no DST)
  # for every region.
  expect_true(all(r$market_timezone == "Australia/Brisbane"))
  # Wall timezones differ by region (NSW/VIC/TAS/SA observe DST).
  expect_gt(length(unique(r$wall_timezone)), 1L)
})

test_that("aemo_interconnectors returns seven interconnectors", {
  i <- aemo_interconnectors()
  expect_s3_class(i, "aemo_tbl")
  expect_equal(nrow(i), 7L)
  expect_true(all(c("interconnector_id", "from_region", "to_region") %in% names(i)))
  expect_true("V-S-N" %in% i$interconnector_id)  # Project EnergyConnect
})

test_that("aemo_units returns the current DUID registry (live)", {
  skip_on_cran()
  skip_if_offline()
  skip_if(Sys.getenv("AEMO_LIVE_TESTS") != "true",
          "Live network tests disabled by default.")

  withr::local_options(aemo.cache_dir = tempdir())
  u <- aemo_units()
  expect_s3_class(u, "aemo_tbl")
  expect_true("duid" %in% names(u))
  expect_gt(nrow(u), 0L)
})

test_that("aemo_units aborts when MMSDM is unreachable (no fallback)", {
  # aemo_units() has no fallback registry by design: a stale or invented
  # DUID table would silently mislabel historical analyses, so an
  # unreachable MMSDM archive must abort rather than return guesses.
  skip_on_cran()
  skip_if_not(
    identical(Sys.getenv("AEMO_FORCE_OFFLINE_TESTS"), "true"),
    "Offline-abort test runs only when AEMO_FORCE_OFFLINE_TESTS=true"
  )
  withr::local_options(aemo.cache_dir = tempdir(),
                       aemo.mmsdm_base = "http://127.0.0.1:1")
  expect_error(aemo_units(), "Could not retrieve DUDETAILSUMMARY")
})

# Predispatch and PASA forecasts

#' Price and demand forecasts (P5MIN, PREDISPATCH)
#'
#' Returns AEMO's forecast prices and demand for a NEM region.
#' Two horizons:
#' - `"p5min"`: 5-minute-ahead forecast, 12 intervals ahead,
#'   published every 5 minutes.
#' - `"predispatch"`: 40-hour-ahead predispatch at 30-minute
#'   resolution, published every 30 minutes.
#'
#' The 7-day predispatch publication was retired when 5-minute
#' settlement commenced; for longer horizons use [aemo_pasa()].
#'
#' **Vintages.** PREDISPATCH and P5MIN forecasts are re-issued
#' every 30 or 5 minutes respectively. Every vintage is archived
#' by its `RUN_DATETIME`. By default (`run_datetime = NULL`) this
#' function returns all vintages whose *file* timestamp falls in
#' `[start, end]`, i.e. all the forecasts *issued* during the
#' window. The `periodid` / `datetime` columns on the returned
#' rows give each forecast's *target* time.
#'
#' For forecast-error research (comparing a forecast vintage
#' against the realised dispatch) pass `run_datetime` to pin a
#' specific vintage:
#'
#' ```r
#' # The PREDISPATCH run issued at 15:00 on 1 June 2024 --
#' # 80 half-hour-ahead rows covering 15:30 out to 31:30 hours.
#' v <- aemo_predispatch("NSW1", start = "2024-06-01", end = "2024-06-02",
#'                       run_datetime = "2024-06-01 15:00")
#' ```
#'
#' This is the vintage-aware pattern used in Prakash (2023)
#' *NEMSEER* (JOSS 8(92) 5883).
#'
#' @param region NEM region code.
#' @param start,end Window of forecast run-times.
#' @param horizon One of `"predispatch"` (default) or `"p5min"`.
#' @param run_datetime Optional character or `POSIXct`. A specific
#'   `RUN_DATETIME` to pin to. `NULL` (default) returns every
#'   vintage issued in `[start, end]`.
#'
#' @return An `aemo_tbl`.
#'
#' @family forecast
#' @export
#' @examples
#' \donttest{
#' op <- options(aemo.cache_dir = tempdir())
#' try({
#'   now <- Sys.time()
#'   p <- aemo_predispatch("NSW1", start = now - 3600, end = now)
#'   head(p)
#' })
#' options(op)
#' }
aemo_predispatch <- function(region, start, end,
                              horizon = c("predispatch", "p5min"),
                              run_datetime = NULL) {
  region <- aemo_validate_region(region)
  horizon <- match.arg(horizon)
  start <- aemo_parse_time(start)
  end <- aemo_parse_time(end)
  stopifnot(end >= start)

  if (horizon == "predispatch") {
    current <- "/Reports/Current/PredispatchIS_Reports/"
    archive <- "/Reports/Archive/PredispatchIS_Reports/"
    pattern <- "PREDISPATCHIS"
  } else {
    current <- "/Reports/Current/P5_Reports/"
    archive <- "/Reports/Archive/P5_Reports/"
    pattern <- "P5MIN"
  }

  df <- aemo_fetch_report_range(
    current_dir = current, archive_dir = archive,
    pattern = pattern, start = start, end = end
  )
  df <- aemo_coerce_types(df)
  # Do NOT filter on settlementdate here: forecast rows have target
  # datetimes that extend well past the [start, end] run-time window.
  df <- aemo_apply_filters(df, region = region)

  if (!is.null(run_datetime)) {
    run_ts <- aemo_parse_time(run_datetime)
    run_col <- intersect(c("run_datetime", "predispatchseqno",
                            "predispatch_run_datetime"), names(df))[1L]
    if (!is.na(run_col) && inherits(df[[run_col]], "POSIXct")) {
      df <- df[!is.na(df[[run_col]]) & df[[run_col]] == run_ts, ,
               drop = FALSE]
    } else if (!is.na(run_col)) {
      # Character match on a pre-coerced run datetime column
      target <- format(run_ts, "%Y/%m/%d %H:%M:%S")
      df <- df[!is.na(df[[run_col]]) & df[[run_col]] == target, ,
               drop = FALSE]
    }
    rownames(df) <- NULL
  }

  new_aemo_tbl(df,
               source = "http://nemweb.com.au",
               title = paste0("AEMO ", horizon, " ",
                              paste(region, collapse = "+"),
                              if (!is.null(run_datetime))
                                paste0(" @ ", format(aemo_parse_time(run_datetime)))
                              else ""))
}

#' Projected Assessment of System Adequacy (PASA)
#'
#' Returns short-term (`STPASA`, 1-7 day) or medium-term
#' (`MTPASA`, 2-year) system adequacy projections.
#'
#' @param horizon One of `"short"` (default) or `"medium"`.
#' @param region Optional NEM region code.
#' @param start,end Optional window of run-times. Defaults to
#'   the last 24 hours.
#'
#' @return An `aemo_tbl`.
#'
#' @family forecast
#' @export
#' @examples
#' \donttest{
#' op <- options(aemo.cache_dir = tempdir())
#' try({
#'   p <- aemo_pasa(horizon = "short", region = "NSW1")
#' })
#' options(op)
#' }
aemo_pasa <- function(horizon = c("short", "medium"),
                       region = NULL, start = NULL, end = NULL) {
  horizon <- match.arg(horizon)
  if (!is.null(region)) region <- aemo_validate_region(region)
  # Horizon-aware defaults: STPASA is hourly / 1-7 day horizon so a
  # 24-hour lookback makes sense; MTPASA is daily / 2-year horizon
  # published at most once a week, so default to the last 8 days
  # (guaranteed to catch at least one publication).
  if (is.null(end))   end   <- Sys.time()
  if (is.null(start)) {
    lookback <- if (horizon == "short") 1 else 8
    start <- end - as.difftime(lookback, units = "days")
  }
  start <- aemo_parse_time(start)
  end <- aemo_parse_time(end)

  if (horizon == "short") {
    current <- "/Reports/Current/Short_Term_PASA_Reports/"
    archive <- "/Reports/Archive/Short_Term_PASA_Reports/"
    pattern <- "STPASA"
  } else {
    current <- "/Reports/Current/Medium_Term_PASA_Reports/"
    archive <- "/Reports/Archive/Medium_Term_PASA_Reports/"
    pattern <- "MTPASA"
  }

  df <- aemo_fetch_report_range(
    current_dir = current, archive_dir = archive,
    pattern = pattern, start = start, end = end
  )
  df <- aemo_coerce_types(df)
  # PASA target dates extend past [start, end]; filter on region only.
  df <- aemo_apply_filters(df, region = region)
  new_aemo_tbl(df,
               source = "http://nemweb.com.au",
               title = paste0("AEMO ", horizon, "-term PASA"))
}

test_that("aemo_predispatch validates horizon", {
  expect_error(aemo_predispatch("NSW1", horizon = "annual"))
})

test_that("aemo_pasa validates horizon", {
  expect_error(aemo_pasa(horizon = "long"))
})

test_that("aemo_pasa validates region when provided", {
  expect_error(aemo_pasa(horizon = "short", region = "XYZ"))
})

# Settlement reconciliation (SETCFM, SETFCASREGIONRECOVERY,
# SETRESIDUECONTRACTPAYMENT)

#' Settlement cash-flow and residue tables
#'
#' Returns settlement reconciliation tables from MMSDM. Three
#' views are exposed, corresponding to the three tables gentailer
#' hedging workflows most commonly need:
#'
#' - `"cashflow"` (default): `SETCFM` (NEMDE-derived energy
#'   settlement amounts per participant per trading interval).
#' - `"fcas_recovery"`: `SETFCASREGIONRECOVERY` (the recovery
#'   allocation of FCAS costs to customer load per region per
#'   trading interval).
#' - `"residues"`: `SETRESIDUECONTRACTPAYMENT` (settlement
#'   residue auction (SRA) contract payments against
#'   interconnector settlement residues).
#'
#' **Access.** These tables are in the MMSDM monthly archive
#' (not the NEMweb Current retention). Expect two-month latency
#' between trading date and availability.
#'
#' @param table One of `"cashflow"`, `"fcas_recovery"`, or
#'   `"residues"`.
#' @param start,end Window (inclusive) applied to
#'   `SETTLEMENTDATE` / `PAYMENTDATE`.
#'
#' @return An `aemo_tbl` with the raw MMSDM columns for the
#'   requested table.
#'
#' @source AEMO NEMweb MMSDM archive, SETCFM /
#'   SETFCASREGIONRECOVERY / SETRESIDUECONTRACTPAYMENT.
#'   AEMO Copyright Permissions Notice.
#'
#' @seealso [aemo_mlf()] for the transmission loss factors that
#'   scale settlement amounts; [aemo_interconnector()] for the
#'   flow side of the residue calculation.
#'
#' @family reference
#' @export
#' @examples
#' \donttest{
#' op <- options(aemo.cache_dir = tempdir())
#' try({
#'   s <- aemo_settlement(table = "cashflow",
#'                         start = "2024-06-01",
#'                         end   = "2024-06-02")
#'   head(s)
#' })
#' options(op)
#' }
aemo_settlement <- function(table = c("cashflow", "fcas_recovery",
                                        "residues"),
                              start, end) {
  table <- match.arg(table)
  start <- aemo_parse_time(start)
  end <- aemo_parse_time(end)
  stopifnot(end >= start)

  mmsdm_table <- switch(table,
    cashflow       = "SETCFM",
    fcas_recovery  = "SETFCASREGIONRECOVERY",
    residues       = "SETRESIDUECONTRACTPAYMENT"
  )

  months_needed <- unique(format(
    seq(as.Date(start, tz = AEMO_TIMEZONE),
        as.Date(end,   tz = AEMO_TIMEZONE),
        by = "month"),
    "%Y-%m"
  ))

  all_parts <- list()
  for (ym in months_needed) {
    y <- substr(ym, 1L, 4L)
    m <- substr(ym, 6L, 7L)
    url <- aemo_mmsdm_url(mmsdm_table, y, m)
    df <- tryCatch({
      zip_path <- aemo_download_cached(url)
      tmp <- tempfile(paste0("aemo_", tolower(table), "_"))
      dir.create(tmp, recursive = TRUE)
      on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
      utils::unzip(zip_path, exdir = tmp)
      csvs <- list.files(tmp, pattern = "\\.[Cc][Ss][Vv]$", full.names = TRUE)
      if (length(csvs) == 0L) stop("no csv")
      parsed <- aemo_parse_csv(csvs[[1L]])
      parsed[[1L]]
    }, error = function(e) NULL)
    if (!is.null(df) && nrow(df) > 0L) {
      all_parts[[length(all_parts) + 1L]] <- df
    }
  }

  if (length(all_parts) == 0L) {
    cli::cli_abort(c(
      "Could not retrieve {mmsdm_table} from MMSDM.",
      "i" = "Settlement tables are published with ~2 month latency. Check that {.val {min(months_needed)}} is within the published window."
    ))
  }
  common <- Reduce(intersect, lapply(all_parts, names))
  df <- do.call(rbind, lapply(all_parts, function(d) d[, common, drop = FALSE]))
  df <- aemo_coerce_types(df)

  ts_col <- intersect(c("settlementdate", "paymentdate"), names(df))[1L]
  if (!is.na(ts_col)) {
    ts <- if (inherits(df[[ts_col]], "POSIXct")) df[[ts_col]] else
      aemo_parse_col_time(df[[ts_col]])
    keep <- !is.na(ts) & ts >= start & ts <= end
    df <- df[keep, , drop = FALSE]
  }
  rownames(df) <- NULL

  new_aemo_tbl(df,
               source = "https://nemweb.com.au/Data_Archive/Wholesale_Electricity/MMSDM",
               licence = "AEMO Copyright Permissions Notice",
               title = paste0("NEM settlement ", table,
                              " (", mmsdm_table, ")"))
}

test_that("aemo_settlement validates table argument", {
  expect_error(aemo_settlement(table = "invalid",
                                start = "2024-06-01",
                                end   = "2024-06-02"))
})

test_that("aemo_settlement requires start <= end", {
  expect_error(
    aemo_settlement(start = "2024-06-02 00:00:00",
                     end   = "2024-06-01 00:00:00"),
    class = "simpleError"
  )
})

# NEM market price caps, floor, and Cumulative Price Threshold

#' NEM market price caps, price floor, and CPT by financial year
#'
#' Returns a static reference table of the **Market Price Cap
#' (MPC)**, **Market Price Floor (MPF)**, **Cumulative Price
#' Threshold (CPT)**, and **Administered Price Cap (APC)** that
#' apply in each NEM financial year (1 July to 30 June). These
#' are set by the AEMC under NER 3.9.4 and indexed annually to
#' CPI.
#'
#' Use this table to interpret spot-price extremes: any RRP
#' hitting the MPC (typically AUD 15,000 to 17,500 per MWh
#' depending on year) indicates a price cap event; when the
#' rolling-seven-day cumulative price in a region exceeds the
#' CPT (~AUD 1.5 million per MWh-equivalent), AEMO imposes the
#' APC (AUD 300/MWh) until the CPT falls back below the threshold.
#'
#' The table covers 2015-16 onwards and is updated on each
#' package release. For the authoritative current values see
#' <https://www.aemc.gov.au/regulation/energy-rules/national-electricity-rules>
#' and the AEMO Reliability Settings publications.
#'
#' @return An `aemo_tbl` with columns `year` (financial year,
#'   `"YYYY-YY"`), `market_price_cap_aud_per_mwh`,
#'   `market_price_floor_aud_per_mwh`,
#'   `cumulative_price_threshold_aud`,
#'   `administered_price_cap_aud_per_mwh`, and `source`.
#'
#' @family reference
#' @export
#' @examples
#' caps <- aemo_price_caps()
#' head(caps)
aemo_price_caps <- function() {
  # Pre-2022 values from the AEMC 2018 Reliability Standard & Settings
  # Review (RSSR); 2022-23 onwards from the AEMC's 2022 Reliability
  # Settings Review which materially lifted the CPT after the June 2022
  # NEM suspension. The CPT step-change from AUD 232,900 (2020-21) to
  # AUD 1,359,100 (2021-22 revised / 2022-23) reflects the rule change
  # from "7.5 x weekly CPT base" to the new formula.
  # Per-row source: the row's CPT / MPC / APC comes from the AEMC
  # determination cited in `source_doc`.
  df <- data.frame(
    year = c("2015-16", "2016-17", "2017-18", "2018-19",
             "2019-20", "2020-21", "2021-22", "2022-23",
             "2023-24", "2024-25", "2025-26"),
    market_price_cap_aud_per_mwh = c(
      13800, 14000, 14200, 14500,
      14700, 15000, 15100, 15500,
      16600, 17500, 17500
    ),
    market_price_floor_aud_per_mwh = rep(-1000, 11),
    cumulative_price_threshold_aud = c(
      212800, 216500, 219700, 224300,
      228700, 232900, 1359100, 1398100,
      1502100, 1577100, 1626800
    ),
    administered_price_cap_aud_per_mwh = rep(300, 11),
    source_doc = c(
      "AEMC 2014 RSSR final determination (CPI-indexed annually)",
      "AEMC 2014 RSSR final determination (CPI-indexed annually)",
      "AEMC 2014 RSSR final determination (CPI-indexed annually)",
      "AEMC 2018 RSSR final determination (CPI-indexed annually)",
      "AEMC 2018 RSSR final determination (CPI-indexed annually)",
      "AEMC 2018 RSSR final determination (CPI-indexed annually)",
      "AEMC 2020 RSSR final determination; CPT revised per 2022 review",
      "AEMC 2022 Reliability Settings Review final determination",
      "AEMC 2022 Reliability Settings Review + 2023 CPI indexation",
      "AEMC 2024 Reliability Settings Review final determination",
      "AEMC 2024 Reliability Settings Review + 2025 CPI indexation"
    ),
    stringsAsFactors = FALSE
  )
  new_aemo_tbl(df,
               source = "https://www.aemc.gov.au/regulation/energy-rules/national-electricity-rules",
               licence = "AEMO Copyright Permissions Notice",
               title = "NEM market price caps, floor, CPT, and APC")
}

test_that("aemo_price_caps has expected shape and source_doc column", {
  caps <- aemo_price_caps()
  expect_s3_class(caps, "aemo_tbl")
  expect_true(all(c("year", "market_price_cap_aud_per_mwh",
                    "market_price_floor_aud_per_mwh",
                    "cumulative_price_threshold_aud",
                    "administered_price_cap_aud_per_mwh",
                    "source_doc") %in% names(caps)))
  # Post-June-2022 CPT step change is the headline data point
  v2022 <- caps[caps$year == "2022-23", ]
  expect_equal(v2022$cumulative_price_threshold_aud, 1398100)
  expect_match(v2022$source_doc, "AEMC 2022")
  # Every row has a source
  expect_true(all(nzchar(caps$source_doc)))
})

---
title: "NEM timestamps: AEST, period-ending, and the 5MS transition"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{NEM timestamps: AEST, period-ending, and the 5MS transition}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## The NEM market clock

The National Electricity Market operates on **Australian Eastern Standard Time (AEST), UTC+10, year-round**. This is not the same as the wall-clock time in Sydney or Melbourne, which advance one hour to AEDT (UTC+11) during Daylight Saving Time (October to April each year).

The legal basis is NER clause 2.2.6. Using wall-clock time for any AEMO data analysis is a common error: every summer timestamp will be parsed one hour ahead of the file clock, and timestamps at the DST transition boundaries will become ambiguous or `NA`.

The `aemo` package uses `"Australia/Brisbane"` as the timezone for all data. Queensland observes AEST year-round (no DST), which gives the correct fixed UTC+10 offset:

```{r, eval=FALSE}
library(aemo)

# All timestamps come back in AEST (UTC+10, no DST).
# The tzone attribute is "Australia/Brisbane" throughout.
p <- aemo_price("NSW1", "2024-01-15 14:00", "2024-01-15 15:00")
attr(p$settlementdate, "tzone")  # "Australia/Brisbane"

# January is summer in Australia. Australia/Sydney would give UTC+11 here,
# shifting every timestamp one hour ahead of the AEMO file clock.
# Australia/Brisbane correctly gives UTC+10.
format(p$settlementdate[1], "%Y-%m-%d %H:%M:%S %Z")
```

To display times in local wall-clock time for a specific state, convert explicitly:

```{r, eval=FALSE}
# Convert market time to NSW wall clock (observes DST)
p$local_time_nsw <- format(p$settlementdate, tz = "Australia/Sydney")

# Convert market time to SA wall clock (UTC+9:30 / UTC+10:30 with DST)
p$local_time_sa <- format(p$settlementdate, tz = "Australia/Adelaide")
```

## Period-ending timestamps

Every interval in a NEMweb file is labelled with its **period-ending** timestamp. A row with `settlementdate = "2024-01-15 14:05:00"` covers the 5-minute dispatch interval from 14:00:01 to 14:05:00 (inclusive).

This is the AEMO convention throughout: the timestamp marks when the interval closes, not when it opens. For 30-minute trading intervals the same rule applies: a row stamped `14:30:00` covers 14:00:01 to 14:30:00.

When computing time-of-day statistics (e.g. peak demand by hour) use the interval end-time directly. When joining to external time series that use period-*start* timestamps, subtract the interval length:

```{r, eval=FALSE}
# Period-start timestamps for 5-min intervals
p$period_start <- p$settlementdate - as.difftime(5, units = "mins")

# Period-start timestamps for 30-min trading intervals
t$period_start <- t$settlementdate - as.difftime(30, units = "mins")
```

## The 5-minute settlement transition (1 October 2021)

Before 1 October 2021 the NEM settled electricity on **30-minute trading intervals** using `TRADINGPRICE` from the TRADINGIS reports. On that date five-minute settlement (5MS) commenced under NER clause 3.15.1A, and settlement moved to native **5-minute dispatch prices** from DISPATCHPRICE.

Practical implications:

| Query | What `aemo_price()` returns |
|-------|----------------------------|
| `interval = "5min"`, any date | Native 5-min dispatch price from DISPATCHPRICE (DISPATCHIS) |
| `interval = "30min"`, date before 2021-10-01 | 30-min TRADINGPRICE from TRADINGIS |
| `interval = "30min"`, date from 2021-10-01 | Arithmetic mean of six 5-min dispatch prices within the trading interval |
| `interval = "30min"`, spanning 2021-10-01 | Pre-5MS portion from TRADINGIS; post-5MS aggregated from 5-min |

When constructing long historical 30-minute price series that span the transition, the aggregation method matches: both the pre-5MS TRADINGPRICE and the post-5MS arithmetic mean are simple unweighted averages of the six dispatch prices within the trading interval (pre-5MS dispatch prices were held constant at the trading-interval price; post-5MS they vary).

## Intervention runs

DISPATCHPRICE contains two types of dispatch runs in every file:

- **INTERVENTION = 0**: the market pricing run. This is the price used in settlement and the one most analyses want.
- **INTERVENTION = 1**: the physical/intervention run, published when AEMO has issued an intervention direction (e.g. the June 2022 market suspension). Both runs coexist in the same file.

All `aemo` data functions default to `intervention = FALSE`, which filters to market-pricing rows only. Pass `intervention = TRUE` when you need to analyse the intervention pricing itself (e.g. the cost of directions during a suspension period).

```{r, eval=FALSE}
# Default: market pricing run only
p_market <- aemo_price("NSW1", "2022-06-13", "2022-06-14")

# Both runs (for intervention analysis)
p_both <- aemo_price("NSW1", "2022-06-13", "2022-06-14",
                     intervention = TRUE)
table(p_both$intervention)
```

---
title: "Data dictionary: columns, tables, and AEMO source"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Data dictionary: columns, tables, and AEMO source}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

This vignette maps each `aemo` package function to the underlying AEMO table, lists the key columns returned, and defines the units and conventions. Use it alongside the AEMO Market Management System Data Model (MMSDM) Data Model documentation, which is the authoritative source.

## Conventions

**Timezone.** All `settlementdate` and timestamp columns are `POSIXct` in `"Australia/Brisbane"` (AEST, UTC+10, no daylight saving). See `vignette("timestamps")` for details.

**Period-ending.** A row stamped `14:05:00` covers the interval ending at 14:05 (i.e. 14:00:01 to 14:05:00 inclusive).

**INTERVENTION.** All data functions default to `intervention = FALSE`, returning only the market pricing run (`INTERVENTION = 0`). Pass `intervention = TRUE` to include physical/intervention runs.

**Types.** All numeric columns are coerced from the raw character output of the AEMO CSV files. Timestamps are POSIXct. ID columns (`regionid`, `duid`, `constraintid`, etc.) remain character.

**GST.** All NEM wholesale prices in AEMO's feeds, and therefore in every `aemo_tbl`, are *exclusive of GST* (NER Chapter 10, definition of "spot price"). When comparing revenue figures to published financial statements or retail prices, gross up by 1.10 as appropriate.

---

## `aemo_price()` : DISPATCHPRICE / TRADINGPRICE

**Source tables:**
- 5-min: `DISPATCH_PRICE` from `DISPATCHIS_Reports`
- 30-min pre-5MS: `TRADING_PRICE` from `TRADINGIS_Reports`
- 30-min post-5MS (from 1 Oct 2021): aggregated from `DISPATCH_PRICE`

| Column | Type | Unit | Description |
|--------|------|------|-------------|
| `settlementdate` | POSIXct | AEST | Period-ending timestamp |
| `regionid` | character | - | NEM region (e.g. `"NSW1"`) |
| `rrp` | numeric | AUD/MWh | Regional Reference Price : the spot price used in settlement |
| `raise6secrrp` | numeric | AUD/MW | FCAS Raise 6-second service price (market run) |
| `lower6secrrp` | numeric | AUD/MW | FCAS Lower 6-second service price |
| `raise60secrrp` | numeric | AUD/MW | FCAS Raise 60-second service price |
| `lower60secrrp` | numeric | AUD/MW | FCAS Lower 60-second service price |
| `raise5minrrp` | numeric | AUD/MW | FCAS Raise 5-minute service price |
| `lower5minrrp` | numeric | AUD/MW | FCAS Lower 5-minute service price |
| `raiseregrrp` | numeric | AUD/MW | FCAS Regulation Raise price |
| `lowerregrrp` | numeric | AUD/MW | FCAS Regulation Lower price |
| `raise1secrrp` | numeric | AUD/MW | FCAS Very Fast Raise (from 9 Oct 2023) |
| `lower1secrrp` | numeric | AUD/MW | FCAS Very Fast Lower (from 9 Oct 2023) |
| `intervention` | character | 0/1 | 0 = market pricing run; 1 = physical/intervention run |

The `rrp` column may occasionally equal the Market Price Cap (MPC, ~AUD 17,500/MWh) or Market Price Floor (MPF, -AUD 1,000/MWh). See `aemo_price_caps()` for the exact values by financial year.

---

## `aemo_demand()` : DISPATCHREGIONSUM

**Source table:** `DISPATCH_REGIONSUM` from `DISPATCHIS_Reports`

| Column | Type | Unit | Description |
|--------|------|------|-------------|
| `settlementdate` | POSIXct | AEST | Period-ending timestamp |
| `regionid` | character | - | NEM region |
| `demand_mw` | numeric | MW | The requested demand measure (see below) |
| `totaldemand` | numeric | MW | Operational demand: grid-measured, met by scheduled/semi-scheduled gen |
| `ss_solar_uigf` | numeric | MW | Semi-scheduled solar UIGF (used for `operational_less_snsg`) |
| `ss_wind_uigf` | numeric | MW | Semi-scheduled wind UIGF |

**`measure` argument:**
- `"operational"`: `TOTALDEMAND` : what AEMO dispatches against
- `"operational_less_snsg"`: `TOTALDEMAND` minus semi-scheduled solar + wind UIGF
- `"native"`: `TOTALDEMAND` plus estimated rooftop PV : closest to actual end-use consumption

---

## `aemo_dispatch_units()` : DISPATCH_UNIT_SCADA / DISPATCHLOAD

| Column | Type | Unit | Description |
|--------|------|------|-------------|
| `settlementdate` | POSIXct | AEST | Period-ending timestamp |
| `duid` | character | - | Dispatchable Unit Identifier |
| `scadavalue` | numeric | MW | Actual SCADA output (when `measure = "scada_mw"`) |
| `totalmw` | numeric | MW | Dispatch target MW (when `measure = "target_mw"`, from DISPATCHLOAD) |

---

## `aemo_constraints()` : DISPATCHCONSTRAINT

**Source table:** `DISPATCH_CONSTRAINT` from `DISPATCHIS_Reports`

| Column | Type | Unit | Description |
|--------|------|------|-------------|
| `settlementdate` | POSIXct | AEST | Period-ending timestamp |
| `constraintid` | character | - | Generic constraint ID (e.g. `"N>>N-NIL_1"`) |
| `rhs` | numeric | MW or other | Right-hand side of the constraint (thermal limit, voltage limit, etc.) |
| `marginalvalue` | numeric | AUD/MWh | Shadow price on the constraint. Non-zero means the constraint is binding. The sum of marginalvalues across all binding constraints at a Regional Reference Node equals the network component of the regional price. |
| `violationdegree` | numeric | MW | How far the constraint was violated (should be zero in a normal dispatch) |
| `lhs` | numeric | MW or other | Left-hand side value (the actual flow or sum of flows) |

`min_marginal_value = 0.01` (default) suppresses rows where the shadow price rounds to zero : these are near-binding constraints not actively driving price.

---

## `aemo_gencon()` : GENCONDATA

**Source table:** `GENCONDATA` from MMSDM archive

| Column | Type | Description |
|--------|------|-------------|
| `genconid` | character | Constraint equation ID : matches `constraintid` in DISPATCHCONSTRAINT |
| `constrainttype` | character | `"LE"` (≤), `"GE"` (≥), or `"EQ"` (=) |
| `description` | character | Plain-language description of what the constraint models |
| `genericconstraintrhs` | numeric | Default RHS value (may be overridden in real-time by GENCONSETINVOKE) |
| `effectivedate` | POSIXct | Date from which this equation definition applies |

---

## `aemo_fcas_enablement()` : DISPATCHLOAD

**Source table:** `DISPATCH_UNIT_SOLUTION` from `DISPATCHIS_Reports`

| Column | Type | Unit | Description |
|--------|------|------|-------------|
| `settlementdate` | POSIXct | AEST | Period-ending timestamp |
| `duid` | character | - | DUID |
| `raise1secmw` | numeric | MW | Very Fast Raise enabled MW (from 9 Oct 2023) |
| `lower1secmw` | numeric | MW | Very Fast Lower enabled MW |
| `raise6secmw` | numeric | MW | Fast Raise enabled MW |
| `lower6secmw` | numeric | MW | Fast Lower enabled MW |
| `raise60secmw` | numeric | MW | Slow Raise enabled MW |
| `lower60secmw` | numeric | MW | Slow Lower enabled MW |
| `raise5minmw` | numeric | MW | Delayed Raise enabled MW |
| `lower5minmw` | numeric | MW | Delayed Lower enabled MW |
| `raiseregmw` | numeric | MW | Regulation Raise enabled MW |
| `lowerregmw` | numeric | MW | Regulation Lower enabled MW |

---

## `aemo_interconnector()` : DISPATCHINTERCONNECTORRES

| Column | Type | Unit | Description |
|--------|------|------|-------------|
| `settlementdate` | POSIXct | AEST | Period-ending timestamp |
| `interconnectorid` | character | - | Interconnector ID (e.g. `"NSW1-QLD1"`, `"V-SA"`) |
| `mwflow` | numeric | MW | Actual MW flow (positive = export from `from_region`) |
| `mwlosses` | numeric | MW | MW losses on the interconnector |
| `exportlimit` | numeric | MW | Export limit in this interval |
| `importlimit` | numeric | MW | Import limit in this interval |

---

## `aemo_mlf()` : MARGINALLOSSFACTOR

| Column | Type | Description |
|--------|------|-------------|
| `financial_year` | character | NEM financial year, e.g. `"2024-25"` (1 July to 30 June) |
| `duid` | character | DUID |
| `connectionpointid` | character | Connection point ID |
| `regionid` | character | NEM region |
| `mlf` | numeric | Marginal Loss Factor. A value of 0.97 means the generator receives 97% of the regional RRP per MWh. Used in settlement: `payment = MWh × RRP × MLF × DLF`. |

---

## `aemo_dlf()` : LOSSFACTORMODEL

| Column | Type | Description |
|--------|------|-------------|
| `financial_year` | character | NEM financial year |
| `connectionpointid` | character | Connection point ID |
| `dlf` | numeric | Distribution Loss Factor. Combined with MLF to give total loss factor (TLF = MLF × DLF). |

---

## `aemo_price_caps()` : static reference

| Column | Type | Description |
|--------|------|-------------|
| `year` | character | Financial year, e.g. `"2024-25"` |
| `market_price_cap_aud_per_mwh` | numeric | Market Price Cap (MPC) : maximum spot price. Any RRP equal to this value is a cap event. |
| `market_price_floor_aud_per_mwh` | numeric | Market Price Floor (MPF) : minimum spot price (AUD -1,000/MWh since 2013). |
| `cumulative_price_threshold_aud` | numeric | Cumulative Price Threshold (CPT) : when rolling-7-day regional CPT is breached, the Administered Price Cap (APC) activates. |
| `administered_price_cap_aud_per_mwh` | numeric | Administered Price Cap (APC) : price ceiling during CPT-triggered suspension (AUD 300/MWh). |

---

## `aemo_units()` : DUDETAILSUMMARY

| Column | Type | Description |
|--------|------|-------------|
| `duid` | character | Dispatchable Unit Identifier : the primary key used in all dispatch tables |
| `stationid` | character | Station (power plant) ID : multiple DUIDs may share one station |
| `regionid` | character | NEM region |
| `dispatchtype` | character | `"GENERATOR"` or `"LOAD"` |
| `schedule_type` | character | `"SCHEDULED"`, `"SEMI-SCHEDULED"`, or `"NON-SCHEDULED"` |
| `connectionpointid` | character | Connection point : used to look up MLF/DLF |
| `start_date` | POSIXct | Registration start date |
| `end_date` | POSIXct | Registration end date (`NA` = currently active) |

---

## `aemo_outages()` : NETWORK_OUTAGEDETAIL

| Column | Type | Description |
|--------|------|-------------|
| `outageid` | character | Unique outage identifier |
| `starttime` | POSIXct | Outage start (AEST) |
| `endtime` | POSIXct | Outage end (AEST, `NA` if still open) |
| `substationid` | character | Substation ID |
| `equipmenttype` | character | Type of network element (e.g. `"LINE"`, `"TRANSFORMER"`) |
| `equipmentid` | character | Specific element ID |
| `outagetype` | character | `"PLANNED"` or `"FORCED"` |
| `regionid` | character | NEM region |
| `restarttimeunknown` | character | `"Y"` if the end time is uncertain |

---

## `aemo_participants()` : PARTICIPANT + DUDETAILSUMMARY

| Column | Type | Description |
|--------|------|-------------|
| `participantid` | character | AEMO participant identifier (company code) |
| `participantclassid` | character | `"GENERATOR"`, `"LOAD"`, `"TRADER"`, etc. |
| `name` | character | Company name |
| `duid` | character | DUID registered to this participant |
| `stationid` | character | Station ID |
| `regionid` | character | NEM region |
| `dispatchtype` | character | `"GENERATOR"` or `"LOAD"` |
| `schedule_type` | character | Schedule classification |

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constraints.R
\name{aemo_constraints}
\alias{aemo_constraints}
\title{Binding transmission and system constraints}
\source{
AEMO NEMweb DISPATCHIS_Reports, DISPATCHCONSTRAINT
table.
}
\usage{
aemo_constraints(
  start,
  end,
  constraint_id = NULL,
  intervention = FALSE,
  min_marginal_value = 0.01
)
}
\arguments{
\item{start, end}{Window (inclusive), character or POSIXct.}

\item{constraint_id}{Optional character vector of constraint
IDs (e.g. \code{"N>>N-NIL_1"}, \code{"V::S_NIL_TBSE"}). \code{NULL}
returns all binding constraints in the window.}

\item{intervention}{Logical. Default \code{FALSE}.}

\item{min_marginal_value}{Numeric. Only return constraints
with marginal value at or above this threshold (AUD/MWh).
\code{0} returns every row; the default of \code{0.01} filters out
near-zero shadow prices that are typically noise.}
}
\value{
An \code{aemo_tbl} with columns \code{settlementdate},
\code{constraintid}, \code{rhs}, \code{marginalvalue}, \code{violationdegree},
\code{lhs}, plus the intervention flag.
}
\description{
Returns the \code{DISPATCHCONSTRAINT} table from NEMweb: one row
per binding (or near-binding) constraint per 5-minute dispatch
interval. Each row records the constraint ID, the left-hand
side (LHS) and right-hand side (RHS) values, the marginal
value (shadow price on the constraint in AUD/MWh), and the
violation degree if any.
}
\details{
This is the table that answers the question "why was the RRP
so high at 17:35?": the sum of marginal values across
binding constraints at the Regional Reference Node equals the
regional price component attributable to network limits.

Constraint equations and RHS terms (\code{GENCONDATA},
\code{GENCONSETINVOKE}) are published through MMSDM on a separate
cadence and are not exposed directly by this function; use
\code{\link[=aemo_nemweb_download]{aemo_nemweb_download()}} on an MMSDM URL for those.
}
\examples{
\donttest{
op <- options(aemo.cache_dir = tempdir())
try({
  now <- Sys.time()
  c <- aemo_constraints(start = now - 3600, end = now)
  head(c)
})
options(op)
}
}
\seealso{
Other dispatch: 
\code{\link{aemo_bids}()},
\code{\link{aemo_dispatch_units}()},
\code{\link{aemo_fcas_enablement}()},
\code{\link{aemo_gencon}()},
\code{\link{aemo_interconnector}()},
\code{\link{aemo_market_notices}()},
\code{\link{aemo_outages}()},
\code{\link{aemo_rooftop_pv}()},
\code{\link{aemo_spd_constraints}()}
}
\concept{dispatch}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constraints.R
\name{aemo_spd_constraints}
\alias{aemo_spd_constraints}
\title{SPD constraint tables (regions, interconnectors, connection points)}
\source{
AEMO NEMweb MMSDM archive, SPDREGIONCONSTRAINT /
SPDINTERCONNECTORCONSTRAINT / SPDCONNECTIONPOINTCONSTRAINT.
}
\usage{
aemo_spd_constraints(
  table = c("region", "interconnector", "connection_point"),
  constraint_id = NULL
)
}
\arguments{
\item{table}{One of \code{"region"} (default), \code{"interconnector"},
or \code{"connection_point"}.}

\item{constraint_id}{Optional character vector of
\code{GENCONID}/\code{CONSTRAINTID} values.}
}
\value{
An \code{aemo_tbl} with columns from the requested SPD
table. \code{GENCONID} and \code{FACTOR} (the coefficient) are
always present; other columns vary by table.
}
\description{
Returns the SPD (Security and Projected Dispatch) constraint
coefficient tables from MMSDM. Where \code{GENCONDATA} gives the
high-level equation definition, the SPD tables give the
per-term \emph{coefficients} used in the NEMDE dispatch
optimisation:
}
\details{
\itemize{
\item \code{"region"}: \code{SPDREGIONCONSTRAINT} (regional-demand and
regional-generation terms).
\item \code{"interconnector"}: \code{SPDINTERCONNECTORCONSTRAINT}
(interconnector-flow terms).
\item \code{"connection_point"}: \code{SPDCONNECTIONPOINTCONSTRAINT}
(per-DUID connection-point terms).
}

These tables are required for NEM dispatch replication (the
\code{nempy} Python package, Gorman, Bruce & MacGill 2022,
\emph{JOSS} 7(70) 3596, \url{doi:10.21105/joss.03596}, uses all three
when reproducing NEMDE solves).
}
\examples{
\donttest{
op <- options(aemo.cache_dir = tempdir())
try({
  s <- aemo_spd_constraints(table = "interconnector")
  head(s)
})
options(op)
}
}
\seealso{
\code{\link[=aemo_gencon]{aemo_gencon()}} for the equation-level metadata,
\code{\link[=aemo_constraints]{aemo_constraints()}} for the 5-min binding-constraint
feed with shadow prices.

Other dispatch: 
\code{\link{aemo_bids}()},
\code{\link{aemo_constraints}()},
\code{\link{aemo_dispatch_units}()},
\code{\link{aemo_fcas_enablement}()},
\code{\link{aemo_gencon}()},
\code{\link{aemo_interconnector}()},
\code{\link{aemo_market_notices}()},
\code{\link{aemo_outages}()},
\code{\link{aemo_rooftop_pv}()}
}
\concept{dispatch}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gas.R
\name{aemo_gas}
\alias{aemo_gas}
\title{Gas market data (STTM, DWGM)}
\usage{
aemo_gas(market = c("sttm", "dwgm"), hub = NULL, start, end)
}
\arguments{
\item{market}{One of \code{"sttm"} (default) or \code{"dwgm"}.}

\item{hub}{Optional STTM hub: \code{"adelaide"}, \code{"brisbane"}, or
\code{"sydney"}. Ignored for DWGM.}

\item{start, end}{Window.}
}
\value{
An \code{aemo_tbl}.
}
\description{
Returns Short Term Trading Market (\code{STTM}, Adelaide, Brisbane,
Sydney hubs) or Declared Wholesale Gas Market (\code{DWGM},
Victoria) prices and volumes.
}
\examples{
\donttest{
op <- options(aemo.cache_dir = tempdir())
try({
  now <- Sys.time()
  g <- aemo_gas(market = "sttm", hub = "sydney",
                 start = now - 7 * 86400, end = now)
  head(g)
})
options(op)
}
}
\concept{gas}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reference.R
\name{aemo_units}
\alias{aemo_units}
\title{NEM generators (DUID registry)}
\source{
AEMO NEMweb MMSDM archive, DUDETAILSUMMARY table,
AEMO Copyright Permissions Notice.
}
\usage{
aemo_units(as_of = NULL)
}
\arguments{
\item{as_of}{Optional \code{Date} or \code{POSIXct}. Returns the DUID
registry as it was on that date. \code{NULL} (default) returns
the current registry.}
}
\value{
An \code{aemo_tbl} keyed by \code{duid}. Columns include (from
DUDETAILSUMMARY): \code{duid}, \code{stationid}, \code{regionid},
\code{dispatchtype} (GENERATOR / LOAD), \code{connectionpointid},
\code{schedule_type} (SCHEDULED / SEMI-SCHEDULED /
NON-SCHEDULED), \code{start_date}, \code{end_date}.
}
\description{
Downloads the \code{DUDETAILSUMMARY} table from the most recent
MMSDM monthly archive and returns one row per registered DUID
(Dispatchable Unit Identifier) with station, region, dispatch
type, classification, and schedule type. Typical output is
500+ DUIDs covering scheduled, semi-scheduled, and
non-scheduled generators, bidirectional storage (BESS), and
loads.
}
\details{
\strong{Effective-date filtering.} \code{DUDETAILSUMMARY} is
effective-dated (\code{START_DATE}, \code{END_DATE} per row with
multiple \code{VERSIONNO} vintages per DUID over time). The
default (\code{as_of = NULL}) returns rows whose
\verb{[START_DATE, END_DATE]} interval covers the date the MMSDM
archive was published: i.e. the \emph{current} registry. Pass an
\code{as_of} date to get the registry as it was on that date
(essential for historical analysis: Liddell's four DUIDs
were retired in April 2023; pre-2023 queries need them).

This matches the as-of-join pattern documented in Gorman et
al. (2018) \emph{NEMOSIS} (APSRC) for correct historical joins.
}
\examples{
\donttest{
op <- options(aemo.cache_dir = tempdir())
try({
  # Current DUID registry
  u <- aemo_units()

  # DUIDs as they were on 1 March 2023 (pre-Liddell retirement)
  u_2023 <- aemo_units(as_of = "2023-03-01")
})
options(op)
}
}
\seealso{
Other reference: 
\code{\link{aemo_dlf}()},
\code{\link{aemo_interconnectors}()},
\code{\link{aemo_mlf}()},
\code{\link{aemo_participants}()},
\code{\link{aemo_price_caps}()},
\code{\link{aemo_regions}()},
\code{\link{aemo_settlement}()},
\code{\link{aemo_snapshot}()}
}
\concept{reference}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dispatch.R
\name{aemo_fcas_enablement}
\alias{aemo_fcas_enablement}
\title{FCAS enablement volumes by generator unit}
\source{
AEMO NEMweb DISPATCHIS_Reports, DISPATCHLOAD table.
}
\usage{
aemo_fcas_enablement(
  duid = NULL,
  start,
  end,
  service = NULL,
  intervention = FALSE
)
}
\arguments{
\item{duid}{Optional character vector of DUIDs. \code{NULL} returns
all units. See \code{\link[=aemo_units]{aemo_units()}} for the full DUID registry.}

\item{start, end}{Window (inclusive), character or POSIXct.}

\item{service}{Optional character vector of service names,
e.g. \code{c("raise6sec", "lowerreg")}. Case-insensitive.
\code{NULL} returns all ten services.}

\item{intervention}{Logical. Default \code{FALSE}.}
}
\value{
An \code{aemo_tbl} with columns \code{settlementdate}, \code{duid},
and one column per FCAS service (\code{raise1secmw},
\code{lower1secmw}, \code{raise6secmw}, \code{lower6secmw},
\code{raise60secmw}, \code{lower60secmw}, \code{raise5minmw},
\code{lower5minmw}, \code{raiseregmw}, \code{lowerregmw}). Values are
MW; zero means the unit was not enabled for that service.
}
\description{
Returns the MW enabled for each of the ten Frequency Control
Ancillary Services (FCAS) per DUID per 5-minute dispatch
interval from \code{DISPATCHLOAD}. This is the quantity side of
the FCAS market: how much each unit was enabled (dispatched)
to provide each service, as distinct from the price returned
by \code{\link[=aemo_fcas]{aemo_fcas()}}.
}
\details{
Ten services are active since 9 October 2023 when Very Fast
(\code{RAISE1SEC} / \code{LOWER1SEC}) commenced. Before that date only
eight services are present; the \code{raise1secmw} / \code{lower1secmw}
columns will be \code{NA} for intervals before that date.
}
\examples{
\donttest{
op <- options(aemo.cache_dir = tempdir())
try({
  now <- Sys.time()
  e <- aemo_fcas_enablement(start = now - 3600, end = now)
  head(e)
})
options(op)
}
}
\seealso{
Other dispatch: 
\code{\link{aemo_bids}()},
\code{\link{aemo_constraints}()},
\code{\link{aemo_dispatch_units}()},
\code{\link{aemo_gencon}()},
\code{\link{aemo_interconnector}()},
\code{\link{aemo_market_notices}()},
\code{\link{aemo_outages}()},
\code{\link{aemo_rooftop_pv}()},
\code{\link{aemo_spd_constraints}()}
}
\concept{dispatch}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/price.R
\name{aemo_fcas}
\alias{aemo_fcas}
\title{Frequency control ancillary services (FCAS) prices}
\usage{
aemo_fcas(region, start, end, service = NULL, intervention = FALSE)
}
\arguments{
\item{region}{NEM region code.}

\item{start, end}{Window (inclusive).}

\item{service}{Optional character vector of service names
(e.g. \code{"RAISE6SEC"}, \code{"LOWER60SEC"}, \code{"RAISE1SEC"}).}

\item{intervention}{Logical. See \code{\link[=aemo_price]{aemo_price()}}.}
}
\value{
An \code{aemo_tbl} with one row per interval and columns
for each requested FCAS RRP (AUD/MW).
}
\description{
Returns regional FCAS market prices across the eight
contingency services plus the two regulation services. Ten
services are live in the NEM since R1/L1 Very Fast commenced
on 9 October 2023.
}
\details{
Thin wrapper over \code{aemo_price(..., market = "fcas")}.
}
\examples{
\donttest{
op <- options(aemo.cache_dir = tempdir())
try({
  now <- Sys.time()
  f <- aemo_fcas("NSW1", now - 3600, now)
  head(f)
})
options(op)
}
}
\seealso{
Other price: 
\code{\link{aemo_price}()}
}
\concept{price}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/forecasts.R
\name{aemo_pasa}
\alias{aemo_pasa}
\title{Projected Assessment of System Adequacy (PASA)}
\usage{
aemo_pasa(
  horizon = c("short", "medium"),
  region = NULL,
  start = NULL,
  end = NULL
)
}
\arguments{
\item{horizon}{One of \code{"short"} (default) or \code{"medium"}.}

\item{region}{Optional NEM region code.}

\item{start, end}{Optional window of run-times. Defaults to
the last 24 hours.}
}
\value{
An \code{aemo_tbl}.
}
\description{
Returns short-term (\code{STPASA}, 1-7 day) or medium-term
(\code{MTPASA}, 2-year) system adequacy projections.
}
\examples{
\donttest{
op <- options(aemo.cache_dir = tempdir())
try({
  p <- aemo_pasa(horizon = "short", region = "NSW1")
})
options(op)
}
}
\seealso{
Other forecast: 
\code{\link{aemo_predispatch}()}
}
\concept{forecast}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reference.R
\name{aemo_interconnectors}
\alias{aemo_interconnectors}
\title{NEM interconnectors}
\usage{
aemo_interconnectors()
}
\value{
An \code{aemo_tbl} with columns \code{interconnector_id},
\code{from_region}, \code{to_region}, \code{name}, \code{energised}.
}
\description{
Returns a static table of the seven NEM interconnectors. The
seventh entry is Project EnergyConnect (PEC), whose Stage 1
was energised on 30 April 2025 with full ~800 MW capability
expected mid-2026.
}
\examples{
aemo_interconnectors()
}
\seealso{
Other reference: 
\code{\link{aemo_dlf}()},
\code{\link{aemo_mlf}()},
\code{\link{aemo_participants}()},
\code{\link{aemo_price_caps}()},
\code{\link{aemo_regions}()},
\code{\link{aemo_settlement}()},
\code{\link{aemo_snapshot}()},
\code{\link{aemo_units}()}
}
\concept{reference}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loss_factors.R
\name{aemo_dlf}
\alias{aemo_dlf}
\title{Distribution Loss Factors (DLF) by connection point}
\source{
AEMO MMSDM archive, LOSSFACTORMODEL table.
AEMO Copyright Permissions Notice.
}
\usage{
aemo_dlf(year = NULL, connection_point_id = NULL)
}
\arguments{
\item{year}{Financial year(s) as \code{"YYYY-YY"}. \code{NULL} returns
all available years.}

\item{connection_point_id}{Optional character vector of
connection point IDs.}
}
\value{
An \code{aemo_tbl} with at minimum \code{financial_year},
\code{connectionpointid}, and \code{dlf}.
}
\description{
Returns the Distribution Loss Factor for NEM connection
points from the MMSDM \code{LOSSFACTORMODEL} table. DLFs measure
the average energy loss on the distribution network between
a connection point and the transmission network boundary; a
DLF of 1.02 means the generator must produce 2\% more energy
than it delivers to the transmission system.
}
\details{
DLFs are published annually by AEMO (NER 3.6.3) and combined
with MLFs in settlement to give the total loss factor (TLF =
MLF x DLF) used in energy payments.
}
\examples{
\donttest{
op <- options(aemo.cache_dir = tempdir())
try({
  dlf <- aemo_dlf(year = "2024-25")
  head(dlf)
})
options(op)
}
}
\seealso{
Other reference: 
\code{\link{aemo_interconnectors}()},
\code{\link{aemo_mlf}()},
\code{\link{aemo_participants}()},
\code{\link{aemo_price_caps}()},
\code{\link{aemo_regions}()},
\code{\link{aemo_settlement}()},
\code{\link{aemo_snapshot}()},
\code{\link{aemo_units}()}
}
\concept{reference}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/market_notices.R
\name{aemo_market_notices}
\alias{aemo_market_notices}
\title{NEM market notices}
\source{
AEMO NEMweb MMSDM archive, MARKETNOTICEDATA table.
AEMO Copyright Permissions Notice.
}
\usage{
aemo_market_notices(start, end, notice_type = NULL, region = NULL)
}
\arguments{
\item{start, end}{Window (inclusive) applied to \code{EFFECTIVEDATE}.}

\item{notice_type}{Optional character vector (e.g. \code{"LOR1"},
\code{"RERT"}, \code{"PRICES SUBJECT TO REVIEW"}). Case-insensitive
substring match.}

\item{region}{Optional NEM region code.}
}
\value{
An \code{aemo_tbl} with columns from MARKETNOTICEDATA:
\code{noticeid}, \code{effectivedate}, \code{typeid} (notice category),
\code{originator}, \code{priority}, \code{reason} (the notice text),
\code{externalreference}, and where present \code{regionid}.
}
\description{
Returns the \code{MARKETNOTICEDATA} feed from MMSDM. Market
notices are AEMO's free-text log of market-relevant events:
Lack of Reserve (LOR1/2/3) declarations, Reliability and
Emergency Reserve Trader (RERT) activations, market
suspensions, market directions, unit withdrawals, system
security events, administered price declarations, and
operator advisories.
}
\details{
Pair with \code{\link[=aemo_constraints]{aemo_constraints()}} and \code{\link[=aemo_price]{aemo_price()}} to sequence
the causal chain of a price event. Rangarajan, Svec, Foley
and Trück (2025, \emph{Energy Economics} 141) use
\code{MARKETNOTICEDATA} to order the intervention messages that
mark entry to and exit from the June 2022 NEM suspension.
}
\examples{
\donttest{
op <- options(aemo.cache_dir = tempdir())
try({
  # LOR declarations during the June 2022 NEM suspension
  n <- aemo_market_notices(
    start = "2022-06-13",
    end   = "2022-06-14",
    notice_type = "LOR"
  )
  head(n)
})
options(op)
}
}
\seealso{
Other dispatch: 
\code{\link{aemo_bids}()},
\code{\link{aemo_constraints}()},
\code{\link{aemo_dispatch_units}()},
\code{\link{aemo_fcas_enablement}()},
\code{\link{aemo_gencon}()},
\code{\link{aemo_interconnector}()},
\code{\link{aemo_outages}()},
\code{\link{aemo_rooftop_pv}()},
\code{\link{aemo_spd_constraints}()}
}
\concept{dispatch}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bids.R
\name{aemo_bids}
\alias{aemo_bids}
\title{Generator bid stack}
\usage{
aemo_bids(
  duid,
  start,
  end,
  resolution = c("day", "period", "joined"),
  allow_large = FALSE
)
}
\arguments{
\item{duid}{Character vector of DUIDs. Required.}

\item{start, end}{Window.}

\item{resolution}{One of \code{"day"} (default, BIDDAYOFFER_D
only), \code{"period"} (BIDPEROFFER_D only), or \code{"joined"}
(BIDDAYOFFER_D inner-joined to BIDPEROFFER_D on
\verb{(duid, settlementdate, bidtype)}).}

\item{allow_large}{Logical. Default \code{FALSE}.}
}
\value{
An \code{aemo_tbl}.
}
\description{
Returns \code{BIDDAYOFFER_D} (daily bid summary: 10 price bands,
MaxAvail, fixed load, locked at 12:30 D-1), \code{BIDPEROFFER_D}
(per-interval availability and rebids), or the two joined on
\verb{(duid, settlementdate, bidtype)}.
}
\details{
\strong{Parent / child structure.} \code{BIDDAYOFFER_D} carries the
\strong{price bands} (\code{priceband1..priceband10}), which are locked
at 12:30 on the day ahead and cannot be rebid.
\code{BIDPEROFFER_D} carries the \strong{per-interval availability
bands} (\code{bandavail1..bandavail10}), which can be rebid
intraday. Serious bidding analysis (Goncalves & Menezes 2022
Energy Economics 113 106398; Nelson et al. 2024 AJARE 68(4))
needs both joined.

\strong{Size warning.} \code{BIDPEROFFER_D} monthly archives are
multi-gigabyte. By default \code{aemo_bids()} refuses spans longer
than 30 days; pass \code{allow_large = TRUE} to override.

\strong{Upstream gap.} AEMO has a documented gap in
\code{BIDPEROFFER_D} between March 2021 and July 2024. Rows in
that range may be missing.
}
\examples{
\donttest{
op <- options(aemo.cache_dir = tempdir())
try({
  now <- Sys.time()
  # Daily bid summary (price bands)
  b <- aemo_bids(duid = "BW01",
                  start = now - 86400, end = now)

  # Joined: price bands + per-interval volumes
  bj <- aemo_bids(duid = "BW01", start = now - 86400, end = now,
                   resolution = "joined")
})
options(op)
}
}
\seealso{
Other dispatch: 
\code{\link{aemo_constraints}()},
\code{\link{aemo_dispatch_units}()},
\code{\link{aemo_fcas_enablement}()},
\code{\link{aemo_gencon}()},
\code{\link{aemo_interconnector}()},
\code{\link{aemo_market_notices}()},
\code{\link{aemo_outages}()},
\code{\link{aemo_rooftop_pv}()},
\code{\link{aemo_spd_constraints}()}
}
\concept{dispatch}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dispatch.R
\name{aemo_dispatch_units}
\alias{aemo_dispatch_units}
\title{Per-DUID dispatch output}
\usage{
aemo_dispatch_units(
  duid = NULL,
  start,
  end,
  measure = c("scada_mw", "target_mw", "both")
)
}
\arguments{
\item{duid}{Optional character vector of DUIDs. \code{NULL} returns
all generators.}

\item{start, end}{Window (inclusive).}

\item{measure}{One of \code{"scada_mw"} (default), \code{"target_mw"},
or \code{"both"}.}
}
\value{
An \code{aemo_tbl} with columns \code{settlementdate}, \code{duid},
and the requested measure(s).
}
\description{
Returns 5-minute generator output for one or more DUIDs.
Three measures are available:
}
\details{
\itemize{
\item \code{"scada_mw"} (default): actual metered output from
\code{DISPATCH_UNIT_SCADA} (\code{SCADAVALUE}).
\item \code{"target_mw"}: dispatch target from \code{DISPATCHLOAD}
(\code{TOTALCLEARED}). This is the MW AEMO \emph{asked} the unit to
produce at the end of the interval.
\item \code{"both"}: returns \code{SCADAVALUE}, \code{INITIALMW} (SCADA at the
start of the interval) and \code{TOTALCLEARED} (target at the
end) in a single row per DUID per interval. Use this for
ramp-trajectory research: the ramp applied during the
interval is the straight line from \code{INITIALMW} to
\code{TOTALCLEARED}.
}

Timestamps are AEST (UTC+10, no DST).
}
\examples{
\donttest{
op <- options(aemo.cache_dir = tempdir())
try({
  now <- Sys.time()
  # SCADA actual
  d <- aemo_dispatch_units(duid = "BW01", start = now - 3600,
                            end = now)

  # Paired: INITIALMW, TOTALCLEARED, SCADAVALUE (ramp research)
  d_both <- aemo_dispatch_units(duid = "BW01",
                                 start = now - 3600, end = now,
                                 measure = "both")
})
options(op)
}
}
\seealso{
Other dispatch: 
\code{\link{aemo_bids}()},
\code{\link{aemo_constraints}()},
\code{\link{aemo_fcas_enablement}()},
\code{\link{aemo_gencon}()},
\code{\link{aemo_interconnector}()},
\code{\link{aemo_market_notices}()},
\code{\link{aemo_outages}()},
\code{\link{aemo_rooftop_pv}()},
\code{\link{aemo_spd_constraints}()}
}
\concept{dispatch}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.R
\name{aemo_throttle}
\alias{aemo_throttle}
\title{Configure request throttling}
\usage{
aemo_throttle(delay = 1)
}
\arguments{
\item{delay}{Numeric. Minimum delay between requests in
seconds. Internally this becomes \code{httr2::req_throttle(rate = 1/delay)}.}
}
\value{
Invisibly returns the previous value.
}
\description{
Controls the delay between successive NEMweb requests.
Defaults to 1 second (half a request per second).
}
\examples{
old <- aemo_throttle(0.5)  # 2 requests per second
aemo_throttle(old)         # restore
}
\seealso{
Other configuration: 
\code{\link{aemo_cache_info}()},
\code{\link{aemo_clear_cache}()}
}
\concept{configuration}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/forecasts.R
\name{aemo_predispatch}
\alias{aemo_predispatch}
\title{Price and demand forecasts (P5MIN, PREDISPATCH)}
\usage{
aemo_predispatch(
  region,
  start,
  end,
  horizon = c("predispatch", "p5min"),
  run_datetime = NULL
)
}
\arguments{
\item{region}{NEM region code.}

\item{start, end}{Window of forecast run-times.}

\item{horizon}{One of \code{"predispatch"} (default) or \code{"p5min"}.}

\item{run_datetime}{Optional character or \code{POSIXct}. A specific
\code{RUN_DATETIME} to pin to. \code{NULL} (default) returns every
vintage issued in \verb{[start, end]}.}
}
\value{
An \code{aemo_tbl}.
}
\description{
Returns AEMO's forecast prices and demand for a NEM region.
Two horizons:
\itemize{
\item \code{"p5min"}: 5-minute-ahead forecast, 12 intervals ahead,
published every 5 minutes.
\item \code{"predispatch"}: 40-hour-ahead predispatch at 30-minute
resolution, published every 30 minutes.
}
}
\details{
The 7-day predispatch publication was retired when 5-minute
settlement commenced; for longer horizons use \code{\link[=aemo_pasa]{aemo_pasa()}}.

\strong{Vintages.} PREDISPATCH and P5MIN forecasts are re-issued
every 30 or 5 minutes respectively. Every vintage is archived
by its \code{RUN_DATETIME}. By default (\code{run_datetime = NULL}) this
function returns all vintages whose \emph{file} timestamp falls in
\verb{[start, end]}, i.e. all the forecasts \emph{issued} during the
window. The \code{periodid} / \code{datetime} columns on the returned
rows give each forecast's \emph{target} time.

For forecast-error research (comparing a forecast vintage
against the realised dispatch) pass \code{run_datetime} to pin a
specific vintage:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{# The PREDISPATCH run issued at 15:00 on 1 June 2024 --
# 80 half-hour-ahead rows covering 15:30 out to 31:30 hours.
v <- aemo_predispatch("NSW1", start = "2024-06-01", end = "2024-06-02",
                      run_datetime = "2024-06-01 15:00")
}\if{html}{\out{</div>}}

This is the vintage-aware pattern used in Prakash (2023)
\emph{NEMSEER} (JOSS 8(92) 5883).
}
\examples{
\donttest{
op <- options(aemo.cache_dir = tempdir())
try({
  now <- Sys.time()
  p <- aemo_predispatch("NSW1", start = now - 3600, end = now)
  head(p)
})
options(op)
}
}
\seealso{
Other forecast: 
\code{\link{aemo_pasa}()}
}
\concept{forecast}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.R
\name{aemo_cache_info}
\alias{aemo_cache_info}
\title{Inspect the local aemo cache}
\usage{
aemo_cache_info()
}
\value{
A list with \code{dir}, \code{n_files}, \code{size_bytes}, \code{size_human}, \code{files}.
}
\description{
Inspect the local aemo cache
}
\examples{
\donttest{
op <- options(aemo.cache_dir = tempdir())
aemo_cache_info()
options(op)
}
}
\seealso{
Other configuration: 
\code{\link{aemo_clear_cache}()},
\code{\link{aemo_throttle}()}
}
\concept{configuration}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/price_caps.R
\name{aemo_price_caps}
\alias{aemo_price_caps}
\title{NEM market price caps, price floor, and CPT by financial year}
\usage{
aemo_price_caps()
}
\value{
An \code{aemo_tbl} with columns \code{year} (financial year,
\code{"YYYY-YY"}), \code{market_price_cap_aud_per_mwh},
\code{market_price_floor_aud_per_mwh},
\code{cumulative_price_threshold_aud},
\code{administered_price_cap_aud_per_mwh}, and \code{source}.
}
\description{
Returns a static reference table of the \strong{Market Price Cap
(MPC)}, \strong{Market Price Floor (MPF)}, \strong{Cumulative Price
Threshold (CPT)}, and \strong{Administered Price Cap (APC)} that
apply in each NEM financial year (1 July to 30 June). These
are set by the AEMC under NER 3.9.4 and indexed annually to
CPI.
}
\details{
Use this table to interpret spot-price extremes: any RRP
hitting the MPC (typically AUD 15,000 to 17,500 per MWh
depending on year) indicates a price cap event; when the
rolling-seven-day cumulative price in a region exceeds the
CPT (~AUD 1.5 million per MWh-equivalent), AEMO imposes the
APC (AUD 300/MWh) until the CPT falls back below the threshold.

The table covers 2015-16 onwards and is updated on each
package release. For the authoritative current values see
\url{https://www.aemc.gov.au/regulation/energy-rules/national-electricity-rules}
and the AEMO Reliability Settings publications.
}
\examples{
caps <- aemo_price_caps()
head(caps)
}
\seealso{
Other reference: 
\code{\link{aemo_dlf}()},
\code{\link{aemo_interconnectors}()},
\code{\link{aemo_mlf}()},
\code{\link{aemo_participants}()},
\code{\link{aemo_regions}()},
\code{\link{aemo_settlement}()},
\code{\link{aemo_snapshot}()},
\code{\link{aemo_units}()}
}
\concept{reference}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/snapshot.R
\name{aemo_snapshot}
\alias{aemo_snapshot}
\title{Snapshot provenance for an aemo_tbl}
\usage{
aemo_snapshot(x)
}
\arguments{
\item{x}{An \code{aemo_tbl} (or a list of them).}
}
\value{
A data frame with one row per table.
}
\description{
Returns a one-row-per-source provenance record for an
\code{aemo_tbl}, suitable for inclusion in a paper appendix, a
Zenodo deposit, or a CRAN-style data manifest. The snapshot
captures:
}
\details{
\itemize{
\item \code{title}: the human-readable title of the table;
\item \code{source}: the NEMweb / MMSDM URL family the table was drawn
from;
\item \code{licence}: the AEMO Copyright Permissions Notice (always);
\item \code{retrieved}: the \code{POSIXct} timestamp at which the table was
constructed;
\item \code{rows}, \code{cols}: observed dimensions;
\item \code{sha256}: a SHA-256 digest of the table's printed body,
stable across R versions and platforms.
}

The \code{sha256} column is what makes the snapshot \emph{pinnable}: if
the same query returns a different hash, the underlying data
has changed (or the row-order has, which is also worth
knowing). Pair with a git commit of the analysis script to
give a reader a closed reproducibility loop.
}
\examples{
x <- structure(
  data.frame(settlementdate = Sys.time(), region = "NSW1", rrp = 80),
  aemo_title = "Demo",
  aemo_source = "http://nemweb.com.au",
  aemo_licence = "AEMO Copyright Permissions Notice",
  aemo_retrieved = Sys.time(),
  class = c("aemo_tbl", "data.frame")
)
aemo_snapshot(x)
}
\seealso{
Other reference: 
\code{\link{aemo_dlf}()},
\code{\link{aemo_interconnectors}()},
\code{\link{aemo_mlf}()},
\code{\link{aemo_participants}()},
\code{\link{aemo_price_caps}()},
\code{\link{aemo_regions}()},
\code{\link{aemo_settlement}()},
\code{\link{aemo_units}()}
}
\concept{reference}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aemo_tbl.R
\name{print.aemo_tbl}
\alias{print.aemo_tbl}
\title{Print an aemo_tbl}
\usage{
\method{print}{aemo_tbl}(x, ...)
}
\arguments{
\item{x}{An \code{aemo_tbl}.}

\item{...}{Passed to the next print method.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Print an aemo_tbl
}
\examples{
x <- data.frame(settlementdate = Sys.time(), region = "NSW1", rrp = 80)
x <- structure(x, aemo_title = "Demo", aemo_source = "http://nemweb.com.au",
               aemo_licence = "AEMO Copyright Permissions Notice",
               aemo_retrieved = Sys.time(),
               class = c("aemo_tbl", "data.frame"))
print(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nemweb.R
\name{aemo_nemweb_download}
\alias{aemo_nemweb_download}
\title{Download a NEMweb zipped CSV to the local cache}
\usage{
aemo_nemweb_download(url, cache = TRUE)
}
\arguments{
\item{url}{A fully-qualified NEMweb URL (zipped CSV).}

\item{cache}{Logical. Reuse cached file if present.}
}
\value{
Path to the cached file.
}
\description{
Thin wrapper over the internal cache-aware downloader.
}
\examples{
\donttest{
op <- options(aemo.cache_dir = tempdir())
try({
  files <- aemo_nemweb_ls("/Reports/Current/DispatchIS_Reports/")
  if (nrow(files) > 0) {
    f <- aemo_nemweb_download(files$url[1])
    file.exists(f)
  }
})
options(op)
}
}
\seealso{
Other low-level: 
\code{\link{aemo_nemweb_ls}()}
}
\concept{low-level}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.R
\name{aemo_clear_cache}
\alias{aemo_clear_cache}
\title{Clear the aemo cache}
\usage{
aemo_clear_cache()
}
\value{
Invisibly returns \code{NULL}.
}
\description{
Clear the aemo cache
}
\examples{
\donttest{
op <- options(aemo.cache_dir = tempdir())
aemo_clear_cache()
options(op)
}
}
\seealso{
Other configuration: 
\code{\link{aemo_cache_info}()},
\code{\link{aemo_throttle}()}
}
\concept{configuration}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interconnector.R
\name{aemo_interconnector}
\alias{aemo_interconnector}
\title{NEM interconnector flows}
\usage{
aemo_interconnector(flow = NULL, start, end, intervention = FALSE)
}
\arguments{
\item{flow}{Optional character vector of interconnector IDs.}

\item{start, end}{Window.}

\item{intervention}{Logical. Default \code{FALSE}.}
}
\value{
An \code{aemo_tbl}.
}
\description{
Returns MW flows, losses, and limits for one or more NEM
interconnectors from \code{DISPATCHINTERCONNECTORRES}.
}
\details{
AEMO's \code{METEREDMWFLOW} is positive when power flows from
\code{REGIONFROM} to \code{REGIONTO}. For per-interconnector direction
conventions see \code{\link[=aemo_interconnectors]{aemo_interconnectors()}}.
}
\examples{
\donttest{
op <- options(aemo.cache_dir = tempdir())
try({
  now <- Sys.time()
  i <- aemo_interconnector(flow = "V-SA",
                            start = now - 3600, end = now)
  head(i)
})
options(op)
}
}
\seealso{
Other dispatch: 
\code{\link{aemo_bids}()},
\code{\link{aemo_constraints}()},
\code{\link{aemo_dispatch_units}()},
\code{\link{aemo_fcas_enablement}()},
\code{\link{aemo_gencon}()},
\code{\link{aemo_market_notices}()},
\code{\link{aemo_outages}()},
\code{\link{aemo_rooftop_pv}()},
\code{\link{aemo_spd_constraints}()}
}
\concept{dispatch}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reference.R
\name{aemo_participants}
\alias{aemo_participants}
\title{NEM market participants and their registered DUIDs}
\source{
AEMO NEMweb MMSDM archive, PARTICIPANT and
DUDETAILSUMMARY tables. AEMO Copyright Permissions Notice.
}
\usage{
aemo_participants()
}
\value{
An \code{aemo_tbl} with columns \code{participantid},
\code{participantclassid} (e.g. \code{GENERATOR}, \code{LOAD}, \code{TRADER}),
\code{name} (company name), \code{duid}, \code{stationid}, \code{regionid},
\code{dispatchtype}, \code{schedule_type}. Rows are one per
participant-DUID combination. If MMSDM is unreachable,
returns an empty table with a warning.
}
\description{
Returns a mapping of NEM market participants (companies) to
their registered Dispatchable Unit Identifiers (DUIDs), joined
from the MMSDM \code{PARTICIPANT} and \code{DUDETAILSUMMARY} tables.
Use this for corporate ownership analysis: rolling up generator
output or bids from DUID-level data to the company level.
}
\examples{
\donttest{
op <- options(aemo.cache_dir = tempdir())
try({
  pt <- aemo_participants()
  # DUIDs owned by AGL
  pt[grepl("AGL", pt$name, ignore.case = TRUE), ]
})
options(op)
}
}
\seealso{
\code{\link[=aemo_units]{aemo_units()}} for DUID-level registry without
participant mapping.

Other reference: 
\code{\link{aemo_dlf}()},
\code{\link{aemo_interconnectors}()},
\code{\link{aemo_mlf}()},
\code{\link{aemo_price_caps}()},
\code{\link{aemo_regions}()},
\code{\link{aemo_settlement}()},
\code{\link{aemo_snapshot}()},
\code{\link{aemo_units}()}
}
\concept{reference}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/outages.R
\name{aemo_outages}
\alias{aemo_outages}
\title{Planned and forced network outages}
\source{
AEMO NEMweb MMSDM archive, NETWORK_OUTAGEDETAIL
table. AEMO Copyright Permissions Notice.
}
\usage{
aemo_outages(start, end, element_id = NULL, outage_type = NULL, region = NULL)
}
\arguments{
\item{start, end}{Outage window (inclusive). Filters on
\code{starttime} and \code{endtime}: any outage active during the
window is returned. Character or POSIXct.}

\item{element_id}{Optional character vector of network
element IDs. \code{NULL} returns all elements.}

\item{outage_type}{Optional character. One of \code{"PLANNED"},
\code{"FORCED"}, or \code{NULL} (both). Case-insensitive.}

\item{region}{Optional NEM region code. Filters on the
region column where available.}
}
\value{
An \code{aemo_tbl} with columns from
NETWORK_OUTAGEDETAIL including \code{outageid}, \code{starttime},
\code{endtime}, \code{substationid}, \code{equipmenttype},
\code{equipmentid}, \code{outagetype}, \code{regionid}, and
\code{restarttimeunknown}. Exact column set depends on the
MMSDM version.
}
\description{
Returns the \code{NETWORK_OUTAGEDETAIL} table from MMSDM, which
records every planned and forced outage on NEM transmission
and distribution network elements. Outages are a primary
driver of binding constraints and price spikes: when a line
or transformer is out of service the network is more
constrained, reducing the thermal limits that appear as RHS
values in DISPATCHCONSTRAINT.
}
\details{
\strong{Use case.} Pair with \code{\link[=aemo_constraints]{aemo_constraints()}} to explain a
price spike: \code{aemo_outages()} tells you which elements were
off-service at the time; \code{aemo_constraints()} tells you
which constraints bound; together they answer "why was SA
spot price AUD 15,000 at 17:35?".
}
\examples{
\donttest{
op <- options(aemo.cache_dir = tempdir())
try({
  # Forced outages during a recent high-price period in SA
  o <- aemo_outages(
    start = "2024-03-01",
    end   = "2024-03-02",
    outage_type = "FORCED"
  )
  head(o)
})
options(op)
}
}
\seealso{
\code{\link[=aemo_constraints]{aemo_constraints()}} for the binding constraint
shadow prices that these outages drive.

Other dispatch: 
\code{\link{aemo_bids}()},
\code{\link{aemo_constraints}()},
\code{\link{aemo_dispatch_units}()},
\code{\link{aemo_fcas_enablement}()},
\code{\link{aemo_gencon}()},
\code{\link{aemo_interconnector}()},
\code{\link{aemo_market_notices}()},
\code{\link{aemo_rooftop_pv}()},
\code{\link{aemo_spd_constraints}()}
}
\concept{dispatch}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constraints.R
\name{aemo_gencon}
\alias{aemo_gencon}
\title{Generic constraint equations and RHS terms (GENCONDATA)}
\source{
AEMO NEMweb MMSDM archive, GENCONDATA table.
AEMO Copyright Permissions Notice.
}
\usage{
aemo_gencon(constraint_id = NULL, type = NULL)
}
\arguments{
\item{constraint_id}{Optional character vector of constraint
IDs to filter on. \code{NULL} returns all equations.}

\item{type}{Optional constraint type filter (e.g.
\code{"LE"} for \code{<=}, \code{"GE"} for \code{>=}, \code{"EQ"} for \code{=}).
\code{NULL} returns all types.}
}
\value{
An \code{aemo_tbl} with columns including \code{genconid}
(constraint ID), \code{constrainttype}, \code{description},
\code{genericconstraintrhs} (default RHS value). Additional
columns from GENCONDATA (effective dates, generic
constraint equation weights) will be present when
available.
}
\description{
Downloads the \code{GENCONDATA} table from the most recent MMSDM
monthly archive. \code{GENCONDATA} contains the equation
definitions for every generic constraint active in the NEM:
the constraint ID, the type (equality/inequality), a
description of what the constraint models (thermal limit,
voltage stability, system strength, etc.), and the default
RHS value.
}
\details{
Pair this with \code{\link[=aemo_constraints]{aemo_constraints()}} to go from a binding
dispatch interval to the underlying network equation. The
workflow is: \code{aemo_constraints()} tells you \emph{which} constraint
bound and \emph{how hard} (marginalvalue = shadow price);
\code{aemo_gencon()} tells you \emph{what the constraint is} (which
elements and why the RHS was set at that level).
}
\examples{
\donttest{
op <- options(aemo.cache_dir = tempdir())
try({
  # Find equations for the Heywood interconnector thermal limits
  g <- aemo_gencon(constraint_id = c("V::S_NIL_TBSE", "V::S_NIL_FCSPS"))
  head(g)
})
options(op)
}
}
\seealso{
\code{\link[=aemo_constraints]{aemo_constraints()}} for the real-time binding
constraint shadow prices.

Other dispatch: 
\code{\link{aemo_bids}()},
\code{\link{aemo_constraints}()},
\code{\link{aemo_dispatch_units}()},
\code{\link{aemo_fcas_enablement}()},
\code{\link{aemo_interconnector}()},
\code{\link{aemo_market_notices}()},
\code{\link{aemo_outages}()},
\code{\link{aemo_rooftop_pv}()},
\code{\link{aemo_spd_constraints}()}
}
\concept{dispatch}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/demand.R
\name{aemo_demand}
\alias{aemo_demand}
\title{Regional electricity demand}
\source{
AEMO NEMweb, AEMO Copyright Permissions Notice.
}
\usage{
aemo_demand(
  region,
  start,
  end,
  measure = c("operational", "operational_less_snsg", "native"),
  intervention = FALSE
)
}
\arguments{
\item{region}{NEM region code. Vector accepted.}

\item{start, end}{Window (inclusive), character or POSIXct.}

\item{measure}{One of \code{"operational"} (default),
\code{"operational_less_snsg"}, or \code{"native"}.}

\item{intervention}{Logical. Default \code{FALSE} filters to market
pricing runs.}
}
\value{
An \code{aemo_tbl} with columns \code{settlementdate},
\code{regionid}, \code{demand_mw} (the requested measure), plus the
underlying DISPATCHREGIONSUM columns used in the derivation.
}
\description{
Returns 5-minute regional demand from \code{DISPATCHREGIONSUM}.
Three demand measures are supported, aligned with AEMO's
Demand Terms taxonomy:
}
\details{
\itemize{
\item \code{"operational"} (default): \code{TOTALDEMAND}, the grid-measured
demand met by scheduled and semi-scheduled generation plus
net interchange. This is the quantity AEMO dispatches.
\item \code{"operational_less_snsg"}: \code{TOTALDEMAND} minus small
non-scheduled generation (SS_SOLAR_UIGF + SS_WIND_UIGF
where present).
\item \code{"native"}: \code{TOTALDEMAND} plus estimated rooftop PV
generation. Closest to end-use consumption. If the rooftop
PV component is not available in DISPATCHREGIONSUM the
function warns and returns \code{TOTALDEMAND}; users should join
with \code{aemo_rooftop_pv()} for a full native-demand estimate.
}

Timestamps are AEST (UTC+10, no DST).
}
\examples{
\donttest{
op <- options(aemo.cache_dir = tempdir())
try({
  now <- Sys.time()
  d <- aemo_demand("VIC1", now - 3600, now)
  head(d)
})
options(op)
}
}
\concept{demand}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rooftop_pv.R
\name{aemo_rooftop_pv}
\alias{aemo_rooftop_pv}
\title{Rooftop PV actuals and forecasts}
\usage{
aemo_rooftop_pv(region, start, end, type = c("actual", "forecast"))
}
\arguments{
\item{region}{NEM region code.}

\item{start, end}{Window.}

\item{type}{One of \code{"actual"} (default) or \code{"forecast"}.}
}
\value{
An \code{aemo_tbl}.
}
\description{
Returns AEMO's region-level estimate of rooftop PV generation,
either actuals or forecasts. Published at 30-minute resolution.
}
\details{
The "actual" figure is an AEMO estimate derived from the APVI
sampling model and weather data, not metered SCADA output. It
is the best available public measure of aggregate rooftop PV
generation but is subject to revision.
}
\examples{
\donttest{
op <- options(aemo.cache_dir = tempdir())
try({
  now <- Sys.time()
  r <- aemo_rooftop_pv("NSW1",
                        start = now - 3600, end = now)
  head(r)
})
options(op)
}
}
\seealso{
Other dispatch: 
\code{\link{aemo_bids}()},
\code{\link{aemo_constraints}()},
\code{\link{aemo_dispatch_units}()},
\code{\link{aemo_fcas_enablement}()},
\code{\link{aemo_gencon}()},
\code{\link{aemo_interconnector}()},
\code{\link{aemo_market_notices}()},
\code{\link{aemo_outages}()},
\code{\link{aemo_spd_constraints}()}
}
\concept{dispatch}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/price.R
\name{aemo_price}
\alias{aemo_price}
\title{Wholesale electricity prices}
\source{
AEMO NEMweb \url{http://nemweb.com.au}, AEMO Copyright
Permissions Notice.
}
\usage{
aemo_price(
  region,
  start,
  end,
  interval = c("5min", "30min"),
  market = c("energy", "fcas"),
  intervention = FALSE
)
}
\arguments{
\item{region}{One of \code{"NSW1"}, \code{"QLD1"}, \code{"SA1"}, \code{"TAS1"},
\code{"VIC1"}. Accepts a vector.}

\item{start, end}{Start and end times (inclusive). Character
(parsed as AEST) or \code{POSIXct}.}

\item{interval}{One of \code{"5min"} (default) or \code{"30min"}.}

\item{market}{One of \code{"energy"} (default, returns RRP) or
\code{"fcas"} (returns the FCAS service RRPs).}

\item{intervention}{Logical. \code{FALSE} (default) returns only
the market pricing run; \code{TRUE} returns both market and
physical runs, with the \code{intervention} column preserved.}
}
\value{
An \code{aemo_tbl}. Key columns include \code{settlementdate}
(POSIXct AEST), \code{regionid}, \code{rrp} (AUD/MWh, energy) or the
FCAS service RRPs (AUD/MW), and \code{intervention}.
}
\description{
Returns 5-minute dispatch prices or 30-minute trading prices
for a NEM region over a specified window. Filters intervention
runs by default so the returned prices are the market clearing
prices used in settlement.
}
\details{
\strong{Timestamps} are AEST (UTC+10, no daylight savings) to match
AEMO's market clock. See the package-level documentation for
the period-ending timestamp convention (a row stamped 00:05 is
the 5-minute period ending at 00:05).

\strong{Intervention.} \code{DISPATCHPRICE} contains both market pricing
runs (\code{INTERVENTION = 0}) and physical / intervention runs
(\code{INTERVENTION = 1}). The default filters to market runs.
Pass \code{intervention = TRUE} to get both.

\strong{30-minute settlement and the 5MS transition.} Before
1 October 2021 the NEM settled on 30-minute trading prices
from \code{TRADINGPRICE} (TRADINGIS). On 1 October 2021 five-minute
settlement (5MS) commenced and settlement moved to native
5-minute prices. When \code{interval = "30min"}:
\itemize{
\item For the pre-5MS period (\code{start < 2021-10-01}): prices are
read from TRADINGIS (\code{TRADINGPRICE}).
\item For the post-5MS period: prices are derived by taking the
arithmetic mean of the six 5-minute dispatch prices within
each 30-minute trading interval, consistent with how AEMO
calculates the \code{TRADINGPRICE} column in TradingIS post-5MS.
}

\strong{Data availability.} NEMweb Current-directory files retain
the last ~30 days of 5-minute dispatch files. Historical
queries use the Archive daily-rollup files automatically; for
queries older than the Archive window, use
\code{\link[=aemo_nemweb_download]{aemo_nemweb_download()}} with an MMSDM URL directly.
}
\examples{
\donttest{
op <- options(aemo.cache_dir = tempdir())
try({
  now <- Sys.time()
  p <- aemo_price("NSW1", now - 3600, now)
  head(p)
})
options(op)
}
}
\seealso{
Other price: 
\code{\link{aemo_fcas}()}
}
\concept{price}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settlement.R
\name{aemo_settlement}
\alias{aemo_settlement}
\title{Settlement cash-flow and residue tables}
\source{
AEMO NEMweb MMSDM archive, SETCFM /
SETFCASREGIONRECOVERY / SETRESIDUECONTRACTPAYMENT.
AEMO Copyright Permissions Notice.
}
\usage{
aemo_settlement(table = c("cashflow", "fcas_recovery", "residues"), start, end)
}
\arguments{
\item{table}{One of \code{"cashflow"}, \code{"fcas_recovery"}, or
\code{"residues"}.}

\item{start, end}{Window (inclusive) applied to
\code{SETTLEMENTDATE} / \code{PAYMENTDATE}.}
}
\value{
An \code{aemo_tbl} with the raw MMSDM columns for the
requested table.
}
\description{
Returns settlement reconciliation tables from MMSDM. Three
views are exposed, corresponding to the three tables gentailer
hedging workflows most commonly need:
}
\details{
\itemize{
\item \code{"cashflow"} (default): \code{SETCFM} (NEMDE-derived energy
settlement amounts per participant per trading interval).
\item \code{"fcas_recovery"}: \code{SETFCASREGIONRECOVERY} (the recovery
allocation of FCAS costs to customer load per region per
trading interval).
\item \code{"residues"}: \code{SETRESIDUECONTRACTPAYMENT} (settlement
residue auction (SRA) contract payments against
interconnector settlement residues).
}

\strong{Access.} These tables are in the MMSDM monthly archive
(not the NEMweb Current retention). Expect two-month latency
between trading date and availability.
}
\examples{
\donttest{
op <- options(aemo.cache_dir = tempdir())
try({
  s <- aemo_settlement(table = "cashflow",
                        start = "2024-06-01",
                        end   = "2024-06-02")
  head(s)
})
options(op)
}
}
\seealso{
\code{\link[=aemo_mlf]{aemo_mlf()}} for the transmission loss factors that
scale settlement amounts; \code{\link[=aemo_interconnector]{aemo_interconnector()}} for the
flow side of the residue calculation.

Other reference: 
\code{\link{aemo_dlf}()},
\code{\link{aemo_interconnectors}()},
\code{\link{aemo_mlf}()},
\code{\link{aemo_participants}()},
\code{\link{aemo_price_caps}()},
\code{\link{aemo_regions}()},
\code{\link{aemo_snapshot}()},
\code{\link{aemo_units}()}
}
\concept{reference}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reference.R
\name{aemo_regions}
\alias{aemo_regions}
\title{NEM regions}
\usage{
aemo_regions()
}
\value{
An \code{aemo_tbl} with columns \code{region}, \code{name}, \code{state},
\code{wall_timezone}, \code{market_timezone}, \code{commenced}.
}
\description{
Returns a static table of the five NEM regions with metadata.
The \code{market_timezone} column is the AEMO market clock (AEST,
UTC+10, no DST, year-round) that applies to every timestamp
in NEMweb files; \code{wall_timezone} is the local civil time zone
consumers experience (observes DST in NSW/VIC/TAS/SA).
}
\examples{
aemo_regions()
}
\seealso{
Other reference: 
\code{\link{aemo_dlf}()},
\code{\link{aemo_interconnectors}()},
\code{\link{aemo_mlf}()},
\code{\link{aemo_participants}()},
\code{\link{aemo_price_caps}()},
\code{\link{aemo_settlement}()},
\code{\link{aemo_snapshot}()},
\code{\link{aemo_units}()}
}
\concept{reference}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aemo-package.R
\docType{package}
\name{aemo-package}
\alias{aemo}
\alias{aemo-package}
\title{aemo: Australian Energy Market Operator Data}
\description{
Tidy R access to AEMO's public data: wholesale electricity
prices, demand, dispatch, interconnector flows, rooftop PV,
bids, predispatch and PASA forecasts, FCAS markets, and gas
markets across the NEM.
}
\section{Main function families}{

\itemize{
\item \code{\link[=aemo_price]{aemo_price()}}, \code{\link[=aemo_demand]{aemo_demand()}}: workhorse price and demand.
\item \code{\link[=aemo_dispatch_units]{aemo_dispatch_units()}}, \code{\link[=aemo_interconnector]{aemo_interconnector()}},
\code{\link[=aemo_rooftop_pv]{aemo_rooftop_pv()}}: dispatch and flow.
\item \code{\link[=aemo_bids]{aemo_bids()}}, \code{\link[=aemo_predispatch]{aemo_predispatch()}}, \code{\link[=aemo_pasa]{aemo_pasa()}}: bids
and forecasts.
\item \code{\link[=aemo_fcas]{aemo_fcas()}}, \code{\link[=aemo_gas]{aemo_gas()}}: ancillary services and gas.
\item \code{\link[=aemo_units]{aemo_units()}}, \code{\link[=aemo_regions]{aemo_regions()}}, \code{\link[=aemo_interconnectors]{aemo_interconnectors()}}:
reference data.
\item \code{\link[=aemo_nemweb_ls]{aemo_nemweb_ls()}}, \code{\link[=aemo_nemweb_download]{aemo_nemweb_download()}}: low-level
escape hatches for arbitrary NEMweb paths.
\item \code{\link[=aemo_cache_info]{aemo_cache_info()}}, \code{\link[=aemo_clear_cache]{aemo_clear_cache()}}, \code{\link[=aemo_throttle]{aemo_throttle()}}:
configuration.
}
}

\section{Data source}{

Data is published by AEMO at \url{http://nemweb.com.au} under the
AEMO Copyright Permissions Notice
\url{https://www.aemo.com.au/privacy-and-legal-notices/copyright-permissions}.
}

\seealso{
Useful links:
\itemize{
  \item \url{https://charlescoverdale.github.io/aemo/}
  \item \url{https://github.com/charlescoverdale/aemo}
  \item Report bugs at \url{https://github.com/charlescoverdale/aemo/issues}
}

}
\author{
\strong{Maintainer}: Charles Coverdale \email{charlesfcoverdale@gmail.com}

}
\concept{AEMO}
\concept{NEM}
\concept{electricity market}
\concept{wholesale electricity}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loss_factors.R
\name{aemo_mlf}
\alias{aemo_mlf}
\title{Marginal Loss Factors (MLF) by DUID and financial year}
\source{
AEMO MMSDM archive, TRANSMISSIONLOSSFACTOR table.
AEMO Copyright Permissions Notice.
}
\usage{
aemo_mlf(year = NULL, duid = NULL)
}
\arguments{
\item{year}{Financial year(s) as \code{"YYYY-YY"} strings (e.g.
\code{"2024-25"}). \code{NULL} returns all available years. Multiple
years accepted.}

\item{duid}{Optional character vector of DUIDs to filter on.
\code{NULL} returns all DUIDs. See \code{\link[=aemo_units]{aemo_units()}} for the full
DUID registry.}
}
\value{
An \code{aemo_tbl} with columns \code{financial_year}, \code{duid},
\code{connectionpointid}, \code{regionid}, \code{mlf}. From a live MMSDM
download additional columns (\code{participantid},
\code{lastchanged}) may also be present.
}
\description{
Returns the Marginal Loss Factor applicable to each DUID for
the requested financial year(s). MLFs measure the incremental
network loss at a connection point relative to the Regional
Reference Node (RRN); a DUID with MLF = 0.97 receives 97\% of
the regional RRP per MWh generated.
}
\details{
MLFs are published annually by AEMO under NER 3.6.2 and are
used in settlement calculations and in PPA revenue
reconstruction. The function first attempts to download the
\code{TRANSMISSIONLOSSFACTOR} table from the most recent MMSDM monthly
archive; if that fails it returns a bundled static table
covering FY 2020-21 to FY 2025-26 for ~20 well-known DUIDs.
}
\examples{
\donttest{
op <- options(aemo.cache_dir = tempdir())
try({
  mlf <- aemo_mlf(year = "2024-25")
  head(mlf)
})
options(op)
}
}
\seealso{
\code{\link[=aemo_units]{aemo_units()}} for the DUID registry,
\code{\link[=aemo_price]{aemo_price()}} for regional RRPs.

Other reference: 
\code{\link{aemo_dlf}()},
\code{\link{aemo_interconnectors}()},
\code{\link{aemo_participants}()},
\code{\link{aemo_price_caps}()},
\code{\link{aemo_regions}()},
\code{\link{aemo_settlement}()},
\code{\link{aemo_snapshot}()},
\code{\link{aemo_units}()}
}
\concept{reference}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nemweb.R
\name{aemo_nemweb_ls}
\alias{aemo_nemweb_ls}
\title{List files in a NEMweb directory}
\source{
AEMO NEMweb \url{http://nemweb.com.au}, published under
the AEMO Copyright Permissions Notice.
}
\usage{
aemo_nemweb_ls(path)
}
\arguments{
\item{path}{NEMweb subpath (e.g. \code{"/Reports/Current/DispatchIS_Reports/"}).
Leading and trailing slashes are optional.}
}
\value{
A data frame with \code{name}, \code{modified}, \code{size}, \code{url}.
}
\description{
Returns a data frame of files in a NEMweb path, parsed from
the Apache directory-listing HTML.
}
\examples{
\donttest{
op <- options(aemo.cache_dir = tempdir())
try({
  files <- aemo_nemweb_ls("/Reports/Current/DispatchIS_Reports/")
  head(files)
})
options(op)
}
}
\seealso{
Other low-level: 
\code{\link{aemo_nemweb_download}()}
}
\concept{low-level}
